
#include "headers.hpp"
#include "dsplib.cpp"
#include "options.cpp"

#include <cctype>

static void setIRLength(std::vector< std::vector< double > >& irBufs,
                        bool powOfTwo = false, bool doubleLen = false)
{
  size_t  maxLen = 2;
  for (size_t i = 0; i < irBufs.size(); i++)
  {
    for (size_t j = irBufs[i].size(); j-- > 0; )
    {
      if (irBufs[i][j] != 0.0)
      {
        if (!powOfTwo)
        {
          maxLen = ((j + 1) > maxLen ? (j + 1) : maxLen);
        }
        else
        {
          while (maxLen < (j + 1))
            maxLen = maxLen << 1;
        }
        break;
      }
    }
  }
  if (doubleLen)
    maxLen = maxLen << 1;
  for (size_t i = 0; i < irBufs.size(); i++)
  {
    if (irBufs[i].size() != maxLen)
      irBufs[i].resize(maxLen, 0.0);
  }
}

static void inverseIR(std::vector< double >& buf, int dynRangeDB)
{
  realFFT(&(buf.front()), buf.size(), false);
  double  sqMaxMag = 0.00000001;
  for (size_t i = 0; i < 2; i++)
  {
    double  sqR = buf[i] * buf[i];
    sqMaxMag = (sqR > sqMaxMag ? sqR : sqMaxMag);
  }
  for (size_t i = 2; i < buf.size(); i += 2)
  {
    double  sqR = (buf[i] * buf[i]) + (buf[i + 1] * buf[i + 1]);
    sqMaxMag = (sqR > sqMaxMag ? sqR : sqMaxMag);
  }
  double  dynRange = std::pow(10.0, double(dynRangeDB) / 10.0);
  double  powOffs = sqMaxMag / (dynRange - 1.0);
  double  powGain = (sqMaxMag + powOffs) / sqMaxMag;
  for (size_t i = 0; i < 2; i++)
  {
    buf[i] = std::sqrt(powGain / (buf[i] * buf[i] + powOffs));
  }
  for (size_t i = 2; i < buf.size(); i += 2)
  {
    double  sqR = (buf[i] * buf[i]) + (buf[i + 1] * buf[i + 1]);
    double  ph = -(std::atan2(buf[i + 1], buf[i]));
    double  r = std::sqrt(powGain / (sqR + powOffs));
    buf[i] = r * std::cos(ph);
    buf[i + 1] = r * std::sin(ph);
  }
  realFFT(&(buf.front()), buf.size(), true);
}

static void linearPhaseIR(std::vector< double >& buf)
{
  realFFT(&(buf.front()), buf.size(), false);
  if (buf.size() == 2)
    buf[1] = -(buf[1]);
  for (size_t i = 2; i < buf.size(); i += 2)
  {
    double  r = std::sqrt((buf[i] * buf[i]) + (buf[i + 1] * buf[i + 1]));
    buf[i] = ((i & 2) == 0 ? r : (-r));
    buf[i + 1] = 0.0;
  }
  realFFT(&(buf.front()), buf.size(), true);
}

int main(int argc, char **argv)
{
  try
  {
    std::string inFileName;
    std::string outFileName;
    std::string irFileName;
    double      isrAdj = 1.0;
    double      startPos = 0.0;
    double      duration = 0.0;
    bool        gainFlag = false;
    bool        inverseFilter = false;
    bool        linearPhaseFilter = false;
    bool        minimumPhaseFilter = false;
    int         outFmt = 0;
    int         ditherSeed = 0;
    int         invFiltDR = 40;
    double      gain = 1.0;
    double      channelGain[16] =
    {
      1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
      1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0
    };
    static const int  formatTable[6] =
    {
      SF_FORMAT_WAV | SF_FORMAT_FLOAT,
      SF_FORMAT_WAV | SF_FORMAT_PCM_U8,
      SF_FORMAT_WAV | SF_FORMAT_PCM_16,
      SF_FORMAT_WAV | SF_FORMAT_PCM_24,
      SF_FORMAT_WAV | SF_FORMAT_PCM_32,
      SF_FORMAT_WAV | SF_FORMAT_DOUBLE
    };
    {
      CommandLineParser opts(argv[0]);
      opts.addFileName("INFILE", "input file name", inFileName);
      opts.addFileName("OUTFILE", "output file name", outFileName);
      opts.addFileName("IRFILE",
                       "impulse response file name; can be a binary number "
                       "to use a\n"
                       "        maximum length sequence as the impulse "
                       "response (e.g. 10100 will\n"
                       "        generate a 31-sample sequence with the "
                       "polynomial x^5+x^3+1), or\n"
                       "        LPF::FREQ,LEN or HPF::FREQ,LEN for a linear "
                       "phase lowpass or\n"
                       "        highpass filter with a -6 dB cutoff frequency "
                       "of FREQ (in Hz)\n"
                       "        and length of LEN (must be even) samples or "
                       "-LEN microseconds",
                       irFileName);
      opts.addOption("format", "f",
                     "output sample format (0: float, 1/2/3/4: "
                     "8/16/24/32 bit PCM,\n"
                     "        5: double)",
                     outFmt, 0, 5);
      opts.addOption("dither-seed", "ds",
                     "random seed for dithering integer sample formats; zero "
                     "seeds\n"
                     "        from the current time in seconds, -1 disables "
                     "dither",
                     ditherSeed, -1, 0x7FFFFFFE);
      opts.addOption("isrm", "m",
                     "factor to multiply assumed input sample rate by; "
                     "it only affects\n"
                     "        the resample ratio, but not the conversion of "
                     "skip time to\n"
                     "        samples",
                     isrAdj, 0.01, 100.0);
      opts.addOption("skip", "k",
                     "start time in seconds (a negative value delays the "
                     "output)",
                     startPos);
      opts.addOption("duration", "d",
                     "if non-zero, duration limit in seconds, or "
                     "-sample frames of\n"
                     "        output",
                     duration);
      opts.addOption("inverse-filter", "inv", "", inverseFilter);
      opts.addOption("linear-phase", "lp", "", linearPhaseFilter);
      opts.addOption("minimum-phase", "mp",
                     "transform impulse response to create an inverse, "
                     "linear phase,\n"
                     "        or minimum phase filter. Enabling any of these "
                     "will pad the\n"
                     "        impulse response to a power of two length, "
                     "except for an inverse\n"
                     "        MLS, which simply reverses the sequence, and "
                     "applies a 1/length\n"
                     "        gain. If more than one transform is enabled, "
                     "they are applied in\n"
                     "        -inv, -lp, -mp order.\n"
                     "        Creating an inverse filter doubles the impulse "
                     "response length,\n"
                     "        and a linear phase inverse filter is also "
                     "windowed. An inverse\n"
                     "        minimum phase filter is first converted to "
                     "linear phase and\n"
                     "        windowed",
                     minimumPhaseFilter);
      opts.addOption("inverse-filter-range", "idr",
                     "limit the dB range of the magnitude response of an "
                     "inverse\n"
                     "        filter",
                     invFiltDR, 10, 120);
      opts.addOption("gain", "g", "overall gain (1.0 = 0 dB) for all channels",
                     gain, -1000.0, 1000.0);
      opts.addOption("channel1-gain", "g1", "",
                     channelGain[0], -1000.0, 1000.0);
      opts.addOption("channel2-gain", "g2", "",
                     channelGain[1], -1000.0, 1000.0);
      opts.addOption("channel3-gain", "g3", "",
                     channelGain[2], -1000.0, 1000.0);
      opts.addOption("channel4-gain", "g4", "",
                     channelGain[3], -1000.0, 1000.0);
      opts.addOption("channel5-gain", "g5", "",
                     channelGain[4], -1000.0, 1000.0);
      opts.addOption("channel6-gain", "g6", "",
                     channelGain[5], -1000.0, 1000.0);
      opts.addOption("channel7-gain", "g7", "",
                     channelGain[6], -1000.0, 1000.0);
      opts.addOption("channel8-gain", "g8", "gain for channel 1 to 8",
                     channelGain[7], -1000.0, 1000.0);
      if (!opts.parseCommandLine(argc, argv))
        return 0;
    }
    std::vector< std::vector< double > >  irBufs;
    double  irSampleRate = -1.0;
    int     irChannels = 0;
    bool    mlsFlag = false;
    if (irFileName.length() >= 5 &&
        irFileName[3] == ':' && irFileName[4] == ':')
    {
      if ((std::tolower(irFileName[0]) == 'l' ||
           std::tolower(irFileName[0]) == 'h') &&
          std::tolower(irFileName[1]) == 'p' &&
          std::tolower(irFileName[2]) == 'f')
      {
        bool    isHighpass = (std::tolower(irFileName[0]) == 'h');
        const char  *p = irFileName.c_str() + 5;
        char    *endp = (char *) 0;
        double  freq = std::strtod(p, &endp);
        if (!endp || endp == p || *endp != ',')
        {
          throw std::runtime_error(std::string("missing or invalid "
                                               "filter parameters"));
        }
        p = endp + 1;
        endp = (char *) 0;
        long    len = std::strtol(p, &endp, 0);
        if (!endp || endp == p || *endp != '\0' ||
            len < -10000000L || len > 4194304L || len == 0L ||
            (len >= 0L && (len & 1L) != 0L))
        {
          throw std::runtime_error(std::string("missing or invalid "
                                               "filter parameters"));
        }
        {
          AudioFilter_FileInput f(inFileName.c_str());
          irSampleRate = f.getSampleRate();
        }
        if (len < 0L)
        {
          len = long(double(len) * irSampleRate * -0.000001 + 0.5);
          len = len + long(!len);
          len = len + (len & 1L);
        }
        irChannels = 1;
        irBufs.resize(1);
        createLowpassIR(irBufs[0], size_t(len), freq / irSampleRate,
                        (isHighpass ? -1.0 : 1.0));
        if (isHighpass)
          irBufs[0][len >> 1] = irBufs[0][len >> 1] + 1.0;
      }
    }
    if (!irChannels)
    {
      mlsFlag = true;
      unsigned int  mlsPoly = 0U;
      int     tapCnt = 0;
      int     maxTap = 0;
      for (size_t i = 0; i < irFileName.length(); i++)
      {
        if (irFileName[i] != '0' && irFileName[i] != '1')
        {
          mlsFlag = false;
          break;
        }
        tapCnt += (int(irFileName[i]) & 1);
        mlsPoly = (mlsPoly << 1) | ((unsigned int) irFileName[i] & 1U);
        if (mlsPoly >= 0x400000U)
          throw std::runtime_error(std::string("MLS length is out of range"));
        maxTap += int(bool(mlsPoly));
      }
      if (mlsFlag)
      {
        if (!tapCnt || (tapCnt & 1) != 0)
          throw std::runtime_error(std::string("invalid MLS polynomial"));
        for (int i = 2; i <= maxTap; i++)
        {
          int     d = 0;
          for (int j = i; j <= maxTap; j += i)
            d += int(bool(mlsPoly & (1U << (j - 1))));
          if (d == tapCnt)
            throw std::runtime_error(std::string("invalid MLS polynomial"));
        }
        irChannels = 1;
        irBufs.resize(1);
        size_t  mlsLen = size_t((1 << maxTap) - 1);
        double  y0 = -1.0;
        double  y1 = 1.0;
        if (inverseFilter)
        {
          y1 = 1.0 / double(long(mlsLen));
          y0 = -y1;
        }
        irBufs[0].resize(mlsLen, y0);
        unsigned int  sr = 0xFFFFFFFFU;
        for (size_t i = 0; i < mlsLen; i++)
        {
          if (sr & 1U)
            irBufs[0][(inverseFilter ? (mlsLen - 1 - i) : i)] = y1;
          unsigned int  tmp = sr & mlsPoly;
          tmp = tmp ^ (tmp >> 16);
          tmp = tmp ^ (tmp >> 8);
          tmp = tmp ^ (tmp >> 4);
          tmp = tmp ^ (tmp >> 2);
          tmp = tmp ^ (tmp >> 1);
          sr = (sr << 1) | (tmp & 1U);
        }
      }
    }
    if (!irChannels)
    {
      AudioFilter_FileInput irFile(irFileName.c_str());
      if (irFile.getFileLength() > 4194304)
        throw std::runtime_error(std::string("impulse response is too long"));
      irSampleRate = irFile.getSampleRate();
      irChannels = irFile.getChannelCount();
      irBufs.resize(size_t(irChannels));
      for (size_t i = 0; i < irBufs.size(); i++)
        irBufs[i].reserve(size_t(irFile.getFileLength()));
      for (size_t i = 0; i < size_t(irFile.getFileLength()); i++)
      {
        irFile.process();
        for (int j = 0; j < irChannels; j++)
          irBufs[j].push_back(irFile.getOutputBuffer()[j]);
      }
    }
    AudioFilter_ResampledFileInput  inFile(inFileName.c_str(), irSampleRate,
                                           isrAdj, startPos, false);
    double  sampleRate = inFile.getSampleRate();
    sf_count_t  maxFrames = -1;
    if (duration > 0.0)
      maxFrames = sf_count_t(duration * sampleRate + 0.5);
    else if (duration < 0.0)
      maxFrames = sf_count_t(-duration + 0.5);
    if (mlsFlag)
      inverseFilter = false;
    else if (inverseFilter && minimumPhaseFilter)
      linearPhaseFilter = true;
    setIRLength(irBufs, inverseFilter | linearPhaseFilter | minimumPhaseFilter,
                inverseFilter);
    for (size_t i = 0; i < irBufs.size(); i++)
    {
      if (inverseFilter)
        inverseIR(irBufs[i], invFiltDR);
      if (linearPhaseFilter)
      {
        linearPhaseIR(irBufs[i]);
        if (inverseFilter)
        {
          CosineGen w(double(long(irBufs[i].size())));
          for (size_t j = 0; j < irBufs[i].size(); j++)
            irBufs[i][j] = irBufs[i][j] * cosineToNuttallWindow(w.sample());
        }
      }
      if (minimumPhaseFilter)
        minimumPhaseIR(irBufs[i]);
    }
    AudioFilter_FIRFilter firFilter(inFile, irBufs);
    int     channels = firFilter.getChannelCount();
    if (channels > int(sizeof(channelGain) / sizeof(double)))
    {
      throw std::runtime_error(std::string("input file has unsupported "
                                           "number of channels"));
    }
    for (int i = 0; i < channels; i++)
    {
      if (gain != 1.0 || channelGain[i] != 1.0)
      {
        gainFlag = true;
        channelGain[i] = channelGain[i] * gain;
      }
    }
    AudioOutputFile outFile(outFileName.c_str(), sampleRate, channels,
                            formatTable[outFmt], ditherSeed);
    if (!gainFlag)
    {
      if (maxFrames >= 0)
      {
        while (maxFrames > 0 && firFilter.process())
        {
          maxFrames--;
          outFile.writeSampleFrame(&(firFilter.getOutputBuffer().front()));
        }
      }
      else
      {
        while (firFilter.process())
        {
          outFile.writeSampleFrame(&(firFilter.getOutputBuffer().front()));
        }
      }
    }
    else
    {
      double  tmpBuf[16];
      const double  *bufrp = &(firFilter.getOutputBuffer().front());
      if (maxFrames >= 0)
      {
        while (maxFrames > 0 && firFilter.process())
        {
          maxFrames--;
          int     i = 0;
          do
          {
            tmpBuf[i] = bufrp[i] * channelGain[i];
          }
          while (++i < channels);
          outFile.writeSampleFrame(&(tmpBuf[0]));
        }
      }
      else
      {
        while (firFilter.process())
        {
          int     i = 0;
          do
          {
            tmpBuf[i] = bufrp[i] * channelGain[i];
          }
          while (++i < channels);
          outFile.writeSampleFrame(&(tmpBuf[0]));
        }
      }
    }
  }
  catch (std::exception& e)
  {
    std::fprintf(stderr, "%s: %s\n", argv[0], e.what());
    return -1;
  }
  return 0;
}

