
#ifndef DSPLIB_CPP_INCLUDED
#define DSPLIB_CPP_INCLUDED

#include "headers.hpp"
#include "dsplib.hpp"

#include <ctime>

// ----------------------------------------------------------------------------

SineGen::SineGen(double n)
{
  long double omega = twoPiL / n;
  s1 = 0.0L;
  s2 = -(std::sin(omega));
  c = std::cos(omega) * 2.0L;
}

SineGen::SineGen(double n, double p)
{
  long double omega = twoPiL / n;
  s1 = std::sin((long double) p);
  s2 = std::sin((long double) p - omega);
  c = std::cos(omega) * 2.0L;
}

CosineGen::CosineGen(double n)
{
  long double omega = twoPiL / n;
  s1 = 1.0L;
  s2 = std::cos(omega);
  c = s2 * 2.0L;
}

CosineGen::CosineGen(double n, double p)
{
  long double omega = twoPiL / n;
  s1 = std::cos((long double) p);
  c = std::cos(omega);
  // this can be compiled using the FSINCOS instruction,
  // and is faster than std::cos(p - omega)
  s2 = (s1 * c) + (std::sin((long double) p) * std::sin(omega));
  c *= 2.0L;
}

// ============================================================================
// |                               FFT LIBRARY                                |
// ============================================================================

FFTSineGen::FFTSineGen(long n)
{
  phase_re = 1.0L;
  phase_im = 0.0L;
  long double   omega = -twoPiL / (long double) n;
  omega_re = std::cos(omega);
  omega_im = std::sin(omega);
}

// ----------------------------------------------------------------------------

// called by complexFFT() to calculate FFT
// the complex data in 'buf' is expected to be in bit-reversed order,
// and 'n' should be greater than 1; 'invFFTLen' is the full length (N)
// for inverse FFT, and zero otherwise

static void recursiveFFT(double *buf, size_t n, int invFFTLen)
{
  if (n > 512)
  {
    // recursively partition large FFT buffers to improve cache efficiency
    recursiveFFT(buf, n >> 1, invFFTLen);
    recursiveFFT(buf + n, n >> 1, invFFTLen);
  }
  else
  {
    double  *endp = buf + (n << 1);
    double  *bufp = buf;
    if (!invFFTLen)
    {
      // optimized code for small values of 'm' (2 and 4)
      do
      {
        double  r1 = bufp[0];
        double  i1 = bufp[1];
        double  r2 = bufp[2];
        double  i2 = bufp[3];
        bufp[0] = r1 + r2;
        bufp[1] = i1 + i2;
        bufp[2] = r1 - r2;
        bufp[3] = i1 - i2;
        bufp = bufp + 4;
      }
      while (bufp < endp);
      if (n < 4)
        return;
      bufp = buf;
      do
      {
        double  r1 = bufp[0];
        double  i1 = bufp[1];
        double  r2 = bufp[4];
        double  i2 = bufp[5];
        bufp[0] = r1 + r2;
        bufp[1] = i1 + i2;
        bufp[4] = r1 - r2;
        bufp[5] = i1 - i2;
        r1 = bufp[2];
        i1 = bufp[3];
        r2 = bufp[7];
        i2 = -(bufp[6]);
        bufp[2] = r1 + r2;
        bufp[3] = i1 + i2;
        bufp[6] = r1 - r2;
        bufp[7] = i1 - i2;
        bufp = bufp + 8;
      }
      while (bufp < endp);
    }
    else
    {
      // inverse FFT: optimized code for small values of 'm' (2 and 4)
      double  a = 1.0 / double(invFFTLen);
      do
      {
        double  r1 = bufp[0];
        double  i1 = bufp[1];
        double  r2 = bufp[2];
        double  i2 = bufp[3];
        // scale inverse FFT by 1 / N
        bufp[0] = (r1 + r2) * a;
        bufp[1] = (i1 + i2) * a;
        bufp[2] = (r1 - r2) * a;
        bufp[3] = (i1 - i2) * a;
        bufp = bufp + 4;
      }
      while (bufp < endp);
      if (n < 4)
        return;
      bufp = buf;
      do
      {
        double  r1 = bufp[0];
        double  i1 = bufp[1];
        double  r2 = bufp[4];
        double  i2 = bufp[5];
        bufp[0] = r1 + r2;
        bufp[1] = i1 + i2;
        bufp[4] = r1 - r2;
        bufp[5] = i1 - i2;
        r1 = bufp[2];
        i1 = bufp[3];
        r2 = -(bufp[7]);
        i2 = bufp[6];
        bufp[2] = r1 + r2;
        bufp[3] = i1 + i2;
        bufp[6] = r1 - r2;
        bufp[7] = i1 - i2;
        bufp = bufp + 8;
      }
      while (bufp < endp);
    }
    if (n < 8)
      return;
    for (size_t m = 8; m < n; m <<= 1)
    {
      FFTSineGen  s(invFFTLen ? -(long(m)) : long(m));
      size_t  i = 0;
      do
      {
        double  re, im;
        s.sample(re, im);
        bufp = buf + i;
        do
        {
          double  r1 = bufp[0];
          double  i1 = bufp[1];
          double  r2 = bufp[m];
          double  i2 = bufp[m + 1];
          double  tmp = (r2 * re) - (i2 * im);
          i2 = (r2 * im) + (i2 * re);
          r2 = tmp;
          bufp[0] = r1 + r2;
          bufp[1] = i1 + i2;
          bufp[m] = r1 - r2;
          bufp[m + 1] = i1 - i2;
          bufp = bufp + (m << 1);
        }
        while (bufp < endp);
        i = i + 2;
      }
      while (i < m);
    }
  }
  FFTSineGen  s(invFFTLen ? -(long(n)) : long(n));
  double  *bufp = buf;
  double  *endp = buf + n;
  do
  {
    double  r1 = bufp[0];
    double  i1 = bufp[1];
    double  r2 = bufp[n];
    double  i2 = bufp[n + 1];
    double  re, im;
    s.sample(re, im);
    double  tmp = (r2 * re) - (i2 * im);
    i2 = (r2 * im) + (i2 * re);
    r2 = tmp;
    bufp[0] = r1 + r2;
    bufp[1] = i1 + i2;
    bufp[n] = r1 - r2;
    bufp[n + 1] = i1 - i2;
    bufp = bufp + 2;
  }
  while (bufp < endp);
}

// 'buf' should contain n * 2 floats (interleaved real/imaginary data)

static void complexFFT(double *buf, size_t n, bool isInverse)
{
  if (!buf)
    throw std::runtime_error(std::string("internal error: NULL FFT buffer"));
  if (n == 1)
    return;
  if (n < 1 || n > 16777216 || (n & (n - 1)) != 0)
    throw std::runtime_error(std::string("internal error: invalid FFT size"));
  // pack data in reverse bit order
  {
    size_t  j = n;
    for (size_t i = 2; j > 1; i += 2)
    {
      if (j > i)
      {
        double  tmp = buf[i];
        buf[i] = buf[j];
        buf[j] = tmp;
        tmp = buf[i + 1];
        buf[i + 1] = buf[j + 1];
        buf[j + 1] = tmp;
      }
      size_t  k = n;
      while (j & k)
      {
        j -= k;
        k >>= 1;
      }
      j += k;
    }
  }
  // calculate FFT
  recursiveFFT(buf, n, (isInverse ? int(n) : 0));
}

static void realFFT(double *buf, size_t n, bool isInverse)
{
  if (!buf)
    throw std::runtime_error(std::string("internal error: NULL FFT buffer"));
  if (n < 2 || n > 33554432 || (n & (n - 1)) != 0)
    throw std::runtime_error(std::string("internal error: invalid FFT size"));
  size_t  nd2 = n >> 1;
  if (isInverse)
  {
    FFTSineGen  s(-(long(n)));
    // DC: re = 1, im = 0
    // Fs / 2 (stored at buf[1]): re = -1, im = 0
    {
      double  re, im;
      s.sample(re, im);
      // even samples (0, 2, ...)
      double  r1 = buf[0];
      // odd samples (1, 3, ...)
      double  r2 = buf[1];
      buf[0] = (r1 + r2) * 0.5;
      buf[1] = (r1 - r2) * 0.5;
    }
    for (size_t i = 2; i < nd2; i += 2)
    {
      size_t  j = n - i;
      // lower half of spectrum
      double  rl = buf[i];
      double  il = buf[i + 1];
      // upper half of spectrum
      double  rh = buf[j];
      double  ih = buf[j + 1];
      // r1,i1 = even samples (0, 2, ...)
      double  r1 = rl + rh;
      double  i1 = il - ih;
      // delay odd samples by -1 sample
      double  re, im;
      s.sample(re, im);
      // r2,i2 = odd samples (1, 3, ...)
      double  r2 = ((rl * re) - (il * im)) + ((rh * -re) - (ih * im));
      double  i2 = ((rl * im) + (il * re)) - ((rh * im) + (ih * -re));
      buf[i] = (r1 - i2) * 0.5;
      buf[i + 1] = (i1 + r2) * 0.5;
      buf[j] = (r1 + i2) * 0.5;
      buf[j + 1] = (-i1 + r2) * 0.5;
    }
    if (n >= 4)
    {
      // Fs / 4 (i = j = nd2): re = 0, im = 1
      buf[nd2 + 1] *= -1.0;
    }
  }
  complexFFT(buf, n >> 1, isInverse);
  if (!isInverse)
  {
    FFTSineGen  s((long) n);
    // DC: re = 1, im = 0
    // Fs / 2 (stored at buf[1]): re = -1, im = 0
    {
      double  re, im;
      s.sample(re, im);
      // even samples (0, 2, ...)
      double  r1 = buf[0];
      // odd samples (1, 3, ...)
      double  r2 = buf[1];
      buf[0] = r1 + r2;
      buf[1] = r1 - r2;
    }
    for (size_t i = 2; i < nd2; i += 2)
    {
      size_t  j = n - i;
      // r1,i1 = even samples (0, 2, ...)
      double  r1 = (buf[i] + buf[j]) * 0.5;
      double  i1 = (buf[i + 1] - buf[j + 1]) * 0.5;
      // r2,i2 = odd samples (1, 3, ...)
      double  r2 = (buf[i + 1] + buf[j + 1]) * 0.5;
      double  i2 = (buf[j] - buf[i]) * 0.5;
      // delay odd samples by 1 sample
      double  re, im;
      s.sample(re, im);
      // lower half of spectrum
      buf[i] = r1 + (r2 * re) - (i2 * im);
      buf[i + 1] = i1 + (r2 * im) + (i2 * re);
      // upper half of spectrum: similar to above, but with -i1, -i2, and -re
      buf[j] = r1 - (r2 * re) + (i2 * im);
      buf[j + 1] = -i1 + (r2 * im) + (i2 * re);
    }
    if (n >= 4)
    {
      // Fs / 4 (i = j = nd2): re = 0, im = -1
      buf[nd2 + 1] *= -1.0;
    }
  }
}

// ============================================================================
// |                              AUDIO FILTERS                               |
// ============================================================================

AudioFilter::AudioFilter()
  : prv((AudioFilter *) 0),
    sampleRate(44100.0),
    channels(1),
    outBuf(1, 0.0)
{
}

AudioFilter::AudioFilter(AudioFilter& prv_)
  : prv(&prv_)
{
  sampleRate = prv->getSampleRate();
  channels = prv->getChannelCount();
  outBuf.resize(size_t(channels), 0.0);
}

AudioFilter::~AudioFilter()
{
}

// ----------------------------------------------------------------------------

void AudioFilter_FileInput::fillBuffer()
{
  sf_count_t  framesLeft = size_t(int(fileBuf.size()) / channels);
  sf_count_t  fileReadPos = filePos;
  double      *bufwp = &(fileBuf.front());
  while (framesLeft > 0)
  {
    if (fileReadPos < 0 || fileReadPos >= fileLength)
    {
      if (!isLooped)
      {
        for (int i = 0; i < channels; i++)
        {
          bufwp[i] = 0.0;
        }
        bufwp = bufwp + channels;
        framesLeft--;
        fileReadPos++;
        continue;
      }
      if (fileReadPos < 0)
        fileReadPos += fileLength;
      else
        fileReadPos -= fileLength;
      if (fileReadPos < 0 || fileReadPos >= fileLength)
      {
        fileReadPos = fileReadPos % fileLength;
        if (fileReadPos < 0)
          fileReadPos += fileLength;
      }
    }
    sf_count_t  framesToRead = fileLength - fileReadPos;
    if (framesToRead > framesLeft)
      framesToRead = framesLeft;
    if (sf_seek(f, fileReadPos, SEEK_SET) != fileReadPos)
      throw std::runtime_error(std::string("error seeking input file"));
    sf_count_t  framesRead = sf_readf_double(f, bufwp, framesToRead);
    if (framesRead != framesToRead)
      throw std::runtime_error(std::string("error reading input file"));
    framesLeft -= framesRead;
    fileReadPos += framesRead;
    bufwp = bufwp + (int(framesRead) * channels);
  }
  bufPos = 0;
}

AudioFilter_FileInput::AudioFilter_FileInput(const char *fname)
  : AudioFilter()
{
  if (!fname || !fname[0])
    throw std::runtime_error(std::string("invalid sound file name"));
  SF_INFO sfinfo;
  std::memset(&sfinfo, 0, sizeof(SF_INFO));
  f = sf_open(fname, SFM_READ, &sfinfo);
  if (!f)
  {
    throw std::runtime_error(std::string("error opening input file \"") + fname
                             + "\": " + sf_strerror(f));
  }
  try
  {
    if (sfinfo.samplerate < 1000 || sfinfo.samplerate > 10000000)
    {
      throw std::runtime_error(std::string("invalid sample rate "
                                           "in input file ") + fname);
    }
    if (sfinfo.channels < 1 || sfinfo.channels > 16)
    {
      throw std::runtime_error(std::string("invalid number of channels "
                                           "in input file ") + fname);
    }
    if (sfinfo.frames < 1)
    {
      throw std::runtime_error(std::string("empty input file: ") + fname);
    }
    sampleRate = double(sfinfo.samplerate);
    channels = sfinfo.channels;
    fileLength = sfinfo.frames;
    fileBuf.resize(size_t(channels) * 16384);
    filePos = 0;
    bufPos = fileBuf.size();
    isLooped = false;
    outBuf.resize(size_t(channels), 0.0);
  }
  catch (...)
  {
    sf_close(f);
    throw;
  }
}

AudioFilter_FileInput::~AudioFilter_FileInput()
{
  sf_close(f);
}

bool AudioFilter_FileInput::process()
{
  if (bufPos >= fileBuf.size())
    fillBuffer();
  copyOutputBuffer(&(fileBuf.front()) + bufPos);
  filePos++;
  bufPos = bufPos + size_t(channels);
  return true;
}

void AudioFilter_FileInput::setPosition(sf_count_t pos, bool looped)
{
  filePos = pos;
  bufPos = fileBuf.size();
  isLooped = looped;
}

// ----------------------------------------------------------------------------

AudioFilter_Delay::AudioFilter_Delay(AudioFilter& prv_,
                                     long delayFrames_, size_t lengthLimit)
  : AudioFilter(prv_),
    delayFrames(delayFrames_),
    framesRemaining(lengthLimit)
{
}

AudioFilter_Delay::~AudioFilter_Delay()
{
}

bool AudioFilter_Delay::process()
{
  if (framesRemaining)
  {
    framesRemaining--;
    if (delayFrames == 0L)
    {
      if (prv->process())
      {
        copyOutputBuffer(prv->getOutputBuffer());
        return true;
      }
    }
    else if (delayFrames > 0L)
    {
      delayFrames--;
      clearOutputBuffer();
      return true;
    }
    else
    {
      long    i = delayFrames - 1L;
      delayFrames = 0L;
      while (prv->process())
      {
        if (++i == 0L)
        {
          copyOutputBuffer(prv->getOutputBuffer());
          return true;
        }
      }
    }
  }
  framesRemaining = 0;
  clearOutputBuffer();
  return false;
}

// ----------------------------------------------------------------------------

AudioFilter_Upsample::AudioFilter_Upsample(AudioFilter& prv_,
                                           int upsampleRatio)
  : AudioFilter(prv_),
    r(upsampleRatio),
    cnt(0),
    isEOF(false)
{
  if (r < 1 || r > 1000)
    throw std::runtime_error(std::string("invalid upsample ratio"));
  sampleRate = prv->getSampleRate() * double(r);
}

AudioFilter_Upsample::~AudioFilter_Upsample()
{
}

bool AudioFilter_Upsample::process()
{
  if (cnt > 0)
  {
    cnt--;
    clearOutputBuffer();
    return true;
  }
  if (isEOF)
  {
    clearOutputBuffer();
    return false;
  }
  if (!prv->process())
  {
    isEOF = true;
    clearOutputBuffer();
    return false;
  }
  cnt = r - 1;
  copyOutputBuffer(prv->getOutputBuffer());
  return true;
}

// ----------------------------------------------------------------------------

AudioFilter_Downsample::AudioFilter_Downsample(AudioFilter& prv_,
                                               int downsampleRatio)
  : AudioFilter(prv_),
    r(downsampleRatio),
    isEOF(false)
{
  if (r < 1 || r > 1000)
    throw std::runtime_error(std::string("invalid downsample ratio"));
  sampleRate = prv->getSampleRate() / double(r);
}

AudioFilter_Downsample::~AudioFilter_Downsample()
{
}

bool AudioFilter_Downsample::process()
{
  if (!isEOF)
  {
    if (prv->process())
    {
      copyOutputBuffer(prv->getOutputBuffer());
      for (int i = 1; i < r; i++)
      {
        if (!prv->process())
        {
          isEOF = true;
          break;
        }
      }
      return true;
    }
    isEOF = true;
  }
  clearOutputBuffer();
  return false;
}

// ----------------------------------------------------------------------------

static void minimumPhaseIR(std::vector< double >& ir)
{
  // pad IR length to the nearest power of two, and at least 2
  size_t  len = 2;
  while (len < ir.size())
    len = len << 1;
  size_t  lenM2 = len << 1;
  std::vector< double > phaseBuf(lenM2, 0.0);
  ir.resize(lenM2, 0.0);
  // calculate magnitude response
  realFFT(&(ir.front()), lenM2, false);
  ir[0] = std::fabs(ir[0]);
  ir[1] = std::fabs(ir[1]);
  ir[0] = (ir[0] >= 1.0e-50 ? ir[0] : 1.0e-50);         // limit magnitude
  ir[1] = (ir[1] >= 1.0e-50 ? ir[1] : 1.0e-50);         // to at least -1000 dB
  phaseBuf[0] = std::log(ir[0]);
  phaseBuf[lenM2 >> 1] = std::log(ir[1]);
  for (size_t i = 2; i < lenM2; i += 2)
  {
    double  re = ir[i];
    double  im = ir[i + 1];
    double  r2 = (re * re) + (im * im);
    r2 = (r2 >= 1.0e-100 ? r2 : 1.0e-100);
    double  r = std::sqrt(r2);
    ir[i] = r;
    double  ln_r = std::log(r);
    phaseBuf[i >> 1] = ln_r;
    phaseBuf[lenM2 - (i >> 1)] = ln_r;
  }
  // calculate the Hilbert transform of the logarithm of the magnitude response
  realFFT(&(phaseBuf.front()), lenM2, false);
  phaseBuf[0] = 0.0;
  phaseBuf[1] = 0.0;
  for (size_t i = 2; i < lenM2; i += 2)
  {
    // since the magnitude response is mirrored around Fs/2 and symmetrical,
    // the imaginary part is always zero here
    phaseBuf[i + 1] = phaseBuf[i];
    phaseBuf[i] = 0.0;
  }
  realFFT(&(phaseBuf.front()), lenM2, true);
  // convert magnitude and phase response to real/imaginary format
  for (size_t i = 2; i < lenM2; i += 2)
  {
    double  r = ir[i];
    double  ph = phaseBuf[i >> 1];
    ir[i] = r * std::cos(ph);
    ir[i + 1] = r * std::sin(ph);
  }
  // convert frequency response to impulse response
  realFFT(&(ir.front()), lenM2, true);
  for (size_t i = 0; i < len; i++)
    ir[i] += ir[len + i];
  ir.resize(len);
}

// f = cutoff frequency as a fraction (0 to 0.5) of the sample rate
//     a negative value will create a minimum phase response
// a = gain (1.0 = 0 dB)

static void createLowpassIR(std::vector< double >& ir, size_t irLen,
                            double f, double a)
{
  if (irLen < 4 || irLen > 4194304 || (irLen & 1) != 0)
  {
    throw std::runtime_error(std::string("invalid lowpass filter IR length"));
  }
  bool    minPhase = false;
  if (f < 0.0)
  {
    minPhase = true;
    f = -f;
  }
  if (!(f >= 0.000001 && f <= 0.5))
  {
    throw std::runtime_error(std::string("invalid lowpass filter frequency"));
  }
  ir.resize(irLen, 0.0);
  size_t  irLenD2 = irLen >> 1;
  ir[0] = 0.0;
  ir[irLenD2] = f * 2.0 * a;
  a /= pi;
  CosineGen w(double(int(irLen)), pi);
  SineGen   s(1.0 / f);
  (void) w.sample();
  (void) s.sample();
  for (size_t i = 1; i < irLenD2; i++)
  {
    double  y = cosineToNuttallWindow(w.sample()) * a * s.sample()
                / double(int(i));
    ir[irLenD2 - i] = y;
    ir[irLenD2 + i] = y;
  }
  if (minPhase)
    minimumPhaseIR(ir);
}

// called by the constructors to initialize the filter

void AudioFilter_FIRFilter::initFilter(
    const std::vector< std::vector< double > >& impulseResponses)
{
  if (impulseResponses.size() != 1)
  {
    if (channels == 1)
    {
      channels = int(impulseResponses.size());
    }
    else if (impulseResponses.size() != size_t(channels) &&
             impulseResponses.size() != size_t(channels * channels))
    {
      throw std::runtime_error(std::string("invalid impulse response "
                                           "channel count"));
    }
  }
  irLength = impulseResponses[0].size();
  if (irLength < 2 || irLength > 4194304)
  {
    throw std::runtime_error(std::string("invalid impulse response length"));
  }
  for (size_t i = 1; i < impulseResponses.size(); i++)
  {
    if (impulseResponses[i].size() != irLength)
    {
      throw std::runtime_error(std::string("impulse response lengths do not "
                                           "match in multi-channel IR"));
    }
  }
  blockLength = 1;
  while (blockLength < irLength)
    blockLength = blockLength << 1;
  fftLength = blockLength << 1;
  fftBuf.resize(fftLength, 0.0);
  irFFTs.resize(impulseResponses.size());
  for (size_t i = 0; i < impulseResponses.size(); i++)
  {
    irFFTs[i].resize(fftLength, 0.0);
    for (size_t j = 0; j < irLength; j++)
    {
      irFFTs[i][j] = impulseResponses[i][j];
    }
    realFFT(&(irFFTs[i].front()), fftLength, false);
  }
  inBufs.resize(size_t(prv->getChannelCount()));
  for (int i = 0; i < prv->getChannelCount(); i++)
    inBufs[i].resize(blockLength, 0.0);
  outBufs.resize(size_t(channels));
  for (int i = 0; i < channels; i++)
    outBufs[i].resize(fftLength, 0.0);
  outBufPos = 0;
  framesRemaining = 0;
  isEOF = false;
}

AudioFilter_FIRFilter::AudioFilter_FIRFilter(
    AudioFilter& prv_,
    const std::vector< std::vector< double > >& impulseResponses)
  : AudioFilter(prv_)
{
  initFilter(impulseResponses);
}

AudioFilter_FIRFilter::AudioFilter_FIRFilter(AudioFilter& prv_,
                                             double f, int irLen, double a,
                                             double *d)
  : AudioFilter(prv_)
{
  std::vector< std::vector< double > >  impulseResponses;
  impulseResponses.resize(1);
  if (irLen < 0)
    irLen = int(double(irLen) * sampleRate * -0.000001 + 0.5);
  irLen = irLen + (irLen & 1);
  createLowpassIR(impulseResponses[0], size_t(irLen), f, a);
  initFilter(impulseResponses);
  if (d)
  {
    // calculate delay (NOTE: FFT length is always at least 4)
    if (f < 0.0)
    {
      // minimum phase
      *d = -(std::atan2(irFFTs[0][3], irFFTs[0][2]))
           * double(long(irFFTs[0].size())) / twoPi;
    }
    else
    {
      // linear phase
      *d = double(irLen >> 1);
    }
  }
}

AudioFilter_FIRFilter::~AudioFilter_FIRFilter()
{
}

bool AudioFilter_FIRFilter::process()
{
  if (framesRemaining > 0)
  {
    framesRemaining--;
    int     i = 0;
    do
    {
      outBuf[i] = outBufs[i][outBufPos];
      outBufs[i][outBufPos] = 0.0;
    }
    while (++i < channels);
    outBufPos = (outBufPos + 1) & (fftLength - 1);
    return true;
  }
  if (isEOF)
  {
    clearOutputBuffer();
    return false;
  }
  const double  *bufrp = &(prv->getOutputBuffer().front());
  int     inChannels = prv->getChannelCount();
  size_t  framesReceived = 0;
  for ( ; framesReceived < blockLength; framesReceived++)
  {
    if (!prv->process())
    {
      isEOF = true;
      break;
    }
    int     i = 0;
    do
    {
      inBufs[i][framesReceived] = bufrp[i];
    }
    while (++i < inChannels);
  }
  if (framesReceived > 0)
  {
    size_t  fftCnt = (inBufs.size() > irFFTs.size() ?
                      inBufs.size() : irFFTs.size());
    for (size_t i = 0; i < fftCnt; i++)
    {
      size_t  inChn = i;
      size_t  outChn = i;
      size_t  irChn = i;
      if (irFFTs.size() == 1)
      {
        // convolve all channels with the same (mono) impulse response
        irChn = 0;
      }
      else if (irFFTs.size() > size_t(inChannels))
      {
        // there are separate impulse responses for every possible
        // combination of input and output channel (IR channels =
        // audio channels ^ 2); for example, with stereo audio, the
        // IR channels are L->L, L->R, R->L, and R->R
        // FIXME: the implementation of this case is not optimal,
        // because the FFT of the same input channel is unnecessarily
        // calculated multiple times
        inChn = i / size_t(channels);
        outChn = i % size_t(channels);
      }
      // copy the input data to the FFT buffer
      for (size_t j = 0; j < framesReceived; j++)
        fftBuf[j] = inBufs[inChn][j];
      // pad to FFT length with zeros
      for (size_t j = framesReceived; j < fftLength; j++)
        fftBuf[j] = 0.0;
      // compute the FFT
      realFFT(&(fftBuf.front()), fftLength, false);
      // multiply the result with the FFT of the IR
      // to perform the convolution
      // there is no imaginary part for DC and the Nyquist frequency
      fftBuf[0] = fftBuf[0] * irFFTs[irChn][0];
      fftBuf[1] = fftBuf[1] * irFFTs[irChn][1];
      for (size_t j = 2; j < fftLength; j += 2)
      {
        double  re = fftBuf[j];
        double  im = fftBuf[j + 1];
        double  ir_re = irFFTs[irChn][j];
        double  ir_im = irFFTs[irChn][j + 1];
        fftBuf[j] = (re * ir_re) - (im * ir_im);
        fftBuf[j + 1] = (re * ir_im) + (im * ir_re);
      }
      // compute inverse FFT to get the convolved audio data
      realFFT(&(fftBuf.front()), fftLength, true);
      // add the result to the mix buffer
      for (size_t j = 0; j < fftLength; j++)
      {
        double& y = outBufs[outChn][(outBufPos + j) & (fftLength - 1)];
        y = y + fftBuf[j];
      }
    }
  }
  framesRemaining = framesReceived + (isEOF ? (irLength - 1) : 0) - 1;
  for (int i = 0; i < channels; i++)
  {
    outBuf[i] = outBufs[i][outBufPos];
    outBufs[i][outBufPos] = 0.0;
  }
  outBufPos = (outBufPos + 1) & (fftLength - 1);
  return true;
}

// ----------------------------------------------------------------------------

AudioFilter_Interpolate::AudioFilter_Interpolate(
    AudioFilter& prv_,
    double downsampleRatio, double startPos, int winSize)
  : AudioFilter(prv_)
{
  if (!(downsampleRatio >= 0.01 && downsampleRatio <= 100.0))
    throw std::runtime_error(std::string("invalid resample ratio"));
  if (!(startPos >= -1000000000.0 && startPos <= 1000000000.0))
    throw std::runtime_error(std::string("invalid resample start position"));
  if (winSize < 8 || winSize > 4096 || (winSize & 3) != 0)
    throw std::runtime_error(std::string("invalid resample window size"));
  framesRemaining = 0xFFFFFFFFUL;
  frameCntIn = 0;
  frameCntOut = 0;
  isEOF = false;
  windowSize = winSize;
  inBuf.resize(size_t(windowSize * channels), 0.0);
  writePos = 0;
  readOffs = (long double) (windowSize >> 1) + (long double) startPos;
  r = downsampleRatio;
  window_c = std::cos(twoPiL / (long double) windowSize);
  window_s = std::sin(twoPiL / (long double) windowSize);
  skipFrames = startPos;
  sampleRate = prv->getSampleRate() / r;
}

AudioFilter_Interpolate::~AudioFilter_Interpolate()
{
}

bool AudioFilter_Interpolate::process()
{
  clearOutputBuffer();
  if (!framesRemaining)
    return false;
  while (readOffs >= 0.0L)
  {
    if (!isEOF)
    {
      if (!prv->process())
      {
        isEOF = true;
        long    tmp = long((((double(frameCntIn) - skipFrames) / r)
                            - double(frameCntOut)) + 0.5);
        if (!(tmp > 0L))
        {
          framesRemaining = 0;
          return false;
        }
        framesRemaining = size_t(tmp);
      }
    }
    double  *bufwp = &(inBuf.front()) + writePos;
    int     i = 0;
    if (isEOF)
    {
      do
      {
        bufwp[i] = 0.0;
      }
      while (++i < channels);
    }
    else
    {
      const double  *bufrp = &(prv->getOutputBuffer().front());
      do
      {
        bufwp[i] = bufrp[i];
      }
      while (++i < channels);
      frameCntIn++;
    }
    writePos = ((writePos + channels) < int(inBuf.size()) ?
                (writePos + channels) : 0);
    readOffs -= 1.0L;
  }
  framesRemaining--;
  frameCntOut++;
  double  winPos_f = -(double(readOffs));
  readOffs += (long double) r;
  int     readPos = writePos;
  if (winPos_f < 0.00000003 || winPos_f > 0.99999997)
  {
    readPos = readPos + (((windowSize >> 1) - int(winPos_f > 0.5)) * channels);
    readPos = (readPos < int(inBuf.size()) ?
               readPos : (readPos - int(inBuf.size())));
    copyOutputBuffer(&(inBuf.front()) + readPos);
    return true;
  }
  double        piDsinX = pi / std::sin(winPos_f * pi);
  CosineGen     w(window_c, window_s, winPos_f * twoPi / double(windowSize));
  long double   x = ((long double) winPos_f - (long double) (windowSize >> 1))
                    * piDsinX;
  const double  *bufrp = &(inBuf.front()) + readPos;
  double        *bufwp = &(outBuf.front());
  int     i = windowSize >> 1;
  switch (channels)
  {
  case 1:
    do
    {
      double  a = cosineToNuttallWindow(w.sample()) / double(x);
      x += piDsinX;
      *bufwp += (*bufrp * a);
      bufrp++;
      bufrp = (bufrp < &(*(inBuf.end())) ? bufrp : &(inBuf.front()));
      a = cosineToNuttallWindow(w.sample()) / double(x);
      x += piDsinX;
      *bufwp -= (*bufrp * a);
      bufrp++;
      bufrp = (bufrp < &(*(inBuf.end())) ? bufrp : &(inBuf.front()));
    }
    while (--i);
    break;
  case 2:
    do
    {
      double  a = cosineToNuttallWindow(w.sample()) / double(x);
      x += piDsinX;
      bufwp[0] += (bufrp[0] * a);
      bufwp[1] += (bufrp[1] * a);
      bufrp = bufrp + 2;
      bufrp = (bufrp < &(*(inBuf.end())) ? bufrp : &(inBuf.front()));
      a = cosineToNuttallWindow(w.sample()) / double(x);
      x += piDsinX;
      bufwp[0] -= (bufrp[0] * a);
      bufwp[1] -= (bufrp[1] * a);
      bufrp = bufrp + 2;
      bufrp = (bufrp < &(*(inBuf.end())) ? bufrp : &(inBuf.front()));
    }
    while (--i);
    break;
  default:
    do
    {
      double  a = cosineToNuttallWindow(w.sample()) / double(x);
      x += piDsinX;
      int     j = 0;
      do
      {
        bufwp[j] += (bufrp[j] * a);
      }
      while (++j < channels);
      bufrp = bufrp + channels;
      bufrp = (bufrp < &(*(inBuf.end())) ? bufrp : &(inBuf.front()));
      a = cosineToNuttallWindow(w.sample()) / double(x);
      x += piDsinX;
      j = 0;
      do
      {
        bufwp[j] -= (bufrp[j] * a);
      }
      while (++j < channels);
      bufrp = bufrp + channels;
      bufrp = (bufrp < &(*(inBuf.end())) ? bufrp : &(inBuf.front()));
    }
    while (--i);
    break;
  }
  return true;
}

// ----------------------------------------------------------------------------

AudioFilter_ResampledFileInput::AudioFilter_ResampledFileInput(
    const char *fname, double outputSampleRate, double isrAdj, double skipTime,
    bool isLooped, int interpLen, int lpFilterLen, double lpFilterFreq)
  : AudioFilter()
{
  if (!(isrAdj >= 0.1 && isrAdj <= 10.0))
  {
    throw std::runtime_error(std::string("invalid resample "
                                         "sample rate adjust factor"));
  }
  filters.resize(5, (AudioFilter *) 0);
  AudioFilter*& lastFilter = filters[filters.size() - 1];
  try
  {
    AudioFilter_FileInput&  sf = *(new AudioFilter_FileInput(fname));
    filters[0] = &sf;
    sampleRate = outputSampleRate;
    if (sampleRate < 0.0)
      sampleRate = -sampleRate * sf.getSampleRate();
    if (!(sampleRate >= 1000.0 && sampleRate <= 1000000.0))
    {
      throw std::runtime_error(std::string("invalid resample "
                                           "output sample rate"));
    }
    double  isr = sf.getSampleRate() * isrAdj;
    double  r = sampleRate / isr;
    if (!(r >= 0.01 && r <= 100.0))
      throw std::runtime_error(std::string("invalid resample ratio"));
    channels = sf.getChannelCount();
    outBuf.resize(size_t(channels), 0.0);
    double  skipFrames = sf.getSampleRate() * skipTime;
    double  skipFrames_i = 0.0;
    double  skipFrames_f = std::modf(skipFrames, &skipFrames_i);
    if (skipFrames_f < 0.0)
    {
      skipFrames_i -= 1.0;
      skipFrames_f += 1.0;
    }
    if (skipFrames_f < 0.0000002 || skipFrames_f > 0.9999998)
    {
      // integer sample delay
      skipFrames_i += double(int(skipFrames_f > 0.5));
      skipFrames = skipFrames_i;
      skipFrames_f = 0.0;
    }
    sf.setPosition(sf_count_t(skipFrames_i), isLooped);
    sf.setSampleRate(isr);
    int     r_i = 0;            // integer downsample or upsample ratio
    {
      double  tmp = (r < 1.0 ? (1.0 / r) : r);
      r_i = int(tmp + 0.5);
      if (std::fabs(tmp - double(r_i)) > 5.0e-13)
        r_i = 0;                // not integer resample ratio
      else if (r < 1.0)
        r = 1.0 / double(r_i);
      else
        r = double(r_i);
    }
    sf_count_t  outFrames = 0;
    if (isLooped)
    {
      // unlimited length in the case of looped playback
      outFrames--;
    }
    else
    {
      double  tmp = (double(sf.getFileLength()) - skipFrames) * r + 0.5;
      if (tmp >= 0.0)
        outFrames = sf_count_t(tmp);
    }
    if (r_i > 0 && skipFrames_f == 0.0)
    {
      // no interpolation is needed
      if (r_i == 1)
      {
        // no sample rate conversion
        lastFilter = new AudioFilter_Delay(sf, 0L, size_t(outFrames));
        return;
      }
      int     offs = 0;
      if (isLooped || skipFrames_i > 0.0)
      {
        sf.setPosition(sf_count_t(skipFrames_i) - 100, isLooped);
        offs = 100;
      }
      if (r < 1.0)
      {
        // downsampling by integer ratio
        double  gd = 0.0;
        filters[1] = new AudioFilter_FIRFilter(sf, lpFilterFreq * r,
                                               lpFilterLen, 1.0, &gd);
        filters[2] = new AudioFilter_Delay(*(filters[1]),
                                           long(-(gd + 0.5)) - offs,
                                           size_t(outFrames * r_i));
        lastFilter = new AudioFilter_Downsample(*(filters[2]), r_i);
      }
      else
      {
        // upsampling by integer ratio
        double  gd = 0.0;
        filters[1] = new AudioFilter_Upsample(sf, r_i);
        filters[2] = new AudioFilter_FIRFilter(*(filters[1]),
                                               lpFilterFreq / r, lpFilterLen,
                                               r, &gd);
        lastFilter = new AudioFilter_Delay(*(filters[2]),
                                           long(-(gd + 0.5)) - offs,
                                           size_t(outFrames));
      }
      return;
    }
    // fractional resample ratio and/or delay
    if (isLooped || skipFrames_i > 0.0)
    {
      sf.setPosition(sf_count_t(skipFrames_i) - 100, isLooped);
      skipFrames_f += 100.0;
    }
    if (r > 0.5)
    {
      // need to upsample first
      int     r_us = int(std::ceil(r / 0.5));
      double  r_ds = double(r_us) * isr / sampleRate;
      double  gd = 0.0;
      filters[1] = new AudioFilter_Upsample(sf, r_us);
      filters[2] = new AudioFilter_FIRFilter(*(filters[1]),
                                             lpFilterFreq / double(r_us),
                                             lpFilterLen, double(r_us), &gd);
      filters[3] = new AudioFilter_Interpolate(*(filters[2]), r_ds,
                                               skipFrames_f * double(r_us) + gd,
                                               interpLen);
      lastFilter = new AudioFilter_Delay(*(filters[3]), 0L, size_t(outFrames));
    }
    else
    {
      // downsample only
      double  gd = 0.0;
      filters[1] = new AudioFilter_FIRFilter(sf, lpFilterFreq * r, lpFilterLen,
                                             1.0, &gd);
      filters[2] = new AudioFilter_Interpolate(*(filters[1]), isr / sampleRate,
                                               skipFrames_f + gd, interpLen);
      lastFilter = new AudioFilter_Delay(*(filters[2]), 0L, size_t(outFrames));
    }
  }
  catch (...)
  {
    for (size_t i = 0; i < filters.size(); i++)
    {
      if (filters[i])
      {
        delete filters[i];
        filters[i] = (AudioFilter *) 0;
      }
    }
    throw;
  }
}

AudioFilter_ResampledFileInput::~AudioFilter_ResampledFileInput()
{
  for (size_t i = 0; i < filters.size(); i++)
  {
    if (filters[i])
    {
      delete filters[i];
      filters[i] = (AudioFilter *) 0;
    }
  }
}

bool AudioFilter_ResampledFileInput::process()
{
  AudioFilter&  lastFilter = *(filters[filters.size() - 1]);
  if (!lastFilter.process())
  {
    clearOutputBuffer();
    return false;
  }
  copyOutputBuffer(lastFilter.getOutputBuffer());
  return true;
}

// ============================================================================

RandomGenerator::RandomGenerator(int seed_)
{
  if (!seed_)
    seed_ = int((unsigned int) std::time((std::time_t *) 0) & 0x7FFFFFFFU);
  while (seed_ < 1)
    seed_ += 0x7FFFFFFE;
  while (seed_ >= 0x7FFFFFFF)
    seed_ -= 0x7FFFFFFE;
  seed = seed_;
  (void) random();
  (void) random();
}

double RandomGenerator::random()
{
  long long     tmp = seed * (long long) rndMult;
  unsigned int  tmp2 = (unsigned int) (tmp & 0x7FFFFFFFLL)
                       + (unsigned int) (tmp >> 31);
  seed = int(tmp2 < 0x80000000U ? tmp2 : (tmp2 - 0x7FFFFFFFU));
  return (double(seed) * (1.0 / double(0x7FFFFFFF)));
}

// ----------------------------------------------------------------------------

void AudioOutputFile::flushBuffer()
{
  if (bufPos < 1)
    return;
  if (ditherPrvState.size() > 0)
  {
    int     chn = 0;
    for (size_t i = 0; i < bufPos; i++)
    {
      double  d = ditherRNG.random();
      buf[i] = buf[i] + ((d - ditherPrvState[chn]) * ditherLevel);
      ditherPrvState[chn] = d;
      chn = ((chn + 1) < channels ? (chn + 1) : 0);
    }
  }
  if (sf_write_double(f, &(buf.front()), sf_count_t(bufPos))
      != sf_count_t(bufPos))
  {
    throw std::runtime_error(std::string("error writing audio output file"));
  }
  bufPos = 0;
}

AudioOutputFile::AudioOutputFile(const char *fname,
                                 double sampleRate, int channels_,
                                 int sndfile_format, int ditherSeed)
  : ditherRNG(ditherSeed)
{
  if (!(sampleRate >= 1000.0 && sampleRate <= 10000000.0))
  {
    throw std::runtime_error(std::string("invalud audio output file "
                                         "sample rate"));
  }
  channels = channels_;
  if (channels < 1 || channels > 16)
  {
    throw std::runtime_error(std::string("invalud audio output file "
                                         "channel count"));
  }
  buf.resize(size_t(channels) * 4096, 0.0);
  bufPos = 0;
  ditherLevel = 0.0;
  SF_INFO sfinfo;
  std::memset(&sfinfo, 0, sizeof(SF_INFO));
  sfinfo.frames = sf_count_t(-1);
  sfinfo.samplerate = int(sampleRate + 0.5);
  sfinfo.channels = channels;
  sfinfo.format = sndfile_format;
  f = sf_open(fname, SFM_WRITE, &sfinfo);
  if (!f)
  {
    throw std::runtime_error(std::string("error opening audio output file: ")
                             + sf_strerror(f));
  }
  switch (sndfile_format & SF_FORMAT_SUBMASK)
  {
  case SF_FORMAT_PCM_S8:
  case SF_FORMAT_PCM_U8:
  case SF_FORMAT_PCM_16:
  case SF_FORMAT_PCM_24:
    if (ditherSeed >= 0)
    {
      ditherPrvState.resize(size_t(channels));
      for (int i = 0; i < channels; i++)
        ditherPrvState[i] = ditherRNG.random();
      switch (sndfile_format & SF_FORMAT_SUBMASK)
      {
      case SF_FORMAT_PCM_S8:
      case SF_FORMAT_PCM_U8:
        ditherLevel = 1.0 / 128.0;
        break;
      case SF_FORMAT_PCM_16:
        ditherLevel = 1.0 / 32768.0;
        break;
      case SF_FORMAT_PCM_24:
        ditherLevel = 1.0 / 8388608.0;
        break;
      }
    }
  case SF_FORMAT_PCM_32:
    // clip integer formats
    sf_command(f, SFC_SET_CLIPPING, NULL, SF_TRUE);
    break;
  }
  sf_command(f, SFC_SET_UPDATE_HEADER_AUTO, NULL, 0);
}

AudioOutputFile::~AudioOutputFile()
{
  try
  {
    flushBuffer();
  }
  catch (...)
  {
  }
  sf_close(f);
}

void AudioOutputFile::writeSampleFrame(const double *srcBuf)
{
  double  *bufwp = &(buf.front()) + bufPos;
  int     i = 0;
  do
  {
    bufwp[i] = srcBuf[i];
  }
  while (++i < channels);
  bufPos = bufPos + size_t(i);
  if (bufPos >= buf.size())
    flushBuffer();
}

#endif  // DSPLIB_CPP_INCLUDED

