
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

static long lfsr(long r, long p, long seqLen)
{
  long    tmp = (unsigned long) (r & p);
  if (sizeof(unsigned long) >= 8)
    tmp = tmp ^ (tmp >> 32);
  tmp = tmp ^ (tmp >> 16);
  tmp = tmp ^ (tmp >> 8);
  tmp = tmp ^ (tmp >> 4);
  tmp = tmp ^ (tmp >> 2);
  tmp = (tmp ^ (tmp >> 1)) & 1;
  return (((r << 1) & seqLen) | (long) tmp);
}

static void resample_window(float* w, long n, long ws)
{
  const double  pi = atan(1.0) * 4.0;
  long    i;
  for (i = 0; i <= n; i++)
  {
    if (i != (n >> 1))
    {
      double  x = pi * ws * ((double) (i - (n >> 1)) / n);
      double  y = sin(pi * (double) i / n);
      y = (y * y) * sin(x) / x;
      w[i] = (float) y;
    }
    else
    {
      w[i] = 1.0f;
    }
  }
}

#define WINSIZE (1024)
#define WINRES  (512)

static int lfsr_resampled(FILE* f, long p, long seqLen)
{
  unsigned char*  buf;
  float*  w;
  long    i, r;
  buf = (unsigned char*) malloc(sizeof(unsigned char) * seqLen);
  if (!buf)
    return -2;
  w = (float*) malloc(sizeof(float) * (WINSIZE * WINRES + 1));
  if (!w)
  {
    free(buf);
    return -2;
  }
  resample_window(w, WINSIZE * WINRES, WINSIZE);
  r = seqLen;
  for (i = 0; i < seqLen; i++)
  {
    r = lfsr(r, p, seqLen);
    buf[i] = (unsigned char) (r & 1);
  }
  for (i = 0; i <= seqLen; i++)
  {
    double  x = (double) i * seqLen / (seqLen + 1);
    double  y = 0.0;
    double  wf = (double) WINRES * (1.0 - (x - (double) ((long) x)));
    long    wi = (long) wf;
    long    xi = (long) x - (WINSIZE >> 1);
    float   yf;
    while (xi < 0)
      xi += seqLen;
    wf = wf - (double) wi;
    while (wi < (WINSIZE * WINRES))
    {
      double  w0 = w[wi];
      double  w1 = w[wi + 1];
      y += ((double) buf[xi] * (w0 + ((w1 - w0) * wf)));
      xi++;
      xi = (xi < seqLen ? xi : 0L);
      wi = wi + WINRES;
    }
    yf = (float) ((y - 0.5) * sqrt(0.5));
    if (fwrite(&yf, sizeof(float), 1, f) != 1)
    {
      free(buf);
      free(w);
      return -1;
    }
  }
  free(buf);
  free(w);
  return 0;
}

int main(int argc, char** argv)
{
  FILE*   f = NULL;
  long    p = 0;
  long    r, n, seqLen;
  unsigned char*  buf = NULL;
  int     nBits = 0;
  char*   endp = NULL;
  int     resample = 0;
  if (argc != 3 && argc != 4)
  {
    fprintf(stderr, "Usage: %s FNAME POLY [RESAMPLE]\n", argv[0]);
    return -1;
  }
  if (argc > 3)
    resample = atoi(argv[3]);
  p = strtol(argv[2], &endp, 2);
  if (p < 1)
  {
    fprintf(stderr, "%s: invalid polynomial: %ld\n", argv[0], p);
    return -1;
  }
  while ((1L << nBits) <= p)
    nBits++;
  buf = (unsigned char*) calloc((((size_t) 1 << nBits) + 7) >> 3,
                                sizeof(unsigned char));
  if (!buf)
  {
    fprintf(stderr, "%s: error allocating memory\n", argv[0]);
    return -1;
  }
  seqLen = (1L << nBits) - 1;
  r = seqLen;
  for (n = 0; n < seqLen; n++)
  {
    long    i = r >> 3;
    unsigned char m = (unsigned char) (1 << (r & 7));
    if (buf[i] & m)
    {
      fprintf(stderr, "%s: short sequence (loop after %ld samples)\n",
              argv[0], n);
      free(buf);
      return -1;
    }
    buf[i] = buf[i] | m;
    r = lfsr(r, p, seqLen);
  }
  free(buf);
  buf = NULL;
  f = fopen(argv[1], "wb");
  if (!f)
  {
    fprintf(stderr, "%s: error opening output file '%s'\n", argv[0], argv[1]);
    return -1;
  }
  r = seqLen;
  if (resample)
  {
    switch (lfsr_resampled(f, p, seqLen))
    {
      case -1:
        fclose(f);
        remove(argv[1]);
        fprintf(stderr, "%s: error writing output file '%s'\n",
                argv[0], argv[1]);
        return -1;
      case -2:
        fclose(f);
        remove(argv[1]);
        fprintf(stderr, "%s: error allocating memory\n", argv[0]);
        return -1;
    }
  }
  else
  {
    for (n = 0; n < seqLen; n++)
    {
      int     tmp;
      r = lfsr(r, p, seqLen);
      tmp = ((int) (r & 1) << 15) - 0x4000;
      if (fputc(tmp & 0xFF, f) == EOF || fputc((tmp >> 8) & 0xFF, f) == EOF)
      {
        fclose(f);
        remove(argv[1]);
        fprintf(stderr, "%s: error writing output file '%s'\n",
                argv[0], argv[1]);
        return -1;
      }
    }
  }
  if (fflush(f) != 0)
  {
    fclose(f);
    remove(argv[1]);
    fprintf(stderr, "%s: error writing output file '%s'\n", argv[0], argv[1]);
    return -1;
  }
  fclose(f);
  return 0;
}

