
#ifndef OPTIONS_CPP_INCLUDED
#define OPTIONS_CPP_INCLUDED

#include "headers.hpp"

class CommandLineParser
{
 protected:
  enum CfgVarType
  {
    CFG_VAR_TYPE_BOOL = 1,
    CFG_VAR_TYPE_INT = 2,
    CFG_VAR_TYPE_LONG_LONG = 3,
    CFG_VAR_TYPE_FLOAT = 4,
    CFG_VAR_TYPE_DOUBLE = 5,
    CFG_VAR_TYPE_STRING = 6
  };
  struct ConfigVariable
  {
    const char  *longName;
    const char  *shortName;
    const char  *description;
    void        *value;
    int         type;
    double      minVal;
    double      maxVal;
  };
  std::vector< std::string * >  fileNames;
  std::vector< const char * >   fileNameNames;
  std::vector< const char * >   fileNameDescriptions;
  std::vector< ConfigVariable > cfgVariables;
  const char  *programName;
  // --------
  void parseOption(const std::string& arg, const std::string& valueStr);
  void readConfigurationFile(const std::string& fileName);
 public:
  CommandLineParser(const char *prgName);
  virtual ~CommandLineParser();
  void addOption(const char *longName, const char *shortName, const char *desc,
                 bool& value);
  void addOption(const char *longName, const char *shortName, const char *desc,
                 int& value, int minValue = INT_MIN, int maxValue = INT_MAX);
  void addOption(const char *longName, const char *shortName, const char *desc,
                 long long& value, long long minValue = LLONG_MIN,
                 long long maxValue = LLONG_MAX);
  void addOption(const char *longName, const char *shortName, const char *desc,
                 float& value,
                 float minValue = -1.0e30f, float maxValue = 1.0e30f);
  void addOption(const char *longName, const char *shortName, const char *desc,
                 double& value,
                 double minValue = -1.0e30, double maxValue = 1.0e30);
  void addOption(const char *longName, const char *shortName, const char *desc,
                 std::string& value);
  void addFileName(const char *name, const char *description,
                   std::string& value);
  // return value is false if --help is used
  bool parseCommandLine(int argc, const char * const *argv);
};

void CommandLineParser::parseOption(const std::string& arg,
                                    const std::string& valueStr)
{
  ConfigVariable  *cv = (ConfigVariable *) 0;
  for (size_t j = 0; j < cfgVariables.size(); j++)
  {
    if (arg == cfgVariables[j].shortName || arg == cfgVariables[j].longName)
    {
      cv = &(cfgVariables[j]);
      break;
    }
  }
  if (!cv)
    throw std::runtime_error(std::string("invalid option: -") + arg);
  if (valueStr.empty())
  {
    throw std::runtime_error(std::string("option -") + arg
                             + " value string is empty");
  }
  switch (cv->type)
  {
  case CFG_VAR_TYPE_BOOL:
    if (valueStr == "0" ||
        valueStr == "no" || valueStr == "No" || valueStr == "NO" ||
        valueStr == "false" || valueStr == "False" || valueStr == "FALSE")
    {
      *(reinterpret_cast< bool * >(cv->value)) = false;
    }
    else if (valueStr == "1" ||
             valueStr == "yes" || valueStr == "Yes" || valueStr == "YES" ||
             valueStr == "true" || valueStr == "True" || valueStr == "TRUE")
    {
      *(reinterpret_cast< bool * >(cv->value)) = true;
    }
    else
    {
      throw std::runtime_error(std::string("invalid value for -") + arg);
    }
    break;
  case CFG_VAR_TYPE_INT:
    {
      char    *endp = (char *) 0;
      long    n = std::strtol(valueStr.c_str(), &endp, 0);
      if (!endp || endp == valueStr.c_str() || *endp)
      {
        throw std::runtime_error(std::string("invalid integer format for -")
                                 + arg);
      }
      if (double(n) < cv->minVal || double(n) > cv->maxVal)
      {
        throw std::runtime_error(std::string("value is out of range for -")
                                 + arg);
      }
      *(reinterpret_cast< int * >(cv->value)) = int(n);
    }
    break;
  case CFG_VAR_TYPE_LONG_LONG:
    {
      char    *endp = (char *) 0;
      long long&  n = *(reinterpret_cast< long long * >(cv->value));
      n = std::strtoll(valueStr.c_str(), &endp, 0);
      if (!endp || endp == valueStr.c_str() || *endp)
      {
        throw std::runtime_error(std::string("invalid integer format for -")
                                 + arg);
      }
      if (double(n) < cv->minVal || double(n) > cv->maxVal)
      {
        throw std::runtime_error(std::string("value is out of range for -")
                                 + arg);
      }
    }
    break;
  case CFG_VAR_TYPE_FLOAT:
    {
      char    *endp = (char *) 0;
      double  n = std::strtod(valueStr.c_str(), &endp);
      if (!endp || endp == valueStr.c_str() || *endp)
      {
        throw std::runtime_error(std::string("invalid number format for -")
                                 + arg);
      }
      if (!(n >= cv->minVal && n <= cv->maxVal))
      {
        throw std::runtime_error(std::string("value is out of range for -")
                                 + arg);
      }
      *(reinterpret_cast< float * >(cv->value)) = float(n);
    }
    break;
  case CFG_VAR_TYPE_DOUBLE:
    {
      char    *endp = (char *) 0;
      double& n = *(reinterpret_cast< double * >(cv->value));
      n = std::strtod(valueStr.c_str(), &endp);
      if (!endp || endp == valueStr.c_str() || *endp)
      {
        throw std::runtime_error(std::string("invalid number format for -")
                                 + arg);
      }
      if (!(n >= cv->minVal && n <= cv->maxVal))
      {
        throw std::runtime_error(std::string("value is out of range for -")
                                 + arg);
      }
    }
    break;
  case CFG_VAR_TYPE_STRING:
    *(reinterpret_cast< std::string * >(cv->value)) = valueStr;
    break;
  }
}

void CommandLineParser::readConfigurationFile(const std::string& fileName)
{
  if (fileName.empty())
    throw std::runtime_error(std::string("invalid configuration file name"));
  std::FILE *f = std::fopen(fileName.c_str(), "rb");
  if (!f)
    throw std::runtime_error(std::string("error opening configuration file"));
  try
  {
    std::string optName;
    std::string optValue;
    bool        isEOF = false;
    bool        skippingComment = false;
    bool        quoteFlag = false;
    bool        wasSpace = true;
    int         argCnt = 0;
    do
    {
      int     c = std::fgetc(f);
      if (c == EOF)
      {
        isEOF = true;
        c = '\n';
      }
      c = c & 0xFF;
      if (c != '\r' && c != '\n')
      {
        if (c == '"')
        {
          quoteFlag = !quoteFlag;
          continue;
        }
        if (!quoteFlag)
        {
          if (c == '#' || c == ';')
            skippingComment = true;
          if (skippingComment)
            continue;
          if (c < ' ')
            c = '\t';
          if (c == ' ' || c == '\t')
          {
            wasSpace = true;
            continue;
          }
          if (wasSpace)
          {
            argCnt++;
            if (c == '=' && argCnt == 2)
              continue;
            argCnt += int(argCnt == 2);
            wasSpace = false;
            if (argCnt > 3)
            {
              throw std::runtime_error(std::string("invalid syntax "
                                                   "in configuration file"));
            }
          }
        }
        if (argCnt == 1)
          optName += char(c);
        else if (argCnt > 1)
          optValue += char(c);
      }
      else
      {
        if (quoteFlag)
        {
          throw std::runtime_error(std::string("unmatched quotes "
                                               "in configuration file"));
        }
        if (argCnt == 1)
        {
          throw std::runtime_error(std::string("missing option value for '")
                                   + optName + "' in configuration file");
        }
        if (argCnt > 1)
          parseOption(optName, optValue);
        optName.clear();
        optValue.clear();
        skippingComment = false;
        quoteFlag = false;
        wasSpace = true;
        argCnt = 0;
      }
    }
    while (!isEOF);
  }
  catch (...)
  {
    std::fclose(f);
    throw;
  }
  std::fclose(f);
}

CommandLineParser::CommandLineParser(const char *prgName)
  : programName(prgName)
{
}

CommandLineParser::~CommandLineParser()
{
}

void CommandLineParser::addOption(const char *longName, const char *shortName,
                                  const char *desc, bool& value)
{
  ConfigVariable  cv;
  cv.longName = longName;
  cv.shortName = shortName;
  cv.description = desc;
  cv.type = CFG_VAR_TYPE_BOOL;
  cv.value = &value;
  cv.minVal = 0.0;
  cv.maxVal = 1.0;
  cfgVariables.push_back(cv);
}

void CommandLineParser::addOption(const char *longName, const char *shortName,
                                  const char *desc, int& value,
                                  int minValue, int maxValue)
{
  ConfigVariable  cv;
  cv.longName = longName;
  cv.shortName = shortName;
  cv.description = desc;
  cv.type = CFG_VAR_TYPE_INT;
  cv.value = &value;
  cv.minVal = double(minValue);
  cv.maxVal = double(maxValue);
  cfgVariables.push_back(cv);
}

void CommandLineParser::addOption(const char *longName, const char *shortName,
                                  const char *desc, long long& value,
                                  long long minValue, long long maxValue)
{
  ConfigVariable  cv;
  cv.longName = longName;
  cv.shortName = shortName;
  cv.description = desc;
  cv.type = CFG_VAR_TYPE_LONG_LONG;
  cv.value = &value;
  cv.minVal = double(minValue);
  cv.maxVal = double(maxValue);
  cfgVariables.push_back(cv);
}

void CommandLineParser::addOption(const char *longName, const char *shortName,
                                  const char *desc, float& value,
                                  float minValue, float maxValue)
{
  ConfigVariable  cv;
  cv.longName = longName;
  cv.shortName = shortName;
  cv.description = desc;
  cv.type = CFG_VAR_TYPE_FLOAT;
  cv.value = &value;
  cv.minVal = double(minValue);
  cv.maxVal = double(maxValue);
  cfgVariables.push_back(cv);
}

void CommandLineParser::addOption(const char *longName, const char *shortName,
                                  const char *desc, double& value,
                                  double minValue, double maxValue)
{
  ConfigVariable  cv;
  cv.longName = longName;
  cv.shortName = shortName;
  cv.description = desc;
  cv.type = CFG_VAR_TYPE_DOUBLE;
  cv.value = &value;
  cv.minVal = minValue;
  cv.maxVal = maxValue;
  cfgVariables.push_back(cv);
}

void CommandLineParser::addOption(const char *longName, const char *shortName,
                                  const char *desc, std::string& value)
{
  ConfigVariable  cv;
  cv.longName = longName;
  cv.shortName = shortName;
  cv.description = desc;
  cv.type = CFG_VAR_TYPE_STRING;
  cv.value = &value;
  cv.minVal = 0.0;
  cv.maxVal = 0.0;
  cfgVariables.push_back(cv);
}

void CommandLineParser::addFileName(const char *name, const char *description,
                                    std::string& value)
{
  fileNames.push_back(&value);
  fileNameNames.push_back(name);
  fileNameDescriptions.push_back(description);
}

bool CommandLineParser::parseCommandLine(int argc, const char * const *argv)
{
  for (size_t i = 0; i < fileNames.size(); i++)
    fileNames[i]->clear();
  std::string arg;
  std::string valueStr;
  bool        fileNamesOnly = false;
  for (int i = 1; i < argc; i++)
  {
    if (!argv[i][0])
      throw std::runtime_error(std::string("empty command line argument"));
    if (fileNamesOnly || argv[i][0] != '-')
    {
      size_t  n = 0;
      while (n < fileNames.size() && !fileNames[n]->empty())
        n++;
      if (n >= fileNames.size())
        throw std::runtime_error(std::string("too many file names"));
      *(fileNames[n]) = argv[i];
      continue;
    }
    if (argv[i][1] == '-')
    {
      if (!argv[i][2])
      {
        fileNamesOnly = true;
        continue;
      }
      arg = &(argv[i][2]);
    }
    else if (argv[i][1])
    {
      arg = &(argv[i][1]);
    }
    else
    {
      throw std::runtime_error(std::string("invalid option: \"-\""));
    }
    if (arg == "?" || arg == "h" || arg == "help")
    {
      std::printf("Usage: %s [OPTIONS...]", programName);
      for (size_t j = 0; j < fileNameNames.size(); j++)
        std::printf(" %s", fileNameNames[j]);
      std::printf("\n");
      for (size_t j = 0; j < fileNameNames.size(); j++)
      {
        if (fileNameDescriptions[j] && fileNameDescriptions[j][0])
        {
          std::printf("    %s:\n        %s\n",
                      fileNameNames[j], fileNameDescriptions[j]);
        }
      }
      std::printf("Options (can be specified in any of the formats "
                  "-OPTION VALUE,\n");
      std::printf("--OPTION VALUE, -OPTION=VALUE, and --OPTION=VALUE):\n");
      std::printf("    -? | -h | -help\n");
      std::printf("        print usage information\n");
      std::printf("    --\n");
      std::printf("        interpret any remaining arguments as file names\n");
      std::printf("    -cfg | -config    (string)\n");
      std::printf("        read options from a configuration file\n");
      for (size_t j = 0; j < cfgVariables.size(); j++)
      {
        const ConfigVariable& cv = cfgVariables[j];
        std::printf("    -%s | -%s", cv.shortName, cv.longName);
        switch (cv.type)
        {
        case CFG_VAR_TYPE_BOOL:
          std::printf("    (boolean, value: %d)\n",
                      int(*(reinterpret_cast< bool * >(cv.value))));
          break;
        case CFG_VAR_TYPE_INT:
          std::printf("    (integer");
          if (cv.minVal > double(INT_MIN))
            std::printf(", min: %.0f", cv.minVal);
          if (cv.maxVal < double(INT_MAX))
            std::printf(", max: %.0f", cv.maxVal);
          std::printf(", value: %d)\n",
                      *(reinterpret_cast< int * >(cv.value)));
          break;
        case CFG_VAR_TYPE_LONG_LONG:
          std::printf("    (integer");
          if (cv.minVal > (double(LLONG_MIN) * 0.99999999))
            std::printf(", min: %.0f", cv.minVal);
          if (cv.maxVal < (double(LLONG_MAX) * 0.99999999))
            std::printf(", max: %.0f", cv.maxVal);
          std::printf(", value: %lld)\n",
                      *(reinterpret_cast< long long * >(cv.value)));
          break;
        case CFG_VAR_TYPE_FLOAT:
          std::printf("    (float");
          if (cv.minVal > -9.99999e29)
            std::printf(", min: %g", cv.minVal);
          if (cv.maxVal < 9.99999e29)
            std::printf(", max: %g", cv.maxVal);
          std::printf(", value: %g)\n",
                      double(*(reinterpret_cast< float * >(cv.value))));
          break;
        case CFG_VAR_TYPE_DOUBLE:
          std::printf("    (float");
          if (cv.minVal > -9.9999999e29)
            std::printf(", min: %.8g", cv.minVal);
          if (cv.maxVal < 9.9999999e29)
            std::printf(", max: %.8g", cv.maxVal);
          std::printf(", value: %.8g)\n",
                      *(reinterpret_cast< double * >(cv.value)));
          break;
        case CFG_VAR_TYPE_STRING:
          std::printf("    (string, value: \"%s\")\n",
                      reinterpret_cast< std::string * >(cv.value)->c_str());
          break;
        default:
          std::printf("\n");
          break;
        }
        if (cv.description && cv.description[0])
          std::printf("        %s\n", cv.description);
      }
      return false;
    }
    for (size_t j = 0; true; j++)
    {
      if (j >= arg.length())
      {
        if (++i >= argc)
        {
          throw std::runtime_error(std::string("missing value for option -")
                                   + arg);
        }
        valueStr = argv[i];
        break;
      }
      if (arg[j] == '=')
      {
        valueStr = arg.c_str() + (j + 1);
        arg.resize(j);
        break;
      }
    }
    if (arg == "cfg" || arg == "config")
      readConfigurationFile(valueStr);
    else
      parseOption(arg, valueStr);
  }
  for (size_t i = 0; i < fileNames.size(); i++)
  {
    if (fileNames[i]->empty())
      throw std::runtime_error(std::string("missing ") + fileNameNames[i]);
  }
  return true;
}

#endif  // OPTIONS_CPP_INCLUDED

