
#include "headers.hpp"
#include "dsplib.cpp"
#include "options.cpp"

struct AnalysisResults
{
  double  freq;
  double  amp;
  double  phase;
  double  freqCorrection;
  double  ampCorrection;
  double  startPos;             // in sample frames
  // --------
  AnalysisResults()
    : freq(0.0),
      amp(0.0),
      phase(0.0),
      freqCorrection(0.0),
      ampCorrection(0.0),
      startPos(0.0)
  {
  }
};

static double getMagnitude(const std::vector< double >& fftBuf, long n)
{
  if (n < 0L || n > long(fftBuf.size() >> 1))
    return 0.0;
  if (n == 0L)
    return fftBuf[0];
  if (n == long(fftBuf.size() >> 1))
    return fftBuf[1];
  double  re = fftBuf[n << 1];
  double  im = fftBuf[(n << 1) + 1];
  return std::sqrt((re * re) + (im * im));
}

static inline double getPhase(const std::vector< double >& fftBuf, long n)
{
  if (n <= 0L || n >= long(fftBuf.size() >> 1))
    return 0.0;
  return std::atan2(fftBuf[(n << 1) + 1], fftBuf[n << 1]);
}

static void analyzeSignal(double& freq, double& amp, double& phase,
                          const double *buf, size_t bufSize, double sampleRate,
                          int winParam)
{
  freq = 0.0;
  amp = 0.0;
  phase = 0.0;
  if (bufSize < 4)
    return;
  size_t  fftSize = 16;
  while (fftSize < bufSize)
    fftSize = fftSize << 1;
  std::vector< double > fftBuf(fftSize, 0.0);
  double  wsum = 0.0;
  double  c = 1.0;
  if (winParam < 0)
    c = std::pow(10.0, double(winParam) / 20.0);
  else
    c = 0.2 / double(long(bufSize));
  c = std::sqrt(-(std::log(c))) * 2.0;
  for (size_t i = 0; i < bufSize; i++)
  {
    double  x = ((double(long(i)) / double(long(bufSize - 1))) - 0.5) * c;
    double  w = std::exp(-x * x);
    wsum += (w * w);
    fftBuf[i] = buf[i] * w;
  }
  realFFT(&(fftBuf.front()), fftSize, false);
  double  maxMag = 0.0;
  long    maxPos = 0;
  for (long i = 0L; i <= long(fftSize >> 1); i++)
  {
    double  r = getMagnitude(fftBuf, i);
    if (r > maxMag)
    {
      maxMag = r;
      maxPos = i;
    }
  }
  if (maxMag < 1.0e-30)
    return;
  phase = getPhase(fftBuf, maxPos);
  for (long i = -9L; i <= 9L; i++)
  {
    double  r = getMagnitude(fftBuf, maxPos + i);
    amp += (r * r);
  }
  amp = std::sqrt(amp / (wsum * double(long(fftSize >> 2))));
  double  rm1 = getMagnitude(fftBuf, maxPos - 1L);
  double  r0 = maxMag;
  double  r1 = getMagnitude(fftBuf, maxPos + 1L);
  if (rm1 >= 1.0e-30 && r1 >= 1.0e-30 && ((r0 * 2.0) - (rm1 + r1)) >= 1.0e-30)
  {
    rm1 = std::log(rm1);
    r0 = std::log(r0);
    r1 = std::log(r1);
    double  d0 = r0 - rm1;
    double  d1 = r1 - r0;
    double  x = (d0 / (d0 - d1)) - 0.5;
    freq = (double(maxPos) + x) * sampleRate / double(long(fftSize));
    phase -= (x * pi * double(long(bufSize) - 1L) / double(long(fftSize)));
    if (phase < -pi)
      phase += twoPi;
    else if (phase >= pi)
      phase -= twoPi;
  }
  else
  {
    freq = double(maxPos) * sampleRate / double(long(fftSize));
  }
}

static void analyzeBuffer(AnalysisResults& results,
                          const std::vector< double >& buf,
                          double sampleRate, int winParam,
                          sf_count_t startPos_i,
                          double refFreq, double refAmp, double refPhase)
{
  double  maxVal = 0.0;
  for (size_t i = 0; i < buf.size(); i++)
  {
    double  x = std::fabs(buf[i]);
    maxVal = (x > maxVal ? x : maxVal);
  }
  size_t  startPos_ = buf.size();
  size_t  endPos_ = 0;
  double  minVal = maxVal * 0.25;
  for (size_t i = 0; i < buf.size(); i++)
  {
    double  x = std::fabs(buf[i]);
    if (x > minVal)
    {
      startPos_ = (i < startPos_ ? i : startPos_);
      endPos_ = i + 1;
    }
  }
  if (endPos_ > startPos_)
  {
    double  freq = 0.0;
    double  amp = 0.0;
    double  phase = 0.0;
    analyzeSignal(freq, amp, phase,
                  &(buf.front()) + startPos_, endPos_ - startPos_,
                  sampleRate, winParam);
    if (freq > 0.0 && amp > 0.0)
    {
      phase = (phase / twoPi) + 0.25;
      double  t = (refPhase / 360.0) - phase;
      phase = phase - (double(long(startPos_)) * freq / sampleRate);
      double  tmp = 0.0;
      phase = std::modf(phase, &tmp);
      if (phase < 0.0)
        phase += 1.0;
      phase *= 360.0;
      while (t < -0.5)
        t += 1.0;
      while (t >= 0.5)
        t -= 1.0;
      t = (t * sampleRate / freq) + double(startPos_i + sf_count_t(startPos_));
      results.freq = freq;
      results.amp = amp;
      results.phase = phase;
      if (refFreq > 0.0)
        results.freqCorrection = refFreq / freq;
      if (refAmp > 0.0)
        results.ampCorrection = refAmp / amp;
      results.startPos = t;
    }
  }
}

int main(int argc, char **argv)
{
  try
  {
    std::string fileName;
    double      startTime = 0.0;
    double      duration = 30.0;
    double      startTime2 = 0.0;
    double      duration2 = 0.0;
    int         channel = 0;
    int         winParam = 0;
    double      refFreq = 0.0;
    double      refAmp = 0.0;
    double      refPhase = 0.0;
    double      channelRefFreq[8] = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
    {
      CommandLineParser opts(argv[0]);
      opts.addFileName("FILENAME", "input file name", fileName);
      opts.addOption("skip", "k", "start time in seconds or -samples",
                     startTime);
      opts.addOption("duration", "d", "duration limit in seconds or -samples",
                     duration);
      opts.addOption("skip2", "k2",
                     "start time #2 in seconds or -samples (for calculating "
                     "frequency\n"
                     "        correction with two test signals, it needs to "
                     "be greater than\n"
                     "        the first start time by the exact amount of "
                     "delay required\n"
                     "        between the test tones)",
                     startTime2);
      opts.addOption("duration2", "d2",
                     "duration limit #2 in seconds or -samples; a second "
                     "signal is\n"
                     "        analyzed only if its start time is greater than "
                     "that of the\n"
                     "        first one, and its duration is greater than "
                     "zero",
                     duration2);
      opts.addOption("channel", "c", "channel to test (0 = all)",
                     channel, 0, 16);
      opts.addOption("winparam", "w",
                     "Gaussian window minimum level (-dB); "
                     "set automatically if zero", winParam, -200, 0);
      opts.addOption("freq", "f", "", refFreq, 0.0, 1000000.0);
      opts.addOption("amp", "a", "", refAmp, -120.0, 1000.0);
      opts.addOption("phase", "p",
                     "frequency (Hz), amplitude (+raw (0 dBFS = 1.0) "
                     "or -dBFS), and\n"
                     "        phase (degrees) of reference signal to compare "
                     "input against,\n"
                     "        if the frequency and amplitude are not zero",
                     refPhase, -180.0, 180.0);
      opts.addOption("channel1-freq", "f1", "",
                     channelRefFreq[0], 0.0, 1000000.0);
      opts.addOption("channel2-freq", "f2", "",
                     channelRefFreq[1], 0.0, 1000000.0);
      opts.addOption("channel3-freq", "f3", "",
                     channelRefFreq[2], 0.0, 1000000.0);
      opts.addOption("channel4-freq", "f4", "",
                     channelRefFreq[3], 0.0, 1000000.0);
      opts.addOption("channel5-freq", "f5", "",
                     channelRefFreq[4], 0.0, 1000000.0);
      opts.addOption("channel6-freq", "f6", "",
                     channelRefFreq[5], 0.0, 1000000.0);
      opts.addOption("channel7-freq", "f7", "",
                     channelRefFreq[6], 0.0, 1000000.0);
      opts.addOption("channel8-freq", "f8",
                     "reference frequency for channel 1 to 8, overrides -freq "
                     "if\n"
                     "        greater than zero",
                     channelRefFreq[7], 0.0, 1000000.0);
      if (!opts.parseCommandLine(argc, argv))
        return 0;
    }
    if (refAmp < 0.0)
      refAmp = std::pow(10.0, refAmp / 20.0);
    if (refFreq > 0.0)
    {
      for (size_t i = 0; i < (sizeof(channelRefFreq) / sizeof(double)); i++)
      {
        if (!(channelRefFreq[i] > 0.0))
          channelRefFreq[i] = refFreq;
      }
    }
    std::vector< std::vector< AnalysisResults > > results(2);
    double      sampleRate = 0.0;
    size_t      channels = 0;
    {
      std::vector< std::vector< double > >  buffers;
      AudioFilter_FileInput inFile(fileName.c_str());
      sampleRate = inFile.getSampleRate();
      channels = size_t(inFile.getChannelCount());
      if (channels > (sizeof(channelRefFreq) / sizeof(double)))
      {
        throw std::runtime_error(std::string("input file has unsupported "
                                             "number of channels"));
      }
      if (size_t(channel) > channels)
        throw std::runtime_error(std::string("channel number is out of range"));
      buffers.resize(channel < 1 ? channels : 1);
      for (int n = 0; n < 2; n++)
      {
        sf_count_t  startPos_i = 0;
        sf_count_t  duration_i = 0;
        {
          double& k = (n == 0 ? startTime : startTime2);
          double& d = (n == 0 ? duration : duration2);
          startPos_i = sf_count_t((k < 0.0 ? (-k) : (k * sampleRate)) + 0.5);
          duration_i = sf_count_t((d < 0.0 ? (-d) : (d * sampleRate)) + 0.5);
          sf_count_t  maxDuration = inFile.getFileLength() - startPos_i;
          duration_i = (duration_i < maxDuration ? duration_i : maxDuration);
        }
        if (n > 0 && !(startTime2 > startTime && duration_i > 0))
        {
          results.resize(1);
          break;
        }
        if (duration_i < 1)
          throw std::runtime_error(std::string("no data to analyze"));
        results[n].resize(channel < 1 ? channels : 1);
        inFile.setPosition(startPos_i, false);
        for (size_t i = 0; i < buffers.size(); i++)
        {
          buffers[i].resize(0);
          buffers[i].reserve(size_t(duration_i));
        }
        do
        {
          inFile.process();
          if (channel > 0)
          {
            buffers[0].push_back(inFile.getOutputBuffer()[channel - 1]);
          }
          else
          {
            for (size_t i = 0; i < channels; i++)
              buffers[i].push_back(inFile.getOutputBuffer()[i]);
          }
        }
        while (--duration_i > 0);
        for (size_t i = 0; i < buffers.size(); i++)
        {
          analyzeBuffer(results[n][i], buffers[i], sampleRate, winParam,
                        startPos_i,
                        channelRefFreq[(channel > 0 ? size_t(channel - 1) : i)],
                        refAmp, refPhase);
        }
      }
    }
    double  avgFreqCorrection1 = 0.0;
    double  avgFreqCorrection2 = 0.0;
    double  avgFreqCorrection = 0.0;
    std::vector< double > avgAmpCorrection(results[0].size(), 0.0);
    double  avgSkip1 = 0.0;
    double  avgSkip2 = 0.0;
    {
      int     fcCnt = 0;
      for (size_t i = 0; i < results[0].size(); i++)
      {
        if (results.size() < 2)
        {
          if (results[0][i].freqCorrection > 0.0)
          {
            avgFreqCorrection1 += results[0][i].freqCorrection;
            fcCnt++;
          }
          avgSkip1 += results[0][i].startPos;
        }
        else
        {
          if (results[0][i].freqCorrection > 0.0 &&
              results[1][i].freqCorrection > 0.0)
          {
            avgFreqCorrection1 += results[0][i].freqCorrection;
            avgFreqCorrection2 += results[1][i].freqCorrection;
            fcCnt++;
          }
          if (results[0][i].ampCorrection > 0.0 &&
              results[1][i].ampCorrection > 0.0)
          {
            avgAmpCorrection[i] = (0.5 * (results[0][i].ampCorrection
                                          + results[1][i].ampCorrection));
          }
          avgFreqCorrection +=
              ((results[1][i].startPos - results[0][i].startPos)
               / ((startTime2 - startTime) * sampleRate));
          avgSkip1 += results[0][i].startPos;
          avgSkip2 += results[1][i].startPos;
        }
      }
      if (fcCnt > 0)
      {
        avgFreqCorrection1 /= double(fcCnt);
        avgFreqCorrection2 /= double(fcCnt);
      }
      avgSkip1 /= double(int(results[0].size()));
      if (results.size() > 1)
      {
        avgSkip2 /= double(int(results[0].size()));
        avgFreqCorrection /= double(int(results[0].size()));
      }
      else
      {
        avgFreqCorrection = avgFreqCorrection1;
      }
    }
    for (size_t i = 0; i < results.size(); i++)
    {
      for (size_t j = 0; j < results[i].size(); j++)
      {
        int     chn = (channel > 0 ? channel : int(j + 1));
        if (results.size() < 2)
          std::printf("Channel #%d", chn);
        else if (i == 0)
          std::printf("Chn #%d (1)", chn);
        else
          std::printf("Chn #%d (2)", chn);
        std::printf(": frequency = %.8f Hz\n", results[i][j].freq);
        std::printf("\t    amplitude = %.8f (%.6f dBFS)\n",
                    results[i][j].amp, std::log10(results[i][j].amp) * 20.0);
        std::printf("\t    phase = %.6f degrees\n", results[i][j].phase);
        if (results[i][j].freqCorrection > 0.0)
        {
          std::printf("\t    frequency correction = %.12f\n",
                      results[i][j].freqCorrection);
        }
        if (results[i][j].ampCorrection > 0.0)
        {
          std::printf("\t    amplitude correction = %.8f (%.8f dB)\n",
                      results[i][j].ampCorrection,
                      std::log10(results[i][j].ampCorrection) * 20.0);
        }
        std::printf("\t    start time = %.12f s (%.7f samples)\n",
                    results[i][j].startPos / sampleRate,
                    results[i][j].startPos);
      }
    }
    if (results.size() > 1 || results[0].size() > 1)
    {
      std::printf("Overall");
      if (results.size() > 1)
      {
        if (avgFreqCorrection1 > 0.0)
        {
          std::printf("\tfrequency correction (1) = %.12f\n",
                      avgFreqCorrection1);
        }
        if (avgFreqCorrection2 > 0.0)
        {
          std::printf("\tfrequency correction (2) = %.12f\n",
                      avgFreqCorrection2);
        }
      }
      if (avgFreqCorrection > 0.0)
        std::printf("\tfrequency correction = %.14f\n", avgFreqCorrection);
      for (size_t i = 0; i < avgAmpCorrection.size(); i++)
      {
        if (avgAmpCorrection[i] > 0.0)
        {
          std::printf("\tamplitude correction (channel #%d) = %.8f (%.8f dB)\n",
                      (channel > 0 ? channel : int(i + 1)),
                      avgAmpCorrection[i],
                      std::log10(avgAmpCorrection[i]) * 20.0);
        }
      }
      if (results.size() < 2)
      {
        std::printf("\tstart time = %.12f s (%.7f samples)\n",
                    avgSkip1 / sampleRate, avgSkip1);
      }
      else
      {
        std::printf("\tstart time (1) = %.12f s (%.7f samples)\n",
                    avgSkip1 / sampleRate, avgSkip1);
        std::printf("\tstart time (2) = %.12f s (%.7f samples)\n",
                    avgSkip2 / sampleRate, avgSkip2);
      }
    }
  }
  catch (std::exception& e)
  {
    std::fprintf(stderr, "%s: %s\n", argv[0], e.what());
    return -1;
  }
  return 0;
}

