
#include "headers.hpp"
#include "dsplib.cpp"
#include "options.cpp"
#include "image.cpp"

static const int  imageWidth = 1024;
static const int  imageHeight = 768;
static const int  displayWidth = 776;
static const int  displayHeight = 352;
static const int  displayFrames = 320;
static const int  displayOriginX = 184;
static const int  displayOriginY = 408;

static const int  MAX_FFT_SIZE = 32768;

struct AnalysisParameters
{
  double  sampleRate;
  double  minFreq;
  double  maxFreq;
  double  startPos;
  double  duration;
  double  bandWidthMin;
  double  bandWidthMax;
  double  bandWidthMult;
  double  windowRiseTime;               // as a fraction of the total length,
                                        // or -ms
  int     dynRange;                     // in dB
  int     dBRangeMax;
  double  frameOffset;
};

class CSDAnalyzer
{
 protected:
  std::vector< std::vector< double > >  frameBufs;
  // --------
  static size_t analyzeWindow(std::vector< double >& fftBuf,
                              const std::vector< double >& buf, double t,
                              double bandWidth,
                              const AnalysisParameters& params);
 public:
  CSDAnalyzer(const std::vector< double >& buf,
              const AnalysisParameters& params);
  virtual ~CSDAnalyzer();
  double getMaxMagnitude() const;
  void getFrame(std::vector< double >& magnitudes, size_t n) const;
};

size_t CSDAnalyzer::analyzeWindow(std::vector< double >& fftBuf,
                                  const std::vector< double >& buf, double t,
                                  double bandWidth,
                                  const AnalysisParameters& params)
{
  double  windowLength = 4.0 * params.sampleRate / bandWidth;
  windowLength = (windowLength > 32.0 ?
                  (windowLength < double(MAX_FFT_SIZE - 1) ?
                   windowLength : double(MAX_FFT_SIZE - 1)) : 32.0);
  t = t * params.sampleRate + params.frameOffset;
  double  winRiseLen = (params.windowRiseTime < 0.0 ?
                        (-0.002 * params.sampleRate * params.windowRiseTime)
                        : (windowLength * params.windowRiseTime));
  winRiseLen = (winRiseLen > 1.0 ?
                (winRiseLen < (windowLength - 1.0) ?
                 winRiseLen : (windowLength - 1.0))
                : 1.0);
  double  winFallLen = windowLength - winRiseLen;
  long    startPos = long(std::floor(t - winRiseLen));
  long    endPos = long(std::ceil(t + winFallLen)) + 1L;
  long    peakPos = long(std::ceil(t));
  if (startPos < 0L)
    startPos = 0L;
  if (endPos > long(buf.size()))
    endPos = long(buf.size());
  size_t  fftSize = 64;
  while (fftSize < size_t(endPos - startPos))
    fftSize = fftSize << 1;
  if (fftBuf.size() < fftSize)
    fftBuf.resize(fftSize, 0.0);
  if (peakPos > startPos)
  {
    GaussianWindow  w(winRiseLen * 2.0, 0.000000654869,
                      t - double(peakPos - 1L));
    for (long i = peakPos; --i >= startPos; )
      fftBuf[i - startPos] = buf[i] * w.sample();
  }
  if (peakPos < endPos)
  {
    GaussianWindow  w(winFallLen * 2.0, 0.000000654869, double(peakPos) - t);
    for (long i = peakPos; i < endPos; i++)
      fftBuf[i - startPos] = buf[i] * w.sample();
  }
  for (size_t i = size_t(endPos - startPos); i < fftSize; i++)
    fftBuf[i] = 0.0;
  realFFT(&(fftBuf.front()), fftSize, false);
  fftBuf[0] = std::fabs(fftBuf[0]);
  double  tmp = std::fabs(fftBuf[1]);   // save magnitude at Nyquist frequency
  for (size_t i = 2; i < fftSize; i += 2)
  {
    double  re = fftBuf[i];
    double  im = fftBuf[i + 1];
    fftBuf[i >> 1] = std::sqrt((re * re) + (im * im));
  }
  fftBuf[fftSize >> 1] = tmp;
  return fftSize;
}

CSDAnalyzer::CSDAnalyzer(const std::vector< double >& buf,
                         const AnalysisParameters& params)
{
  std::vector< double > freqTable((size_t) displayWidth, 0.0);
  for (size_t i = 0; i < freqTable.size(); i++)
  {
    freqTable[i] = params.minFreq
                   * std::pow(params.maxFreq / params.minFreq,
                              double(long(i)) / double(displayWidth - 1));
  }
  std::vector< double > bandWidthTable;
  if (!(params.bandWidthMax > (params.bandWidthMin * 1.125)))
  {
    bandWidthTable.push_back(std::sqrt(params.bandWidthMin
                                       * params.bandWidthMax));
  }
  else
  {
    // n = 0 for window length = 31.5, n = 30 for window length = 32256
    int     n0 = int(std::floor(3.0 * std::log(4.0 * params.sampleRate
                                               / (params.bandWidthMax * 31.5))
                                / std::log(2.0) + 0.0625));
    int     n1 = int(std::ceil(3.0 * std::log(4.0 * params.sampleRate
                                              / (params.bandWidthMin * 31.5))
                               / std::log(2.0) - 0.0625));
    int     nMax = int(std::floor(3.0 * std::log(double(MAX_FFT_SIZE - 1)
                                                 / 31.5)
                                  / std::log(2.0) + 0.5));
    n0 = (n0 > 0 ? (n0 < nMax ? n0 : nMax) : 0);
    n1 = (n1 > 0 ? (n1 < nMax ? n1 : nMax) : 0);
    for (int n = n1; n >= n0; n--)
    {
      double  bw =
          4.0 * params.sampleRate / (31.5 * std::pow(2.0, double(n) / 3.0));
      bandWidthTable.push_back(bw);
    }
  }
  frameBufs.resize(size_t(displayFrames));
  std::vector< double > fftBuf0;
  std::vector< double > fftBuf1;
  for (size_t i = 0; i < frameBufs.size(); i++)
  {
    double  t = params.duration * double(long(i)) / double(displayFrames - 1);
    frameBufs[i].resize(freqTable.size(), 0.0);
    std::vector< double > *fftbp0 = &fftBuf0;
    std::vector< double > *fftbp1 = &fftBuf1;
    size_t  fftSize0 = 0;
    size_t  fftSize1 = 0;
    int     n0 = -1;            // bandwidth index
    int     n1 = -1;
    for (size_t j = 0; j < freqTable.size(); j++)
    {
      double  freq = freqTable[j];
      if (freq < 0.0 || freq > (params.sampleRate * 0.5))
      {
        frameBufs[i][j] = 0.0;
        continue;
      }
      double  bw = freq * params.bandWidthMult;
      bw = (bw > params.bandWidthMin ?
            (bw < params.bandWidthMax ?
             bw : params.bandWidthMax)
            : params.bandWidthMin);
      int     n = 0;
      while (n < int(bandWidthTable.size()) && bw >= bandWidthTable[n])
        n++;
      n--;
      if (n == n1)
      {
        std::vector< double > *tmp = fftbp0;
        fftbp0 = fftbp1;
        fftbp1 = tmp;
        fftSize0 = fftSize1;
        fftSize1 = 0;
        n0 = n1;
        n1 = -1;
      }
      else if (n != n0)
      {
        n0 = n;
        if (n0 >= 0)
          fftSize0 = analyzeWindow(*fftbp0, buf, t, bandWidthTable[n0], params);
        else
          fftSize0 = 0;
      }
      if ((n + 1) != n1)
      {
        n1 = n + 1;
        if (n1 < int(bandWidthTable.size()))
          fftSize1 = analyzeWindow(*fftbp1, buf, t, bandWidthTable[n1], params);
        else
          fftSize1 = 0;
      }
      double  r0 = 0.0;
      if (fftSize0 > 0)
      {
        double  fftBin = double(long(fftSize0)) * freq / params.sampleRate;
        int     bin0 = int(fftBin);
        int     bin1 = bin0 + 1;
        double  f = fftBin - double(bin0);
        if (bin1 <= 0)
          r0 = (*fftbp0)[0];
        else if (bin0 >= int(fftSize0 >> 1))
          r0 = (*fftbp0)[fftSize0 >> 1];
        else
          r0 = (*fftbp0)[bin0] + (((*fftbp0)[bin1] - (*fftbp0)[bin0]) * f);
      }
      double  r1 = 0.0;
      if (fftSize1 > 0)
      {
        double  fftBin = double(long(fftSize1)) * freq / params.sampleRate;
        int     bin0 = int(fftBin);
        int     bin1 = bin0 + 1;
        double  f = fftBin - double(bin0);
        if (bin1 <= 0)
          r1 = (*fftbp1)[0];
        else if (bin0 >= int(fftSize1 >> 1))
          r1 = (*fftbp1)[fftSize1 >> 1];
        else
          r1 = (*fftbp1)[bin0] + (((*fftbp1)[bin1] - (*fftbp1)[bin0]) * f);
      }
      if (fftSize0 > 0 && fftSize1 > 0)
      {
        double  f = std::log(bw / bandWidthTable[n0])
                    / std::log(bandWidthTable[n1] / bandWidthTable[n0]);
        frameBufs[i][j] = r0 + ((r1 - r0) * f);
      }
      else
      {
        frameBufs[i][j] = (fftSize0 > 0 ? r0 : r1);
      }
    }
  }
}

CSDAnalyzer::~CSDAnalyzer()
{
}

double CSDAnalyzer::getMaxMagnitude() const
{
  double  maxMagnitude = 0.0;
  for (size_t i = 0; i < frameBufs.size(); i++)
  {
    for (size_t j = 0; j < frameBufs[i].size(); j++)
    {
      if (frameBufs[i][j] > maxMagnitude)
        maxMagnitude = frameBufs[i][j];
    }
  }
  return maxMagnitude;
}

void CSDAnalyzer::getFrame(std::vector< double >& magnitudes, size_t n) const
{
  size_t  i = 0;
  if (n < frameBufs.size())
  {
    for ( ; i < magnitudes.size() && i < frameBufs[n].size(); i++)
      magnitudes[i] = frameBufs[n][i];
  }
  for ( ; i < magnitudes.size(); i++)
    magnitudes[i] = 0.0;
}

// ----------------------------------------------------------------------------

static void drawHVPixel(Image& img, int x, int y, float h, float v)
{
  float   d = img.getDitherOffset(x, y);
  h = (h > 0.0f ? (h < 1.0f ? h : 1.0f) : 0.0f) * 30.0f + d;
  v = (v > 0.0f ? (v < 1.0f ? v : 1.0f) : 0.0f) * 7.0f + d;
  img.setPixel(x, y, (unsigned char) ((int(h) << 3) + int(v) + 8));
}

// buf[(y * imageWidth + x) * 3] = Z distance from peak at this sample
// buf[(y * imageWidth + x) * 3 + 1] = Y coordinate (frame number)
// buf[(y * imageWidth + x) * 3 + 2] = Z coordinate + 1,
//                                     or 0 for unused pixel,
//                                     or -1 for grid line

static void drawGraph(Image& img, const std::vector< short >& buf)
{
  float   hScale = 1.0f / float(displayHeight - 1);
  {
    int     y = 0;
    do
    {
      for (int i = 0; i < 8; i++)
      {
        float   h = float(y) * hScale;
        drawHVPixel(img, displayOriginX + i - (displayFrames >> 1) - 16,
                    displayOriginY + displayFrames - y, h, 1.0f);
        drawHVPixel(img, displayOriginX + i + displayWidth + 8,
                    displayOriginY - y, h, 1.0f);
      }
    }
    while (++y < displayHeight);
  }

  for (int y = 0; y < imageHeight; y++)
  {
    for (int x = 0; x < imageWidth; x++)
    {
      const short *bufp = &(buf.front()) + ((y * imageWidth + x) * 3);
      if (bufp[2] == 0)
        continue;
      if (bufp[2] < 0)
      {
        img.setPixel(x, y, 1);
        continue;
      }
      float   h = float(bufp[2]) * hScale;
      float   v = (float(bufp[1]) + float((displayFrames - 1) * 3))
                  / (float((displayFrames - 1) * 4) * float(bufp[0] + 1));
      drawHVPixel(img, x, y, h, v);
    }
  }
}

static void drawFrame(std::vector< short >& imgBuf,
                      const std::vector< double >& magnitudes,
                      int x0, int y0, int h, double dBmin, double dBmax,
                      const std::vector< unsigned char >& gridMapX,
                      unsigned char gridY)
{
  double  minVal = std::pow(10.0, dBmin / 20.0);
  double  maxVal = std::pow(10.0, dBmax / 20.0);
  double  c1 = 1.0 / minVal;
  double  c2 = double(h) / std::log(maxVal * c1);
  int     w = int(magnitudes.size());
  for (int i = 0; i < w; i++)
  {
    double  r = magnitudes[i];
    if (!(r >= minVal))
      continue;
    int     n = h;
    if (r < maxVal)
      n = int(std::log(r * c1) * c2 + 0.5);
    if (n < 0)
      continue;
    else if (n > h)
      n = h;
    int     x = x0 + i;
    int     y = y0 - n;
    short   *bufp0 = &(imgBuf.front()) + ((y * imageWidth + x) * 3);
    short   *bufp = bufp0;
    do
    {
      bufp[0] = short((y0 - y) - n);
      bufp[1] = short(y0 - displayOriginY);
      bufp[2] = short(n + 1);
      bufp = bufp + (imageWidth * 3);
    }
    while (--n >= 0);
    if (gridMapX[i] > 4 || (gridMapX[i] && (y & 1) != 0) || gridY >= 6)
    {
      bufp0[2] = -1;
      if (y < y0)
        bufp0[imageWidth * 3 + 2] = -1;
    }
  }
}

// axis = 0: X (frequency in Hz)
// axis = 1: Y (time in s)
// axis = 2: Z (magnitude in dB)

static void drawTick(Image& img, const AnalysisParameters& params, double n,
                     int axis, unsigned char c,
                     std::vector< unsigned char >& gridMap)
{
  char    labelBuf[256];
  if (axis == 0 || axis == 1)
  {
    if (n < 0.0009995)
      std::sprintf(labelBuf, "%gu", n * 1000000.0);
    else if (n < 0.9995)
      std::sprintf(labelBuf, "%gm", n * 1000.0);
    else if (n < 999.5)
      std::sprintf(labelBuf, "%g", n);
    else if (n < 999500.0)
      std::sprintf(labelBuf, "%gk", n * 0.001);
    else
      std::sprintf(labelBuf, "%gM", n * 0.000001);
  }
  switch (axis)
  {
  case 0:
    {
      if (!(n >= params.minFreq && n <= params.maxFreq))
        return;
      n = double(displayWidth - 1) * std::log(n / params.minFreq)
          / std::log(params.maxFreq / params.minFreq);
      gridMap[int(n + 0.5)] = c;
      int     x0 = int(n + 0.5) + displayOriginX;
      int     y0 = displayOriginY;
      int     x1 = x0 - (displayFrames >> 1);
      int     y1 = y0 + displayFrames;
      if (c == 7 || labelBuf[0] == '2' || labelBuf[0] == '5')
      {
        img.drawText(x1 - 8, y1 + 8, 1, labelBuf);
        img.drawText(x0 - 8, y0 - displayHeight - 21, 1, labelBuf);
      }
      img.drawLine(x0, y0, x1, y1, c);
      img.drawLine(x0, y0, x0, y0 - displayHeight, c);
      img.drawLine(x1, y1, x1, y1 + 4, 1);
      img.drawLine(x0, y0 - displayHeight, x0, y0 - displayHeight - 4, 1);
    }
    break;
  case 1:
    {
      if (!(n >= 0.0 && n <= params.duration))
        return;
      n = double(displayFrames - 1) * n / params.duration;
      gridMap[int(n + 0.5)] = c;
      int     x0 = displayOriginX - ((int(n + 0.5) + 1) >> 1);
      int     y0 = displayOriginY + int(n + 0.5);
      int     x1 = x0 + displayWidth;
      int     y1 = y0;
      img.drawLine(x0, y0, x1, y1, c);
      img.drawLine(x0, y0, x0, y0 - displayHeight, c);
      if (c == 7)
      {
        if (y1 != displayOriginY)
          img.drawText(x1 + 12, y1 - 6, 1, labelBuf);
        img.drawLine(x1, y1, x1 + 4, y1, 1);
        img.drawLine(x0, y0 - displayHeight, x0 - 4, y0 - displayHeight, 1);
      }
    }
    break;
  case 2:
    {
      if (!(n >= double(params.dBRangeMax - params.dynRange) &&
            n <= double(params.dBRangeMax)))
        return;
      std::sprintf(labelBuf, "%g", n);
      n = double(displayHeight - 1)
          * (n - double(params.dBRangeMax - params.dynRange))
          / double(params.dynRange);
      gridMap[int(n + 0.5)] = c;
      int     x0 = displayOriginX;
      int     y0 = displayOriginY - int(n + 0.5);
      int     x1 = x0 + displayWidth;
      int     y1 = y0;
      img.drawText(x1 + 20, y1 - 6, 1, labelBuf);
      img.drawLine(x0, y0, x1, y1, c);
      img.drawLine(x0, y0, x0 - (displayFrames >> 1), y0 + displayFrames, c);
      img.drawLine(x1, y1, x1 + 4, y1, 1);
      img.drawLine(x0 - (displayFrames >> 1), y0 + displayFrames,
                   x0 - (displayFrames >> 1) - 4, y0 + displayFrames, 1);
    }
    break;
  }
}

int main(int argc, char **argv)
{
  try
  {
    std::string inFileName = "";
    std::string outFileName = "";
    std::string titleString = "";
    std::string descriptionString = "";
    AnalysisParameters  params;
    params.sampleRate = 96000.0;
    params.minFreq = 100.0;
    params.maxFreq = 22000.0;
    params.startPos = 0.0;
    params.duration = 0.01;
    params.bandWidthMin = 50.0;
    params.bandWidthMax = 2500.0;
    params.bandWidthMult = 0.25;
    params.windowRiseTime = 0.125;
    params.dynRange = 48;
    params.dBRangeMax = 1000;
    params.frameOffset = 0.0;
    int     channel = 1;
    bool    blackBackground = false;
    {
      CommandLineParser opts(argv[0]);
      opts.addFileName("INFILE", "input (audio) file name", inFileName);
      opts.addFileName("OUTFILE", "output (BMP image) file name", outFileName);
      opts.addOption("title", "ts",
                     "title string (at most 114 characters), the default is "
                     "the input\n"
                     "        file name",
                     titleString);
      opts.addOption("description", "ds",
                     "description string (at most 126 characters), defaults "
                     "to the\n"
                     "        analysis parameters",
                     descriptionString);
      opts.addOption("black-background", "bk",
                     "use alternate color map with black background",
                     blackBackground);
      opts.addOption("skip", "k", "start time in seconds or -samples",
                     params.startPos);
      opts.addOption("duration", "d", "analysis duration in seconds",
                     params.duration, 0.0005, 10.0);
      opts.addOption("channel", "c", "channel to analyze",
                     channel, 1, 8);
      opts.addOption("minfreq", "fmin", "minimum frequency to display (in Hz)",
                     params.minFreq, 1.0, 10000.0);
      opts.addOption("maxfreq", "fmax", "maximum frequency to display",
                     params.maxFreq, 100.0, 100000.0);
      opts.addOption("min-bandwidth", "bwmin", "minimum bandwidth (in Hz)",
                     params.bandWidthMin, 5.0, 10000.0);
      opts.addOption("max-bandwidth", "bwmax", "maximum bandwidth",
                     params.bandWidthMax, 5.0, 10000.0);
      opts.addOption("bandwidth-mult", "bwm", "bandwidth / frequency",
                     params.bandWidthMult, 0.01, 4.0);
      opts.addOption("window-rise-time", "wt",
                     "window rise time as a fraction of the total length, or "
                     "in -ms\n"
                     "        (the window is symmetrical with -wt 0.5 or "
                     "-1000/bandwidth)",
                     params.windowRiseTime, -200.0, 1.0);
      opts.addOption("dynamic-range", "dr", "dB range to be displayed",
                     params.dynRange, 1, 160);
      opts.addOption("maximum-level", "dbmax",
                     "maximum dB level displayed (the default is to "
                     "normalize the\n"
                     "        display)",
                     params.dBRangeMax, -120, 120);
      if (!opts.parseCommandLine(argc, argv))
        return 0;
      if (!(params.maxFreq > (params.minFreq * 1.001)))
        throw std::runtime_error(std::string("invalid frequency range"));
      if (!(params.bandWidthMax >= params.bandWidthMin))
        throw std::runtime_error(std::string("invalid bandwidth range"));
    }

    Image   img(imageWidth, imageHeight);
    if (blackBackground)
    {
      img.setColor(0, 0, 0, 0);
      img.setColor(1, 255, 255, 255);
      img.setColor(2, 64, 64, 64);
      img.setColor(3, 96, 96, 96);
      img.setColor(4, 128, 128, 128);
      img.setColor(5, 160, 160, 160);
      img.setColor(6, 192, 192, 192);
      img.setColor(7, 224, 224, 224);
    }
    else
    {
      img.setColor(0, 255, 255, 255);
      img.setColor(1, 0, 0, 0);
      img.setColor(2, 224, 224, 224);
      img.setColor(3, 192, 192, 192);
      img.setColor(4, 160, 160, 160);
      img.setColor(5, 128, 128, 128);
      img.setColor(6, 96, 96, 96);
      img.setColor(7, 64, 64, 64);
    }
    for (int i = 8; i <= 255; i++)
    {
      double  h = double((i - 8) >> 3) * pi / 30.0;
      double  v = ((double(i & 7) / 7.0) * 2.0 + 1.0) / 3.0;
      double  r = -(std::cos(h));
      double  g = std::sin(h);
      double  b = std::cos(h);
      v /= (r > g ? (r > b ? r : b) : (g > b ? g : b));
      r = (r > 0.0 ? (r * v) : 0.0) * 255.0 + 0.5;
      g = (g > 0.0 ? (g * v) : 0.0) * 255.0 + 0.5;
      b = (b > 0.0 ? (b * v) : 0.0) * 255.0 + 0.5;
      img.setColor(i, (unsigned char) r, (unsigned char) g, (unsigned char) b);
    }

    std::vector< double > buf;
    {
      AudioFilter_FileInput f(inFileName.c_str());
      if (channel > f.getChannelCount())
        throw std::runtime_error(std::string("channel number is out of range"));
      params.sampleRate = f.getSampleRate();
      double  minBandWidth = 4.0 * params.sampleRate / double(MAX_FFT_SIZE - 1);
      double  maxBandWidth = double(displayFrames - 1)
                             / (params.duration * 4.0);
      if (!(params.bandWidthMin > minBandWidth))
        params.bandWidthMin = minBandWidth;
      if (!(params.bandWidthMax < maxBandWidth))
        params.bandWidthMax = maxBandWidth;
      if (!(params.bandWidthMin < (params.bandWidthMax * 1.000001)))
        throw std::runtime_error(std::string("invalid bandwidth range"));
      double  startFrame =
          (params.startPos < 0.0 ?
           -(params.startPos) : (params.startPos * params.sampleRate));
      long    fileOffset = long(startFrame + 0.5) - MAX_FFT_SIZE;
      params.frameOffset = startFrame - double(fileOffset);
      f.setPosition(sf_count_t(fileOffset), false);
      buf.resize(size_t(long(params.duration * params.sampleRate + 0.5))
                 + (size_t(MAX_FFT_SIZE) << 1), 0.0);
      for (size_t i = 0; i < buf.size(); i++)
      {
        f.process();
        buf[i] = f.getOutputBuffer()[channel - 1];
      }
    }
    CSDAnalyzer analyzer(buf, params);
    if (params.dBRangeMax > 999)
    {
      double  r = analyzer.getMaxMagnitude();
      r = (r > 1.0e-10 ? r : 1.0e-10);
      params.dBRangeMax = int(std::ceil(20.0 * std::log10(r)));
    }

    {
      size_t  maxLen = size_t(imageWidth >> 3) - 2;
      if (titleString.empty())
        titleString = inFileName;
      if (titleString.length() > (maxLen - 12))
        titleString.resize(maxLen - 12);
      img.drawText(8, 3, 1, titleString.c_str(), true);
      char    strBuf[256];
      std::sprintf(strBuf, "- channel %d", channel);
      img.drawText((int(titleString.length()) << 3) + 16, 3, 1, strBuf);
      if (descriptionString.empty())
      {
        std::sprintf(strBuf,
                     "Bandwidth: min = %g Hz, mult = %g, max = %g Hz    "
                     "Window type: Gaussian    Window rise time%s: %g%s",
                     params.bandWidthMin, params.bandWidthMult,
                     params.bandWidthMax,
                     (params.windowRiseTime < 0.0 ? "" : " / length"),
                     std::fabs(params.windowRiseTime),
                     (params.windowRiseTime < 0.0 ? " ms" : ""));
        img.drawText(8, 19, 6, strBuf);
      }
      else
      {
        if (descriptionString.length() > maxLen)
          descriptionString.resize(maxLen);
        img.drawText(8, 19, 6, descriptionString.c_str());
      }
    }

    std::vector< unsigned char >  gridMapX(size_t(displayWidth), 0);
    std::vector< unsigned char >  gridMapY(size_t(displayFrames), 0);
    std::vector< unsigned char >  gridMapZ(size_t(displayHeight), 0);
    {
      for (int f = 1; f <= 1000000; f *= 10)
      {
        drawTick(img, params, double(f) * 1.5, 0, 3, gridMapX);
        for (int f_ = 2; f_ <= 9; f_++)
          drawTick(img, params, double(f * f_), 0, 5, gridMapX);
      }
      double  tt = 0.000001;
      while (tt < (params.duration * 0.00999) && tt < 9.99)
        tt = tt * 10.0;
      for (int t = 0; t <= 100; t++)
      {
        drawTick(img, params, double(t) * tt, 1, ((t % 10) == 0 ? 7 : 3),
                 gridMapY);
      }
      int     mt = 1;
      while ((mt * 25) < params.dynRange && mt < 100)
        mt *= 10;
      int     r = (params.dBRangeMax - params.dynRange)
                  - ((params.dBRangeMax - params.dynRange) % mt);
      do
      {
        drawTick(img, params, double(r), 2, 6, gridMapZ);
        r = r + mt;
      }
      while (r <= params.dBRangeMax);
      for (int f = 1; f <= 1000000; f *= 10)
        drawTick(img, params, double(f), 0, 7, gridMapX);
    }

    img.drawText(displayOriginX + displayWidth + 36,
                 displayOriginY - displayHeight - 22, 1, "dB", true);
    img.drawText(displayOriginX + displayWidth - 32,
                 displayOriginY + displayFrames - 14, 1, "Time (s)", true);
    img.drawText(displayOriginX + ((displayWidth - displayFrames) >> 1) - 56,
                 displayOriginY + displayFrames + 22, 1, "Frequency (Hz)",
                 true);

    img.drawLine(displayOriginX, displayOriginY,
                 displayOriginX + displayWidth, displayOriginY, 1);
    img.drawLine(displayOriginX, displayOriginY - displayHeight,
                 displayOriginX + displayWidth, displayOriginY - displayHeight,
                 1);
    img.drawLine(displayOriginX, displayOriginY,
                 displayOriginX, displayOriginY - displayHeight, 1);
    img.drawLine(displayOriginX + displayWidth, displayOriginY,
                 displayOriginX + displayWidth, displayOriginY - displayHeight,
                 1);
    img.drawLine(displayOriginX, displayOriginY,
                 displayOriginX - (displayFrames >> 1),
                 displayOriginY + displayFrames, 1);
    img.drawLine(displayOriginX, displayOriginY - displayHeight,
                 displayOriginX - (displayFrames >> 1),
                 displayOriginY + displayFrames - displayHeight, 1);
    img.drawLine(displayOriginX - (displayFrames >> 1),
                 displayOriginY + displayFrames,
                 displayOriginX - (displayFrames >> 1),
                 displayOriginY + displayFrames - displayHeight, 1);
    img.drawLine(displayOriginX - (displayFrames >> 1),
                 displayOriginY + displayFrames,
                 displayOriginX + displayWidth - (displayFrames >> 1),
                 displayOriginY + displayFrames, 1);
    img.drawLine(displayOriginX + displayWidth, displayOriginY,
                 displayOriginX + displayWidth - (displayFrames >> 1),
                 displayOriginY + displayFrames, 1);

    std::vector< short >  imgBuf(size_t(imageWidth * imageHeight * 3), 0);
    std::vector< double > magnitudes(size_t(displayWidth), 0.0);
    for (int i = 0; i < displayFrames; i++)
    {
      analyzer.getFrame(magnitudes, i);
      drawFrame(imgBuf, magnitudes,
                displayOriginX - ((i + 1) >> 1), displayOriginY + i,
                displayHeight - 1,
                double(params.dBRangeMax - params.dynRange),
                double(params.dBRangeMax), gridMapX, gridMapY[i]);
    }
    drawGraph(img, imgBuf);
    img.writeBMPFile(outFileName.c_str());
  }
  catch (std::exception& e)
  {
    std::fprintf(stderr, "%s: %s\n", argv[0], e.what());
    return -1;
  }
  return 0;
}

