
#ifndef DSPLIB_CPP_INCLUDED
#define DSPLIB_CPP_INCLUDED

#include "headers.hpp"
#include "dsplib.hpp"

#include <ctime>

// ----------------------------------------------------------------------------

SineGen::SineGen(double n)
{
  long double omega = twoPiL / n;
  s1 = 0.0L;
  s2 = -(std::sin(omega));
  c = std::cos(omega) * 2.0L;
}

SineGen::SineGen(double n, double p)
{
  long double omega = twoPiL / n;
  s1 = std::sin((long double) p);
  s2 = std::sin((long double) p - omega);
  c = std::cos(omega) * 2.0L;
}

CosineGen::CosineGen(double n)
{
  long double omega = twoPiL / n;
  s1 = 1.0L;
  s2 = std::cos(omega);
  c = s2 * 2.0L;
}

CosineGen::CosineGen(double n, double p)
{
  long double omega = twoPiL / n;
  s1 = std::cos((long double) p);
  c = std::cos(omega);
  // this can be compiled using the FSINCOS instruction,
  // and is faster than std::cos(p - omega)
  s2 = (s1 * c) + (std::sin((long double) p) * std::sin(omega));
  c *= 2.0L;
}

GaussianWindow::GaussianWindow(long n, double wMin)
{
  if (n < 2L)
    throw std::runtime_error(std::string("invalid window size"));
  y = (long double) wMin;
  long double nD2 = (long double) (n - 1L) * 0.5L;
  c = std::log((long double) wMin) / (nD2 * nD2);
  c1 = std::exp((-2.0L * nD2 + 1.0L) * c);
  c2 = std::exp(2.0L * c);
}

GaussianWindow::GaussianWindow(long double nM1, double wMin,
                               long double startPos)
{
  if (nM1 < 1.0L)
    throw std::runtime_error(std::string("invalid window size"));
  c = std::log((long double) wMin) * 4.0L / (nM1 * nM1);
  y = std::exp(startPos * startPos * c);
  c1 = std::exp((2.0L * startPos + 1.0L) * c);
  c2 = std::exp(2.0L * c);
}

GaussianWindowNormalized::GaussianWindowNormalized(long n, double wMin)
  : GaussianWindow(n, wMin),
    offs(-wMin),
    mult(1.0 / (1.0 - wMin))
{
}

// ============================================================================
// |                               FFT LIBRARY                                |
// ============================================================================

FFTSineGen::FFTSineGen(long n)
{
  phase_re = 1.0L;
  phase_im = 0.0L;
  long double   omega = -twoPiL / (long double) n;
  omega_re = std::cos(omega);
  omega_im = std::sin(omega);
}

// ----------------------------------------------------------------------------

#ifndef USE_OOURA_FFT
// called by complexFFT() to calculate FFT
// the complex data in 'buf' is expected to be in bit-reversed order,
// and 'n' should be greater than 1; 'invFFTLen' is the full length (N)
// for inverse FFT, and zero otherwise

static void recursiveFFT(double *buf, size_t n, int invFFTLen)
{
  if (n > 512)
  {
    // recursively partition large FFT buffers to improve cache efficiency
    recursiveFFT(buf, n >> 1, invFFTLen);
    recursiveFFT(buf + n, n >> 1, invFFTLen);
  }
  else
  {
    double  *endp = buf + (n << 1);
    double  *bufp = buf;
    if (!invFFTLen)
    {
      // optimized code for small values of 'm' (2 and 4)
      do
      {
        double  r1 = bufp[0];
        double  i1 = bufp[1];
        double  r2 = bufp[2];
        double  i2 = bufp[3];
        bufp[0] = r1 + r2;
        bufp[1] = i1 + i2;
        bufp[2] = r1 - r2;
        bufp[3] = i1 - i2;
        bufp = bufp + 4;
      }
      while (bufp < endp);
      if (n < 4)
        return;
      bufp = buf;
      do
      {
        double  r1 = bufp[0];
        double  i1 = bufp[1];
        double  r2 = bufp[4];
        double  i2 = bufp[5];
        bufp[0] = r1 + r2;
        bufp[1] = i1 + i2;
        bufp[4] = r1 - r2;
        bufp[5] = i1 - i2;
        r1 = bufp[2];
        i1 = bufp[3];
        r2 = bufp[7];
        i2 = -(bufp[6]);
        bufp[2] = r1 + r2;
        bufp[3] = i1 + i2;
        bufp[6] = r1 - r2;
        bufp[7] = i1 - i2;
        bufp = bufp + 8;
      }
      while (bufp < endp);
    }
    else
    {
      // inverse FFT: optimized code for small values of 'm' (2 and 4)
      double  a = 1.0 / double(invFFTLen);
      do
      {
        double  r1 = bufp[0];
        double  i1 = bufp[1];
        double  r2 = bufp[2];
        double  i2 = bufp[3];
        // scale inverse FFT by 1 / N
        bufp[0] = (r1 + r2) * a;
        bufp[1] = (i1 + i2) * a;
        bufp[2] = (r1 - r2) * a;
        bufp[3] = (i1 - i2) * a;
        bufp = bufp + 4;
      }
      while (bufp < endp);
      if (n < 4)
        return;
      bufp = buf;
      do
      {
        double  r1 = bufp[0];
        double  i1 = bufp[1];
        double  r2 = bufp[4];
        double  i2 = bufp[5];
        bufp[0] = r1 + r2;
        bufp[1] = i1 + i2;
        bufp[4] = r1 - r2;
        bufp[5] = i1 - i2;
        r1 = bufp[2];
        i1 = bufp[3];
        r2 = -(bufp[7]);
        i2 = bufp[6];
        bufp[2] = r1 + r2;
        bufp[3] = i1 + i2;
        bufp[6] = r1 - r2;
        bufp[7] = i1 - i2;
        bufp = bufp + 8;
      }
      while (bufp < endp);
    }
    if (n < 8)
      return;
    for (size_t m = 8; m < n; m <<= 1)
    {
      FFTSineGen  s(invFFTLen ? -(long(m)) : long(m));
      size_t  i = 0;
      do
      {
        double  re, im;
        s.sample(re, im);
        bufp = buf + i;
        do
        {
          double  r1 = bufp[0];
          double  i1 = bufp[1];
          double  r2 = bufp[m];
          double  i2 = bufp[m + 1];
          double  tmp = (r2 * re) - (i2 * im);
          i2 = (r2 * im) + (i2 * re);
          r2 = tmp;
          bufp[0] = r1 + r2;
          bufp[1] = i1 + i2;
          bufp[m] = r1 - r2;
          bufp[m + 1] = i1 - i2;
          bufp = bufp + (m << 1);
        }
        while (bufp < endp);
        i = i + 2;
      }
      while (i < m);
    }
  }
  FFTSineGen  s(invFFTLen ? -(long(n)) : long(n));
  double  *bufp = buf;
  double  *endp = buf + n;
  do
  {
    double  r1 = bufp[0];
    double  i1 = bufp[1];
    double  r2 = bufp[n];
    double  i2 = bufp[n + 1];
    double  re, im;
    s.sample(re, im);
    double  tmp = (r2 * re) - (i2 * im);
    i2 = (r2 * im) + (i2 * re);
    r2 = tmp;
    bufp[0] = r1 + r2;
    bufp[1] = i1 + i2;
    bufp[n] = r1 - r2;
    bufp[n + 1] = i1 - i2;
    bufp = bufp + 2;
  }
  while (bufp < endp);
}
#else   // USE_OOURA_FFT
// use Takuya Ooura's FFT library from http://www.kurims.kyoto-u.ac.jp/~ooura/
#  ifdef NO_FFT_TABLES
#    include "fftsg_h.c"
#  else
#    include "fftsg.c"

static std::vector< int >     fftTable_ip(2, 0);
static std::vector< double >  fftTable_w;

static void resizeFFTTables(size_t n)
{
  int     log2N = 1;
  while ((size_t(1) << log2N) < n)
    log2N++;
  fftTable_ip.resize((size_t(1) << (log2N >> 1)) + 2);
  fftTable_w.resize(n >> 1);
}
#  endif
#endif

// 'buf' should contain n * 2 floats (interleaved real/imaginary data)

static void complexFFT(double *buf, size_t n, bool isInverse)
{
  if (!buf)
    throw std::runtime_error(std::string("internal error: NULL FFT buffer"));
  if (n == 1)
    return;
  if (n < 1 || n > 16777216 || (n & (n - 1)) != 0)
    throw std::runtime_error(std::string("internal error: invalid FFT size"));
#ifndef USE_OOURA_FFT
  // pack data in reverse bit order
  {
    size_t  j = n;
    for (size_t i = 2; j > 1; i += 2)
    {
      if (j > i)
      {
        double  tmp = buf[i];
        buf[i] = buf[j];
        buf[j] = tmp;
        tmp = buf[i + 1];
        buf[i + 1] = buf[j + 1];
        buf[j + 1] = tmp;
      }
      size_t  k = n;
      while (j & k)
      {
        j -= k;
        k >>= 1;
      }
      j += k;
    }
  }
  // calculate FFT
  recursiveFFT(buf, n, (isInverse ? int(n) : 0));
#else   // USE_OOURA_FFT
#  ifdef NO_FFT_TABLES
  cdft(int(n << 1), (isInverse ? -1 : 1), buf);
#  else
  if ((n >> 1) > fftTable_w.size())
    resizeFFTTables(n);
  cdft(int(n << 1), (isInverse ? -1 : 1), buf,
       &(fftTable_ip.front()), &(fftTable_w.front()));
#  endif
  if (isInverse)
  {
    double  scale = 1.0 / double(int(n));
    for (size_t i = 0; i < (n << 1); i += 2)
    {
      buf[i] *= scale;
      buf[i + 1] *= scale;
    }
  }
#endif
}

static void realFFT(double *buf, size_t n, bool isInverse)
{
  if (!buf)
    throw std::runtime_error(std::string("internal error: NULL FFT buffer"));
  if (n < 2 || n > 33554432 || (n & (n - 1)) != 0)
    throw std::runtime_error(std::string("internal error: invalid FFT size"));
#ifndef USE_OOURA_FFT
  size_t  nd2 = n >> 1;
  if (isInverse)
  {
    FFTSineGen  s(-(long(n)));
    // DC: re = 1, im = 0
    // Fs / 2 (stored at buf[1]): re = -1, im = 0
    {
      double  re, im;
      s.sample(re, im);
      // even samples (0, 2, ...)
      double  r1 = buf[0];
      // odd samples (1, 3, ...)
      double  r2 = buf[1];
      buf[0] = (r1 + r2) * 0.5;
      buf[1] = (r1 - r2) * 0.5;
    }
    for (size_t i = 2; i < nd2; i += 2)
    {
      size_t  j = n - i;
      // lower half of spectrum
      double  rl = buf[i];
      double  il = buf[i + 1];
      // upper half of spectrum
      double  rh = buf[j];
      double  ih = buf[j + 1];
      // r1,i1 = even samples (0, 2, ...)
      double  r1 = rl + rh;
      double  i1 = il - ih;
      // delay odd samples by -1 sample
      double  re, im;
      s.sample(re, im);
      // r2,i2 = odd samples (1, 3, ...)
      double  r2 = ((rl * re) - (il * im)) + ((rh * -re) - (ih * im));
      double  i2 = ((rl * im) + (il * re)) - ((rh * im) + (ih * -re));
      buf[i] = (r1 - i2) * 0.5;
      buf[i + 1] = (i1 + r2) * 0.5;
      buf[j] = (r1 + i2) * 0.5;
      buf[j + 1] = (-i1 + r2) * 0.5;
    }
    if (n >= 4)
    {
      // Fs / 4 (i = j = nd2): re = 0, im = 1
      buf[nd2 + 1] *= -1.0;
    }
  }
  complexFFT(buf, n >> 1, isInverse);
  if (!isInverse)
  {
    FFTSineGen  s((long) n);
    // DC: re = 1, im = 0
    // Fs / 2 (stored at buf[1]): re = -1, im = 0
    {
      double  re, im;
      s.sample(re, im);
      // even samples (0, 2, ...)
      double  r1 = buf[0];
      // odd samples (1, 3, ...)
      double  r2 = buf[1];
      buf[0] = r1 + r2;
      buf[1] = r1 - r2;
    }
    for (size_t i = 2; i < nd2; i += 2)
    {
      size_t  j = n - i;
      // r1,i1 = even samples (0, 2, ...)
      double  r1 = (buf[i] + buf[j]) * 0.5;
      double  i1 = (buf[i + 1] - buf[j + 1]) * 0.5;
      // r2,i2 = odd samples (1, 3, ...)
      double  r2 = (buf[i + 1] + buf[j + 1]) * 0.5;
      double  i2 = (buf[j] - buf[i]) * 0.5;
      // delay odd samples by 1 sample
      double  re, im;
      s.sample(re, im);
      // lower half of spectrum
      buf[i] = r1 + (r2 * re) - (i2 * im);
      buf[i + 1] = i1 + (r2 * im) + (i2 * re);
      // upper half of spectrum: similar to above, but with -i1, -i2, and -re
      buf[j] = r1 - (r2 * re) + (i2 * im);
      buf[j + 1] = -i1 + (r2 * im) + (i2 * re);
    }
    if (n >= 4)
    {
      // Fs / 4 (i = j = nd2): re = 0, im = -1
      buf[nd2 + 1] *= -1.0;
    }
  }
#else   // USE_OOURA_FFT
#  ifdef NO_FFT_TABLES
  rdft(int(n), (isInverse ? -1 : 1), buf);
#  else
  if ((n >> 1) > fftTable_w.size())
    resizeFFTTables(n);
  rdft(int(n), (isInverse ? -1 : 1), buf,
       &(fftTable_ip.front()), &(fftTable_w.front()));
#  endif
  if (isInverse)
  {
    double  scale = 2.0 / double(int(n));
    for (size_t i = 0; i < n; i += 2)
    {
      buf[i] *= scale;
      buf[i + 1] *= scale;
    }
  }
#endif
}

// simple version that does not perform any checks, and does not scale the
// results of inverse FFT

static inline void realFFT_Fast(double *buf, size_t n, bool isInverse)
{
#ifndef USE_OOURA_FFT
  realFFT(buf, n, isInverse);
#elif defined(NO_FFT_TABLES)
  rdft(int(n), (isInverse ? -1 : 1), buf);
#else
  rdft(int(n), (isInverse ? -1 : 1), buf,
       &(fftTable_ip.front()), &(fftTable_w.front()));
#endif
}

// ============================================================================
// |                              AUDIO FILTERS                               |
// ============================================================================

AudioFilter::AudioFilter(AudioFilter& prv_)
  : prv(&prv_)
{
  sampleRate = prv->getSampleRate();
  channels = prv->getChannelCount();
  outBuf.resize(size_t(channels), 0.0);
}

AudioFilter::AudioFilter(double sampleRate_, int channels_)
  : prv((AudioFilter *) 0),
    sampleRate(sampleRate_),
    channels(channels_),
    outBuf(size_t(channels_), 0.0)
{
}

AudioFilter::AudioFilter()
  : prv((AudioFilter *) 0),
    sampleRate(44100.0),
    channels(1),
    outBuf(1, 0.0)
{
}

AudioFilter::~AudioFilter()
{
}

// ----------------------------------------------------------------------------

void AudioFilter_FileInput::fillBuffer()
{
  sf_count_t  framesLeft = size_t(int(fileBuf.size()) / channels);
  sf_count_t  fileReadPos = filePos;
  double      *bufwp = &(fileBuf.front());
  while (framesLeft > 0)
  {
    if (fileReadPos < 0 || fileReadPos >= fileLength)
    {
      if (!isLooped)
      {
        for (int i = 0; i < channels; i++)
        {
          bufwp[i] = 0.0;
        }
        bufwp = bufwp + channels;
        framesLeft--;
        fileReadPos++;
        continue;
      }
      if (fileReadPos < 0)
        fileReadPos += fileLength;
      else
        fileReadPos -= fileLength;
      if (fileReadPos < 0 || fileReadPos >= fileLength)
      {
        fileReadPos = fileReadPos % fileLength;
        if (fileReadPos < 0)
          fileReadPos += fileLength;
      }
    }
    sf_count_t  framesToRead = fileLength - fileReadPos;
    if (framesToRead > framesLeft)
      framesToRead = framesLeft;
    if (sf_seek(f, fileReadPos, SEEK_SET) != fileReadPos)
      throw std::runtime_error(std::string("error seeking input file"));
    sf_count_t  framesRead = sf_readf_double(f, bufwp, framesToRead);
    if (framesRead != framesToRead)
      throw std::runtime_error(std::string("error reading input file"));
    framesLeft -= framesRead;
    fileReadPos += framesRead;
    bufwp = bufwp + (int(framesRead) * channels);
  }
  bufPos = 0;
}

AudioFilter_FileInput::AudioFilter_FileInput(const char *fname)
  : AudioFilter()
{
  if (!fname || !fname[0])
    throw std::runtime_error(std::string("invalid sound file name"));
  SF_INFO sfinfo;
  std::memset(&sfinfo, 0, sizeof(SF_INFO));
  f = sf_open(fname, SFM_READ, &sfinfo);
  if (!f)
  {
    throw std::runtime_error(std::string("error opening input file \"") + fname
                             + "\": " + sf_strerror(f));
  }
  try
  {
    if (sfinfo.samplerate < 1000 || sfinfo.samplerate > 10000000)
    {
      throw std::runtime_error(std::string("invalid sample rate "
                                           "in input file ") + fname);
    }
    if (sfinfo.channels < 1 || sfinfo.channels > 16)
    {
      throw std::runtime_error(std::string("invalid number of channels "
                                           "in input file ") + fname);
    }
    if (sfinfo.frames < 1)
    {
      throw std::runtime_error(std::string("empty input file: ") + fname);
    }
    sampleRate = double(sfinfo.samplerate);
    channels = sfinfo.channels;
    fileLength = sfinfo.frames;
    fileBuf.resize(size_t(channels) * 16384);
    filePos = 0;
    bufPos = fileBuf.size();
    isLooped = false;
    outBuf.resize(size_t(channels), 0.0);
  }
  catch (...)
  {
    sf_close(f);
    throw;
  }
}

AudioFilter_FileInput::~AudioFilter_FileInput()
{
  sf_close(f);
}

bool AudioFilter_FileInput::process()
{
  if (bufPos >= fileBuf.size())
    fillBuffer();
  copyOutputBuffer(&(fileBuf.front()) + bufPos);
  filePos++;
  bufPos = bufPos + size_t(channels);
  return true;
}

void AudioFilter_FileInput::setPosition(sf_count_t pos, bool looped)
{
  filePos = pos;
  bufPos = fileBuf.size();
  isLooped = looped;
}

// ----------------------------------------------------------------------------

AudioFilter_Delay::AudioFilter_Delay(AudioFilter& prv_,
                                     long delayFrames_, size_t lengthLimit)
  : AudioFilter(prv_),
    delayFrames(delayFrames_),
    framesRemaining(lengthLimit)
{
}

AudioFilter_Delay::~AudioFilter_Delay()
{
}

bool AudioFilter_Delay::process()
{
  if (framesRemaining)
  {
    framesRemaining--;
    if (delayFrames == 0L)
    {
      if (prv->process())
      {
        copyOutputBuffer(prv->getOutputBuffer());
        return true;
      }
    }
    else if (delayFrames > 0L)
    {
      delayFrames--;
      clearOutputBuffer();
      return true;
    }
    else
    {
      long    i = delayFrames - 1L;
      delayFrames = 0L;
      while (prv->process())
      {
        if (++i == 0L)
        {
          copyOutputBuffer(prv->getOutputBuffer());
          return true;
        }
      }
    }
  }
  framesRemaining = 0;
  clearOutputBuffer();
  return false;
}

// ----------------------------------------------------------------------------

AudioFilter_Upsample::AudioFilter_Upsample(AudioFilter& prv_,
                                           int upsampleRatio)
  : AudioFilter(prv_),
    r(upsampleRatio),
    cnt(0),
    isEOF(false)
{
  if (r < 1 || r > 1000)
    throw std::runtime_error(std::string("invalid upsample ratio"));
  sampleRate = prv->getSampleRate() * double(r);
}

AudioFilter_Upsample::~AudioFilter_Upsample()
{
}

bool AudioFilter_Upsample::process()
{
  if (cnt > 0)
  {
    cnt--;
    clearOutputBuffer();
    return true;
  }
  if (isEOF)
  {
    clearOutputBuffer();
    return false;
  }
  if (!prv->process())
  {
    isEOF = true;
    clearOutputBuffer();
    return false;
  }
  cnt = r - 1;
  copyOutputBuffer(prv->getOutputBuffer());
  return true;
}

// ----------------------------------------------------------------------------

AudioFilter_Downsample::AudioFilter_Downsample(AudioFilter& prv_,
                                               int downsampleRatio)
  : AudioFilter(prv_),
    r(downsampleRatio),
    isEOF(false)
{
  if (r < 1 || r > 1000)
    throw std::runtime_error(std::string("invalid downsample ratio"));
  sampleRate = prv->getSampleRate() / double(r);
}

AudioFilter_Downsample::~AudioFilter_Downsample()
{
}

bool AudioFilter_Downsample::process()
{
  if (!isEOF)
  {
    if (prv->process())
    {
      copyOutputBuffer(prv->getOutputBuffer());
      for (int i = 1; i < r; i++)
      {
        if (!prv->process())
        {
          isEOF = true;
          break;
        }
      }
      return true;
    }
    isEOF = true;
  }
  clearOutputBuffer();
  return false;
}

// ----------------------------------------------------------------------------

static void minimumPhaseIR(std::vector< double >& ir)
{
  // pad IR length to the nearest power of two, and at least 2
  size_t  len = 2;
  while (len < ir.size())
    len = len << 1;
  size_t  lenM2 = len << 1;
  std::vector< double > phaseBuf(lenM2, 0.0);
  ir.resize(lenM2, 0.0);
  // calculate magnitude response
  realFFT(&(ir.front()), lenM2, false);
  ir[0] = std::fabs(ir[0]);
  ir[1] = std::fabs(ir[1]);
  ir[0] = (ir[0] >= 1.0e-50 ? ir[0] : 1.0e-50);         // limit magnitude
  ir[1] = (ir[1] >= 1.0e-50 ? ir[1] : 1.0e-50);         // to at least -1000 dB
  phaseBuf[0] = std::log(ir[0]);
  phaseBuf[lenM2 >> 1] = std::log(ir[1]);
  for (size_t i = 2; i < lenM2; i += 2)
  {
    double  re = ir[i];
    double  im = ir[i + 1];
    double  r2 = (re * re) + (im * im);
    r2 = (r2 >= 1.0e-100 ? r2 : 1.0e-100);
    double  r = std::sqrt(r2);
    ir[i] = r;
    double  ln_r = std::log(r);
    phaseBuf[i >> 1] = ln_r;
    phaseBuf[lenM2 - (i >> 1)] = ln_r;
  }
  // calculate the Hilbert transform of the logarithm of the magnitude response
  realFFT(&(phaseBuf.front()), lenM2, false);
  phaseBuf[0] = 0.0;
  phaseBuf[1] = 0.0;
  for (size_t i = 2; i < lenM2; i += 2)
  {
    // since the magnitude response is mirrored around Fs/2 and symmetrical,
    // the imaginary part is always zero here
    phaseBuf[i + 1] = phaseBuf[i];
    phaseBuf[i] = 0.0;
  }
  realFFT(&(phaseBuf.front()), lenM2, true);
  // convert magnitude and phase response to real/imaginary format
  for (size_t i = 2; i < lenM2; i += 2)
  {
    double  r = ir[i];
    double  ph = phaseBuf[i >> 1];
    ir[i] = r * std::cos(ph);
    ir[i + 1] = r * std::sin(ph);
  }
  // convert frequency response to impulse response
  realFFT(&(ir.front()), lenM2, true);
  for (size_t i = 0; i < len; i++)
    ir[i] += ir[len + i];
  ir.resize(len);
}

// f = cutoff frequency as a fraction (0 to 0.5) of the sample rate
//     a negative value will create a minimum phase response
// a = gain (1.0 = 0 dB)

static void createLowpassIR(std::vector< double >& ir, size_t irLen,
                            double f, double a, int winParam)
{
  if (irLen < 4 || irLen > 4194304 || (irLen & 1) != 0)
  {
    throw std::runtime_error(std::string("invalid lowpass filter IR length"));
  }
  bool    minPhase = false;
  if (f < 0.0)
  {
    minPhase = true;
    f = -f;
  }
  if (!(f >= 0.000001 && f <= 0.5))
    throw std::runtime_error(std::string("invalid lowpass filter frequency"));
  if (winParam < -240 || winParam > 0)
  {
    throw std::runtime_error(std::string("invalid lowpass filter "
                                         "window parameter"));
  }
  ir.resize(irLen, 0.0);
  size_t  irLenD2 = irLen >> 1;
  ir[0] = 0.0;
  ir[irLenD2] = f * 2.0 * a;
  a /= pi;
  SineGen   s(1.0 / f);
  (void) s.sample();
  if (winParam < 0)
  {
    GaussianWindowNormalized  w(long(irLen + 1),
                                std::pow(10.0, double(winParam) * 0.05));
    w.setPosition(1.0);
    for (size_t i = 1; i < irLenD2; i++)
    {
      double  y = w.sample() * a * s.sample() / double(int(i));
      ir[irLenD2 - i] = y;
      ir[irLenD2 + i] = y;
    }
  }
  else
  {
    CosineGen w(double(int(irLen)), pi);
    (void) w.sample();
    for (size_t i = 1; i < irLenD2; i++)
    {
      double  y = cosineToNuttallWindow(w.sample()) * a * s.sample()
                  / double(int(i));
      ir[irLenD2 - i] = y;
      ir[irLenD2 + i] = y;
    }
  }
  if (minPhase)
    minimumPhaseIR(ir);
}

// called by the constructors to initialize the filter

void AudioFilter_FIRFilter::initFilter(
    const std::vector< std::vector< double > >& impulseResponses)
{
  if (impulseResponses.size() != 1)
  {
    if (channels == 1)
    {
      channels = int(impulseResponses.size());
    }
    else if (impulseResponses.size() != size_t(channels) &&
             impulseResponses.size() != size_t(channels * channels))
    {
      throw std::runtime_error(std::string("invalid impulse response "
                                           "channel count"));
    }
  }
  irLength = impulseResponses[0].size();
  if (irLength < 2 || irLength > 8388608)
  {
    throw std::runtime_error(std::string("invalid impulse response length"));
  }
  for (size_t i = 1; i < impulseResponses.size(); i++)
  {
    if (impulseResponses[i].size() != irLength)
    {
      throw std::runtime_error(std::string("impulse response lengths do not "
                                           "match in multi-channel IR"));
    }
  }
  blockLength = 1;
  while (blockLength < irLength)
    blockLength = blockLength << 1;
  fftLength = blockLength << 1;
  fftBuf.resize(fftLength, 0.0);
  irFFTs.resize(impulseResponses.size());
  for (size_t i = 0; i < impulseResponses.size(); i++)
  {
    irFFTs[i].resize(fftLength, 0.0);
    for (size_t j = 0; j < irLength; j++)
    {
      irFFTs[i][j] = impulseResponses[i][j];
    }
    realFFT(&(irFFTs[i].front()), fftLength, false);
  }
  inBufs.resize(size_t(prv->getChannelCount()));
  for (int i = 0; i < prv->getChannelCount(); i++)
    inBufs[i].resize(blockLength, 0.0);
  outBufs.resize(size_t(channels));
  for (int i = 0; i < channels; i++)
    outBufs[i].resize(fftLength, 0.0);
  outBufPos = 0;
  framesRemaining = 0;
  isEOF = false;
}

AudioFilter_FIRFilter::AudioFilter_FIRFilter(
    AudioFilter& prv_,
    const std::vector< std::vector< double > >& impulseResponses)
  : AudioFilter(prv_)
{
  initFilter(impulseResponses);
}

AudioFilter_FIRFilter::AudioFilter_FIRFilter(AudioFilter& prv_,
                                             double f, int irLen, double a,
                                             int winParam, double *d)
  : AudioFilter(prv_)
{
  std::vector< std::vector< double > >  impulseResponses;
  impulseResponses.resize(1);
  if (irLen < 0)
    irLen = int(double(irLen) * sampleRate * -0.000001 + 0.5);
  irLen = irLen + (irLen & 1);
  createLowpassIR(impulseResponses[0], size_t(irLen), f, a, winParam);
  initFilter(impulseResponses);
  if (d)
  {
    // calculate delay (NOTE: FFT length is always at least 4)
    if (f < 0.0)
    {
      // minimum phase
      *d = -(std::atan2(irFFTs[0][3], irFFTs[0][2]))
           * double(long(irFFTs[0].size())) / twoPi;
    }
    else
    {
      // linear phase
      *d = double(irLen >> 1);
    }
  }
}

AudioFilter_FIRFilter::~AudioFilter_FIRFilter()
{
}

bool AudioFilter_FIRFilter::process()
{
  if (framesRemaining > 0)
  {
    framesRemaining--;
    int     i = 0;
    do
    {
      outBuf[i] = outBufs[i][outBufPos];
      outBufs[i][outBufPos] = 0.0;
    }
    while (++i < channels);
    outBufPos = (outBufPos + 1) & (fftLength - 1);
    return true;
  }
  if (isEOF)
  {
    clearOutputBuffer();
    return false;
  }
  const double  *bufrp = &(prv->getOutputBuffer().front());
  int     inChannels = prv->getChannelCount();
  size_t  framesReceived = 0;
  for ( ; framesReceived < blockLength; framesReceived++)
  {
    if (!prv->process())
    {
      isEOF = true;
      break;
    }
    int     i = 0;
    do
    {
      inBufs[i][framesReceived] = bufrp[i];
    }
    while (++i < inChannels);
  }
  if (framesReceived > 0)
  {
    size_t  fftCnt = (inBufs.size() > irFFTs.size() ?
                      inBufs.size() : irFFTs.size());
    for (size_t i = 0; i < fftCnt; i++)
    {
      size_t  inChn = i;
      size_t  outChn = i;
      size_t  irChn = i;
      if (irFFTs.size() == 1)
      {
        // convolve all channels with the same (mono) impulse response
        irChn = 0;
      }
      else if (irFFTs.size() > size_t(inChannels))
      {
        // there are separate impulse responses for every possible
        // combination of input and output channel (IR channels =
        // audio channels ^ 2); for example, with stereo audio, the
        // IR channels are L->L, L->R, R->L, and R->R
        // FIXME: the implementation of this case is not optimal,
        // because the FFT of the same input channel is unnecessarily
        // calculated multiple times
        inChn = i / size_t(channels);
        outChn = i % size_t(channels);
      }
      // copy the input data to the FFT buffer
      for (size_t j = 0; j < framesReceived; j++)
        fftBuf[j] = inBufs[inChn][j];
      // pad to FFT length with zeros
      for (size_t j = framesReceived; j < fftLength; j++)
        fftBuf[j] = 0.0;
      // compute the FFT
      realFFT_Fast(&(fftBuf.front()), fftLength, false);
      // multiply the result with the FFT of the IR
      // to perform the convolution
      // there is no imaginary part for DC and the Nyquist frequency
      fftBuf[0] = fftBuf[0] * irFFTs[irChn][0];
      fftBuf[1] = fftBuf[1] * irFFTs[irChn][1];
      for (size_t j = 2; j < fftLength; j += 2)
      {
        double  re = fftBuf[j];
        double  im = fftBuf[j + 1];
        double  ir_re = irFFTs[irChn][j];
        double  ir_im = irFFTs[irChn][j + 1];
        fftBuf[j] = (re * ir_re) - (im * ir_im);
        fftBuf[j + 1] = (re * ir_im) + (im * ir_re);
      }
      // compute inverse FFT to get the convolved audio data
      realFFT_Fast(&(fftBuf.front()), fftLength, true);
      // add the result to the mix buffer
#ifdef USE_OOURA_FFT
      double  scale = 2.0 / double(int(fftLength));
#endif
      for (size_t j = 0; j < fftLength; j++)
      {
        double& y = outBufs[outChn][(outBufPos + j) & (fftLength - 1)];
#ifndef USE_OOURA_FFT
        y = y + fftBuf[j];
#else
        y = y + (fftBuf[j] * scale);
#endif
      }
    }
  }
  framesRemaining = framesReceived + (isEOF ? (irLength - 1) : 0) - 1;
  for (int i = 0; i < channels; i++)
  {
    outBuf[i] = outBufs[i][outBufPos];
    outBufs[i][outBufPos] = 0.0;
  }
  outBufPos = (outBufPos + 1) & (fftLength - 1);
  return true;
}

// ----------------------------------------------------------------------------

// called by the constructor to initialize the filter

void AudioFilter_OctaveResample::initFilter(
    const std::vector< double >& impulseResponse, int nOct)
{
  irLength = impulseResponse.size();
  if (irLength < 2 || irLength > 8388608)
  {
    throw std::runtime_error(std::string("invalid impulse response length"));
  }
  size_t  blockLength = size_t(1 << std::abs(nOct));
  while (blockLength < irLength)
    blockLength = blockLength << 1;
  blockLength_In = (nOct > 0 ? (blockLength >> nOct) : blockLength);
  blockLength_Out = (nOct < 0 ? (blockLength >> (-nOct)) : blockLength);
  size_t  fftLength = blockLength << 1;
  fftLength_In = blockLength_In << 1;
  fftLength_Out = blockLength_Out << 1;
  fftBuf.resize(fftLength, 0.0);
  irFFT.resize(fftLength, 0.0);
  for (size_t i = 0; i < irLength; i++)
    irFFT[i] = impulseResponse[i];
  realFFT(&(irFFT.front()), fftLength, false);
  inBufs.resize(size_t(channels));
  for (int i = 0; i < channels; i++)
    inBufs[i].resize(blockLength_In, 0.0);
  outBufs.resize(size_t(channels));
  for (int i = 0; i < channels; i++)
    outBufs[i].resize(fftLength_Out, 0.0);
  outBufPos = 0;
  framesRemaining = 0;
  isEOF = false;
}

AudioFilter_OctaveResample::AudioFilter_OctaveResample(
    AudioFilter& prv_, int nOct, double f, int irLen, int winParam, double *d)
  : AudioFilter(prv_)
{
  if (nOct < -8 || nOct > 8)
    throw std::runtime_error(std::string("invalid resample ratio"));
  if (nOct > 0)
    sampleRate *= double(1 << nOct);
  double  r = 1.0;
  if (nOct != 0)
  {
    r = 1.0 / double(1 << std::abs(nOct));
    f *= r;
  }
  std::vector< double > impulseResponse;
  if (irLen < 0)
    irLen = int(double(irLen) * sampleRate * -0.000001 + 0.5);
  irLen = irLen + (irLen & 1);
  createLowpassIR(impulseResponse, size_t(irLen), f,
                  double(nOct > 0 ? (1 << nOct) : 1), winParam);
  initFilter(impulseResponse, nOct);
  if (nOct < 0)
    sampleRate *= r;
  if (d)
  {
    // calculate delay (NOTE: FFT length is always at least 4)
    if (f < 0.0)
    {
      // minimum phase
      *d = -(std::atan2(irFFT[3], irFFT[2]))
           * double(long(irFFT.size())) / twoPi;
    }
    else
    {
      // linear phase
      *d = double(irLen >> 1);
    }
    if (nOct < 0)
      *d *= r;
  }
}

AudioFilter_OctaveResample::~AudioFilter_OctaveResample()
{
}

bool AudioFilter_OctaveResample::process()
{
  if (framesRemaining > 0)
  {
    framesRemaining--;
    int     i = 0;
    do
    {
      outBuf[i] = outBufs[i][outBufPos];
      outBufs[i][outBufPos] = 0.0;
    }
    while (++i < channels);
    outBufPos = (outBufPos + 1) & (fftLength_Out - 1);
    return true;
  }
  if (isEOF)
  {
    clearOutputBuffer();
    return false;
  }
  const double  *bufrp = &(prv->getOutputBuffer().front());
  size_t  framesReceived = 0;
  for ( ; framesReceived < blockLength_In; framesReceived++)
  {
    if (!prv->process())
    {
      isEOF = true;
      break;
    }
    int     i = 0;
    do
    {
      inBufs[i][framesReceived] = bufrp[i];
    }
    while (++i < channels);
  }
  if (framesReceived > 0)
  {
    for (size_t i = 0; i < size_t(channels); i++)
    {
      // copy the input data to the FFT buffer
      for (size_t j = 0; j < framesReceived; j++)
        fftBuf[j] = inBufs[i][j];
      // pad to FFT length with zeros
      for (size_t j = framesReceived; j < fftLength_In; j++)
        fftBuf[j] = 0.0;
      // compute the FFT
      realFFT_Fast(&(fftBuf.front()), fftLength_In, false);
      // upsample input if needed
      for (size_t k = fftLength_In; k < fftLength_Out; k <<= 1)
      {
        fftBuf[k] = fftBuf[1];
        fftBuf[k + 1] = 0.0;
        fftBuf[1] = fftBuf[0];
        for (size_t j = 2; j < k; j += 2)
        {
          fftBuf[k + j] = fftBuf[k - j];
          fftBuf[k + 1 + j] = -(fftBuf[k + 1 - j]);
        }
      }
      // multiply the result with the FFT of the IR
      // to perform the convolution
      // there is no imaginary part for DC and the Nyquist frequency
      fftBuf[0] = fftBuf[0] * irFFT[0];
      fftBuf[1] = fftBuf[1] * irFFT[1];
      for (size_t j = 2; j < fftBuf.size(); j += 2)
      {
        double  re = fftBuf[j];
        double  im = fftBuf[j + 1];
        double  ir_re = irFFT[j];
        double  ir_im = irFFT[j + 1];
        fftBuf[j] = (re * ir_re) - (im * ir_im);
        fftBuf[j + 1] = (re * ir_im) + (im * ir_re);
      }
      // downsample output if needed
      for (size_t k = fftLength_In; k > fftLength_Out; k >>= 1)
      {
        fftBuf[0] = (fftBuf[0] + fftBuf[1]) * 0.5;
        fftBuf[1] = fftBuf[k >> 1];
        for (size_t j = 2; j < (k >> 1); j += 2)
        {
          fftBuf[j] = (fftBuf[j] + fftBuf[k - j]) * 0.5;
          fftBuf[j + 1] = (fftBuf[j + 1] - fftBuf[k + 1 - j]) * 0.5;
        }
      }
      // compute inverse FFT to get the convolved audio data
      realFFT_Fast(&(fftBuf.front()), fftLength_Out, true);
      // add the result to the mix buffer
#ifdef USE_OOURA_FFT
      double  scale = 2.0 / double(int(fftLength_Out));
#endif
      for (size_t j = 0; j < fftLength_Out; j++)
      {
        double& y = outBufs[i][(outBufPos + j) & (fftLength_Out - 1)];
#ifndef USE_OOURA_FFT
        y = y + fftBuf[j];
#else
        y = y + (fftBuf[j] * scale);
#endif
      }
    }
  }
  if (!isEOF)
  {
    framesRemaining = blockLength_Out;          // not EOF yet
  }
  else
  {
    framesRemaining = framesReceived;
    if (blockLength_Out > blockLength_In)
      framesRemaining *= (blockLength_Out / blockLength_In);    // upsample
    framesRemaining += (irLength - 1);
    if (blockLength_Out < blockLength_In)
    {
      size_t  r = blockLength_In / blockLength_Out;             // downsample
      framesRemaining = (framesRemaining + r - 1) / r;
    }
  }
  framesRemaining--;
  for (int i = 0; i < channels; i++)
  {
    outBuf[i] = outBufs[i][outBufPos];
    outBufs[i][outBufPos] = 0.0;
  }
  outBufPos = (outBufPos + 1) & (fftLength_Out - 1);
  return true;
}

// ----------------------------------------------------------------------------

AudioFilter_Interpolate::AudioFilter_Interpolate(
    AudioFilter& prv_,
    double downsampleRatio, double startPos, int winSize)
  : AudioFilter(prv_),
    w(winSize + 1,
      1.0 / double(1L << (((winSize > 8 ? (winSize < 42 ? winSize : 42) : 8)
                           * 10 + 8) >> 4)))
{
  if (!(downsampleRatio >= 0.01 && downsampleRatio <= 100.0))
    throw std::runtime_error(std::string("invalid resample ratio"));
  if (!(startPos >= -1000000000.0 && startPos <= 1000000000.0))
    throw std::runtime_error(std::string("invalid resample start position"));
  if (winSize < 8 || winSize > 2048 || (winSize & 3) != 0)
    throw std::runtime_error(std::string("invalid resample window size"));
  framesRemaining = 0xFFFFFFFFUL;
  isEOF = false;
  windowSize = winSize;
  inBuf.resize(size_t(windowSize * channels), 0.0);
  writePos = 0;
  readOffs = (long double) (windowSize >> 1) + (long double) startPos;
  r = downsampleRatio;
  sampleRate = prv->getSampleRate() / r;
}

AudioFilter_Interpolate::~AudioFilter_Interpolate()
{
}

bool AudioFilter_Interpolate::process()
{
  clearOutputBuffer();
  if (!framesRemaining)
    return false;
  while (readOffs >= 0.0L)
  {
    if (!isEOF)
    {
      if (!prv->process())
      {
        isEOF = true;
        long    tmp =
            long((double(windowSize >> 1) - double(std::floor(readOffs))) / r
                 + 0.5);
        if (!(tmp > 0L))
        {
          framesRemaining = 0;
          return false;
        }
        framesRemaining = size_t(tmp);
      }
    }
    double  *bufwp = &(inBuf.front()) + writePos;
    int     i = 0;
    if (isEOF)
    {
      do
      {
        bufwp[i] = 0.0;
      }
      while (++i < channels);
    }
    else
    {
      const double  *bufrp = &(prv->getOutputBuffer().front());
      do
      {
        bufwp[i] = bufrp[i];
      }
      while (++i < channels);
    }
    writePos = ((writePos + channels) < int(inBuf.size()) ?
                (writePos + channels) : 0);
    readOffs -= 1.0L;
  }
  framesRemaining--;
  double  winPos_f = -(double(readOffs));
  int     x = -(windowSize >> 1);
  if (winPos_f > 0.5)
  {
    winPos_f -= 1.0;
    x++;
  }
  readOffs += (long double) r;
  int     readPos = writePos;
  if (std::fabs(winPos_f) < 3.0e-10)
  {
    readPos = readPos - (x * channels);
    readPos = (readPos < int(inBuf.size()) ?
               readPos : (readPos - int(inBuf.size())));
    copyOutputBuffer(&(inBuf.front()) + readPos);
    return true;
  }
  double  sinXDPi = std::sin(std::fabs(winPos_f) * pi) * double(1.0L / piL);
  w.setPosition((long double) x + (long double) winPos_f);
  const double  *bufrp = &(inBuf.front()) + readPos;
  double        *bufwp = &(outBuf.front());
  int     i = windowSize >> 1;
  switch (channels)
  {
  case 1:
    do
    {
#ifdef USE_SIMD
      Vector2Double w_ = { w.sample(), w.sample() };
      Vector2Double x_ = { double(x) + winPos_f, double(x + 1) + winPos_f };
      w_ = w_ / x_;
      x += 2;
      *bufwp += (*bufrp * ((double *) &w_)[0]);
      bufrp++;
      bufrp = (bufrp < &(*(inBuf.end())) ? bufrp : &(inBuf.front()));
      *bufwp -= (*bufrp * ((double *) &w_)[1]);
#else
      double  a = w.sample() / (double(x) + winPos_f);
      x++;
      *bufwp += (*bufrp * a);
      bufrp++;
      bufrp = (bufrp < &(*(inBuf.end())) ? bufrp : &(inBuf.front()));
      a = w.sample() / (double(x) + winPos_f);
      x++;
      *bufwp -= (*bufrp * a);
#endif
      bufrp++;
      bufrp = (bufrp < &(*(inBuf.end())) ? bufrp : &(inBuf.front()));
    }
    while (--i);
    *bufwp *= sinXDPi;
    break;
  case 2:
    do
    {
#ifdef USE_SIMD
      Vector2Double w_ = { w.sample(), w.sample() };
      Vector2Double x_ = { double(x) + winPos_f, double(x + 1) + winPos_f };
      w_ = w_ / x_;
      x += 2;
      bufwp[0] += (bufrp[0] * ((double *) &w_)[0]);
      bufwp[1] += (bufrp[1] * ((double *) &w_)[0]);
      bufrp = bufrp + 2;
      bufrp = (bufrp < &(*(inBuf.end())) ? bufrp : &(inBuf.front()));
      bufwp[0] -= (bufrp[0] * ((double *) &w_)[1]);
      bufwp[1] -= (bufrp[1] * ((double *) &w_)[1]);
#else
      double  a = w.sample() / (double(x) + winPos_f);
      x++;
      bufwp[0] += (bufrp[0] * a);
      bufwp[1] += (bufrp[1] * a);
      bufrp = bufrp + 2;
      bufrp = (bufrp < &(*(inBuf.end())) ? bufrp : &(inBuf.front()));
      a = w.sample() / (double(x) + winPos_f);
      x++;
      bufwp[0] -= (bufrp[0] * a);
      bufwp[1] -= (bufrp[1] * a);
#endif
      bufrp = bufrp + 2;
      bufrp = (bufrp < &(*(inBuf.end())) ? bufrp : &(inBuf.front()));
    }
    while (--i);
    bufwp[0] *= sinXDPi;
    bufwp[1] *= sinXDPi;
    break;
  default:
    do
    {
#ifdef USE_SIMD
      Vector2Double w_ = { w.sample(), w.sample() };
      Vector2Double x_ = { double(x) + winPos_f, double(x + 1) + winPos_f };
      w_ = w_ / x_;
      x += 2;
      int     j = 0;
      do
      {
        bufwp[j] += (bufrp[j] * ((double *) &w_)[0]);
      }
      while (++j < channels);
      bufrp = bufrp + channels;
      bufrp = (bufrp < &(*(inBuf.end())) ? bufrp : &(inBuf.front()));
      j = 0;
      do
      {
        bufwp[j] -= (bufrp[j] * ((double *) &w_)[1]);
      }
      while (++j < channels);
#else
      double  a = w.sample() / (double(x) + winPos_f);
      x++;
      int     j = 0;
      do
      {
        bufwp[j] += (bufrp[j] * a);
      }
      while (++j < channels);
      bufrp = bufrp + channels;
      bufrp = (bufrp < &(*(inBuf.end())) ? bufrp : &(inBuf.front()));
      a = w.sample() / (double(x) + winPos_f);
      x++;
      j = 0;
      do
      {
        bufwp[j] -= (bufrp[j] * a);
      }
      while (++j < channels);
#endif
      bufrp = bufrp + channels;
      bufrp = (bufrp < &(*(inBuf.end())) ? bufrp : &(inBuf.front()));
    }
    while (--i);
    for (int j = 0; j < channels; j++)
      bufwp[j] *= sinXDPi;
    break;
  }
  return true;
}

// ----------------------------------------------------------------------------

AudioFilter_InterpolateFast::AudioFilter_InterpolateFast(
    AudioFilter& prv_, int n, int d, double startPos, int winSize)
  : AudioFilter(prv_)
{
  if (n < 1 || n > 4096)
    throw std::runtime_error(std::string("invalid number of window tables"));
  if ((n * 100) < d || n > (d * 100))
    throw std::runtime_error(std::string("invalid resample ratio"));
  if (!(startPos >= -1000000000.0 && startPos <= 1000000000.0))
    throw std::runtime_error(std::string("invalid resample start position"));
  if (winSize < 8 || winSize > 2048 || (winSize & 3) != 0)
    throw std::runtime_error(std::string("invalid resample window size"));
  framesRemaining = 0xFFFFFFFFUL;
  isEOF = false;
  windowSize = winSize;
  inBuf.resize(size_t(windowSize * channels), 0.0);
  writePos = 0;
  windowNum = 0;
  long double readOffs_ =
      (long double) (windowSize >> 1) + (long double) startPos;
  readOffs = long(std::floor(readOffs_));
  long double f = readOffs_ - (long double) (readOffs + 1L);
  long double r = (long double) d / (long double) n;
  upsampleRatio = double(1.0L / r);
  sampleRate = prv->getSampleRate() * double(n) / double(d);
  windowTable.resize(size_t(windowSize * n));
  readPosIncTable.resize(size_t(n));
  double  wMin =
      1.0 / double(1L << (((winSize > 8 ? (winSize < 42 ? winSize : 42) : 8)
                           * 10 + 8) >> 4));
  GaussianWindowNormalized  w(winSize + 1, wMin);
  for (int i = 0; i < n; i++)
  {
    double  winPos_f = -(double(f));
    int     x = -(windowSize >> 1);
    if (winPos_f > 0.5)
    {
      winPos_f -= 1.0;
      x++;
    }
    f += r;
    unsigned char readPosInc = 0;
    while (f >= 0.0L)
    {
      f -= 1.0L;
      readPosInc++;
    }
    readPosIncTable[i] = readPosInc;
    d -= int(readPosInc);
    double  sinXDPi = std::sin(std::fabs(winPos_f) * pi) * double(1.0L / piL);
    w.setPosition((long double) x + (long double) winPos_f);
    double  *wtp = &(windowTable.front()) + (i * windowSize);
    for (int j = 0; j < windowSize; )
    {
      if (x < -1 || x > 0)
      {
#ifdef USE_SIMD
        Vector2Double w_ = { w.sample() * sinXDPi, w.sample() * -sinXDPi };
        Vector2Double x_ = { double(x) + winPos_f, double(x + 1) + winPos_f };
        w_ = w_ / x_;
        wtp[j] = ((double *) &w_)[0];
        wtp[j + 1] = ((double *) &w_)[1];
#else
        wtp[j] = w.sample() * sinXDPi / (double(x) + winPos_f);
        wtp[j + 1] = w.sample() * -sinXDPi / (double(x + 1) + winPos_f);
#endif
        x += 2;
        j += 2;
      }
      else
      {
        do
        {
          if (x == 0 && sinXDPi < 1.0e-7)
          {
            (void) w.sample();
            wtp[j] = 1.0;
          }
          else
          {
            wtp[j] = w.sample() * ((j & 1) ? -sinXDPi : sinXDPi)
                     / (double(x) + winPos_f);
          }
          x++;
          j++;
        }
        while (j & 1);
      }
    }
  }
  if (d != 0)
    readPosIncTable[n - 1] = (unsigned char) (int(readPosIncTable[n - 1]) + d);
}

AudioFilter_InterpolateFast::~AudioFilter_InterpolateFast()
{
}

bool AudioFilter_InterpolateFast::process()
{
  clearOutputBuffer();
  if (!framesRemaining)
    return false;
  while (readOffs >= 0L)
  {
    if (!isEOF)
    {
      if (!prv->process())
      {
        isEOF = true;
        long    tmp =
            long(double(long(windowSize >> 1) - readOffs) * upsampleRatio
                 + 0.5);
        if (!(tmp > 0L))
        {
          framesRemaining = 0;
          return false;
        }
        framesRemaining = size_t(tmp);
      }
    }
    double  *bufwp = &(inBuf.front()) + writePos;
    int     i = 0;
    if (isEOF)
    {
      do
      {
        bufwp[i] = 0.0;
      }
      while (++i < channels);
    }
    else
    {
      const double  *bufrp = &(prv->getOutputBuffer().front());
      do
      {
        bufwp[i] = bufrp[i];
      }
      while (++i < channels);
    }
    writePos = ((writePos + channels) < int(inBuf.size()) ?
                (writePos + channels) : 0);
    readOffs--;
  }
  framesRemaining--;
  readOffs += long(readPosIncTable[windowNum]);
  const double  *wtp = &(windowTable.front()) + (windowNum * windowSize);
  windowNum = ((windowNum + 1) < int(readPosIncTable.size()) ?
               (windowNum + 1) : 0);
  const double  *bufrp = &(inBuf.front()) + writePos;   // readPos = writePos
  double        *bufwp = &(outBuf.front());
  int     i = windowSize >> 1;
  switch (channels)
  {
  case 1:
    do
    {
      *bufwp += *bufrp * wtp[0];
      bufrp++;
      bufrp = (bufrp < &(*(inBuf.end())) ? bufrp : &(inBuf.front()));
      *bufwp += *bufrp * wtp[1];
      bufrp++;
      wtp = wtp + 2;
      bufrp = (bufrp < &(*(inBuf.end())) ? bufrp : &(inBuf.front()));
    }
    while (--i);
    break;
  case 2:
    do
    {
      double  w = wtp[0];
      bufwp[0] += bufrp[0] * w;
      bufwp[1] += bufrp[1] * w;
      bufrp = bufrp + 2;
      w = wtp[1];
      bufrp = (bufrp < &(*(inBuf.end())) ? bufrp : &(inBuf.front()));
      bufwp[0] += bufrp[0] * w;
      bufwp[1] += bufrp[1] * w;
      bufrp = bufrp + 2;
      wtp = wtp + 2;
      bufrp = (bufrp < &(*(inBuf.end())) ? bufrp : &(inBuf.front()));
    }
    while (--i);
    break;
  default:
    do
    {
      double  w = wtp[0];
      int     j = 0;
      do
      {
        bufwp[j] += bufrp[j] * w;
      }
      while (++j < channels);
      bufrp = bufrp + channels;
      w = wtp[1];
      bufrp = (bufrp < &(*(inBuf.end())) ? bufrp : &(inBuf.front()));
      j = 0;
      do
      {
        bufwp[j] += bufrp[j] * w;
      }
      while (++j < channels);
      bufrp = bufrp + channels;
      wtp = wtp + 2;
      bufrp = (bufrp < &(*(inBuf.end())) ? bufrp : &(inBuf.front()));
    }
    while (--i);
    break;
  }
  return true;
}

AudioFilter * AudioFilter_InterpolateFast::createInterpolateFilter(
    AudioFilter& prv_, double downsampleRatio, double startPos, int winSize)
{
  if (!(downsampleRatio >= 0.01 && downsampleRatio <= 100.0))
    throw std::runtime_error(std::string("invalid resample ratio"));
  if (winSize < 8 || winSize > 2048 || (winSize & 3) != 0)
    throw std::runtime_error(std::string("invalid resample window size"));
  size_t  windowTableSize =
      size_t(winSize) * sizeof(double) + sizeof(unsigned char);
  size_t  totalWindowTableSize = windowTableSize;
  // limit total table size to 192 kilobytes
  for (int n = 1; totalWindowTableSize < 0x30000; n++)
  {
    double  p = double(n) * downsampleRatio;
    int     d = int(p + 0.5);
    // limit pitch error to +/- 5.0e-15
    if (p > (double(d) * 0.999999999999995) &&
        p < (double(d) * 1.000000000000005))
    {
      return new AudioFilter_InterpolateFast(prv_, n, d, startPos, winSize);
    }
    totalWindowTableSize += windowTableSize;
  }
  return new AudioFilter_Interpolate(prv_, downsampleRatio, startPos, winSize);
}

// ----------------------------------------------------------------------------

AudioFilter_ResampledFileInput::AudioFilter_ResampledFileInput(
    const char *fname, double outputSampleRate, double isrAdj, double skipTime,
    bool isLooped, int interpLen, int lpFilterLen, double lpFilterFreq,
    int lpFilterWindow)
  : AudioFilter(),
    filters(4, (AudioFilter *) 0),
    framesRemaining(size_t(-1))
{
  if (!(isrAdj >= 0.1 && isrAdj <= 10.0))
  {
    throw std::runtime_error(std::string("invalid resample "
                                         "sample rate adjust factor"));
  }
  AudioFilter*& lastFilter = filters[filters.size() - 1];
  try
  {
    AudioFilter_FileInput&  sf = *(new AudioFilter_FileInput(fname));
    filters[0] = &sf;
    sampleRate = outputSampleRate;
    if (sampleRate < 0.0)
      sampleRate = -sampleRate * sf.getSampleRate();
    if (!(sampleRate >= 1000.0 && sampleRate <= 1000000.0))
    {
      throw std::runtime_error(std::string("invalid resample "
                                           "output sample rate"));
    }
    double  isr = sf.getSampleRate() * isrAdj;
    double  r = sampleRate / isr;
    if (!(r >= 0.01 && r <= 100.0))
      throw std::runtime_error(std::string("invalid resample ratio"));
    channels = sf.getChannelCount();
    outBuf.resize(size_t(channels), 0.0);
    double  skipFrames = sf.getSampleRate() * skipTime;
    double  skipFrames_i = 0.0;
    double  skipFrames_f = std::modf(skipFrames, &skipFrames_i);
    if (skipFrames_f < 0.0)
    {
      skipFrames_i -= 1.0;
      skipFrames_f += 1.0;
    }
    if (skipFrames_f < 0.0000002 || skipFrames_f > 0.9999998)
    {
      // integer sample delay
      skipFrames_i += double(int(skipFrames_f > 0.5));
      skipFrames = skipFrames_i;
      skipFrames_f = 0.0;
    }
    sf.setPosition(sf_count_t(skipFrames_i), isLooped);
    sf.setSampleRate(isr);
    int     r_i = 0;            // integer downsample or upsample ratio
    {
      double  tmp = (r < 1.0 ? (1.0 / r) : r);
      r_i = int(tmp + 0.5);
      if (tmp > (double(r_i) * 0.999999999999995) &&
          tmp < (double(r_i) * 1.000000000000005))
      {
        r = (r < 1.0 ? (1.0 / double(r_i)) : double(r_i));
      }
      else
      {
        r_i = 0;                // not integer resample ratio
      }
    }
    if (!isLooped)      // unlimited length in the case of looped playback
    {
      long long tmp =
          (long long) ((double(sf.getFileLength()) - skipFrames) * r + 0.5);
      framesRemaining =
          (tmp >= 0LL ?
           ((long long) size_t(tmp) == tmp ? size_t(tmp) : size_t(-1))
           : size_t(0));
    }
    if (r_i > 0 && skipFrames_f == 0.0)
    {
      // no interpolation is needed
      if (r_i == 1)
      {
        // no sample rate conversion
        lastFilter = &sf;
        filters[0] = (AudioFilter *) 0;
        return;
      }
      int     offs = 0;
      if (isLooped || skipFrames_i > 0.0)
      {
        sf.setPosition(sf_count_t(skipFrames_i) - 128, isLooped);
        offs = 128;
      }
      if (r < 1.0)
      {
        // downsampling by integer ratio
        double  gd = 0.0;
        if ((r_i & (r_i - 1)) || lpFilterFreq < 0.0)
        {
          filters[1] = new AudioFilter_FIRFilter(
              sf, lpFilterFreq * r, lpFilterLen, 1.0, lpFilterWindow, &gd);
          filters[2] = new AudioFilter_Delay(
              *(filters[1]), long(-(gd + 0.5)) - offs);
          lastFilter = new AudioFilter_Downsample(*(filters[2]), r_i);
        }
        else            // downsampling by power of two ratio,
        {               // with linear phase lowpass filter
          int     nOct = 1;
          while ((1 << nOct) < r_i)
            nOct++;
          filters[1] = new AudioFilter_OctaveResample(
              sf, -nOct, lpFilterFreq, lpFilterLen, lpFilterWindow, &gd);
          lastFilter = new AudioFilter_Delay(
              *(filters[1]), long(-(gd + 0.5)) - (offs >> nOct));
        }
      }
      else
      {
        // upsampling by integer ratio
        double  gd = 0.0;
        if (r_i & (r_i - 1))
        {
          filters[1] = new AudioFilter_Upsample(sf, r_i);
          filters[2] = new AudioFilter_FIRFilter(
              *(filters[1]), lpFilterFreq / r, lpFilterLen, r, lpFilterWindow,
              &gd);
          lastFilter = new AudioFilter_Delay(
              *(filters[2]), long(-(gd + 0.5)) - (offs * r_i));
        }
        else            // upsampling by power of two ratio
        {
          int     nOct = 1;
          while ((1 << nOct) < r_i)
            nOct++;
          filters[1] = new AudioFilter_OctaveResample(
              sf, nOct, lpFilterFreq, lpFilterLen, lpFilterWindow, &gd);
          lastFilter = new AudioFilter_Delay(
              *(filters[1]), long(-(gd + 0.5)) - (offs << nOct));
        }
      }
      return;
    }
    // fractional resample ratio and/or delay
    if (isLooped || skipFrames_i > 0.0)
    {
      sf.setPosition(sf_count_t(skipFrames_i) - 100, isLooped);
      skipFrames_f += 100.0;
    }
    if (r > 0.5)
    {
      // need to upsample first
      double  r_ds = 2.0 * isr / sampleRate;
      double  gd = 0.0;
      filters[1] = new AudioFilter_OctaveResample(
          sf, 1, (r < 1.0 ? (lpFilterFreq * r) : lpFilterFreq), lpFilterLen,
          lpFilterWindow, &gd);
      lastFilter = AudioFilter_InterpolateFast::createInterpolateFilter(
          *(filters[1]), r_ds, skipFrames_f * 2.0 + gd, interpLen);
    }
    else
    {
      // downsample only
      double  gd = 0.0;
      filters[1] = new AudioFilter_FIRFilter(
          sf, lpFilterFreq * r, lpFilterLen, 1.0, lpFilterWindow, &gd);
      lastFilter = AudioFilter_InterpolateFast::createInterpolateFilter(
          *(filters[1]), isr / sampleRate, skipFrames_f + gd, interpLen);
    }
  }
  catch (...)
  {
    for (size_t i = 0; i < filters.size(); i++)
    {
      if (filters[i])
      {
        delete filters[i];
        filters[i] = (AudioFilter *) 0;
      }
    }
    throw;
  }
}

AudioFilter_ResampledFileInput::~AudioFilter_ResampledFileInput()
{
  for (size_t i = 0; i < filters.size(); i++)
  {
    if (filters[i])
    {
      delete filters[i];
      filters[i] = (AudioFilter *) 0;
    }
  }
}

bool AudioFilter_ResampledFileInput::process()
{
  if (framesRemaining)
  {
    framesRemaining--;
    AudioFilter&  lastFilter = *(filters[filters.size() - 1]);
    if (lastFilter.process())
    {
      copyOutputBuffer(lastFilter.getOutputBuffer());
      return true;
    }
  }
  framesRemaining = 0;
  clearOutputBuffer();
  return false;
}

// ============================================================================

RandomGenerator::RandomGenerator(int seed_)
{
  if (!seed_)
    seed_ = int((unsigned int) std::time((std::time_t *) 0) & 0x7FFFFFFFU);
  while (seed_ < 1)
    seed_ += 0x7FFFFFFE;
  while (seed_ >= 0x7FFFFFFF)
    seed_ -= 0x7FFFFFFE;
  seed = seed_;
  (void) random();
  (void) random();
}

double RandomGenerator::random()
{
  long long     tmp = seed * (long long) rndMult;
  unsigned int  tmp2 = (unsigned int) (tmp & 0x7FFFFFFFLL)
                       + (unsigned int) (tmp >> 31);
  seed = int(tmp2 < 0x80000000U ? tmp2 : (tmp2 - 0x7FFFFFFFU));
  return (double(seed) * (1.0 / double(0x7FFFFFFF)));
}

// ----------------------------------------------------------------------------

void AudioOutputFile::flushBuffer()
{
  if (bufPos < 1)
    return;
  if (ditherPrvState.size() > 0)
  {
    int     chn = 0;
    for (size_t i = 0; i < bufPos; i++)
    {
      double  d = ditherRNG.random();
      buf[i] = buf[i] + ((d - ditherPrvState[chn]) * ditherLevel);
      ditherPrvState[chn] = d;
      chn = ((chn + 1) < channels ? (chn + 1) : 0);
    }
  }
  if (sf_write_double(f, &(buf.front()), sf_count_t(bufPos))
      != sf_count_t(bufPos))
  {
    throw std::runtime_error(std::string("error writing audio output file"));
  }
  bufPos = 0;
}

AudioOutputFile::AudioOutputFile(const char *fname,
                                 double sampleRate, int channels_,
                                 int sndfile_format, int ditherSeed)
  : ditherRNG(ditherSeed)
{
  if (!(sampleRate >= 1000.0 && sampleRate <= 10000000.0))
  {
    throw std::runtime_error(std::string("invalud audio output file "
                                         "sample rate"));
  }
  channels = channels_;
  if (channels < 1 || channels > 16)
  {
    throw std::runtime_error(std::string("invalud audio output file "
                                         "channel count"));
  }
  buf.resize(size_t(channels) * 4096, 0.0);
  bufPos = 0;
  ditherLevel = 0.0;
  SF_INFO sfinfo;
  std::memset(&sfinfo, 0, sizeof(SF_INFO));
  sfinfo.frames = sf_count_t(-1);
  sfinfo.samplerate = int(sampleRate + 0.5);
  sfinfo.channels = channels;
  sfinfo.format = sndfile_format;
  f = sf_open(fname, SFM_WRITE, &sfinfo);
  if (!f)
  {
    throw std::runtime_error(std::string("error opening audio output file: ")
                             + sf_strerror(f));
  }
  switch (sndfile_format & SF_FORMAT_SUBMASK)
  {
  case SF_FORMAT_PCM_S8:
  case SF_FORMAT_PCM_U8:
  case SF_FORMAT_PCM_16:
  case SF_FORMAT_PCM_24:
    if (ditherSeed >= 0)
    {
      ditherPrvState.resize(size_t(channels));
      for (int i = 0; i < channels; i++)
        ditherPrvState[i] = ditherRNG.random();
      switch (sndfile_format & SF_FORMAT_SUBMASK)
      {
      case SF_FORMAT_PCM_S8:
      case SF_FORMAT_PCM_U8:
        ditherLevel = 1.0 / 128.0;
        break;
      case SF_FORMAT_PCM_16:
        ditherLevel = 1.0 / 32768.0;
        break;
      case SF_FORMAT_PCM_24:
        ditherLevel = 1.0 / 8388608.0;
        break;
      }
    }
  case SF_FORMAT_PCM_32:
    // clip integer formats
    sf_command(f, SFC_SET_CLIPPING, NULL, SF_TRUE);
    break;
  }
  sf_command(f, SFC_SET_UPDATE_HEADER_AUTO, NULL, 0);
}

AudioOutputFile::~AudioOutputFile()
{
  try
  {
    flushBuffer();
  }
  catch (...)
  {
  }
  sf_close(f);
}

void AudioOutputFile::writeSampleFrame(const double *srcBuf)
{
  double  *bufwp = &(buf.front()) + bufPos;
  int     i = 0;
  do
  {
    bufwp[i] = srcBuf[i];
  }
  while (++i < channels);
  bufPos = bufPos + size_t(i);
  if (bufPos >= buf.size())
    flushBuffer();
}

#endif  // DSPLIB_CPP_INCLUDED

