
#ifndef DSPLIB_HPP_INCLUDED
#define DSPLIB_HPP_INCLUDED

#include "headers.hpp"

// c1 = cos(twoPi * n / (N - 1))

static inline double cosineToNuttallWindow(double c1)
{
  // Nuttall window, continuous first derivative
#if 0
  double  c2 = c1 * (c1 * 2.0) - 1.0;
  double  c3 = c2 * (c1 * 2.0) - c1;
  return (0.355768 - (0.487396 * c1) + (0.144232 * c2) - (0.012604 * c3));
#endif
  // optimized version
  return (((-0.050416 * c1 + 0.288464) * c1 - 0.449584) * c1 + 0.211536);
}

class SineGen
{
 protected:
  long double s1, s2, c;
  // --------
  inline SineGen()
  {
  }
 public:
  // n = cycle time in samples (omega = 2 * PI / n)
  SineGen(double n);
  // p = start phase (0 = sine, PI / 2 = cosine, PI = -sine, etc.)
  SineGen(double n, double p);
  inline double sample()
  {
    double      y = double(s1);
    long double s = s1 * c - s2;
    s2 = s1;
    s1 = s;
    return y;
  }
};

class CosineGen : public SineGen
{
 public:
  // n = cycle time in samples (omega = 2 * PI / n)
  CosineGen(double n);
  // p = start phase (0 = cosine, PI / 2 = -sine, PI = -cosine, etc.)
  CosineGen(double n, double p);
  // cosOmega = cos(omega)
  // sinOmega = sin(omega)
  CosineGen(long double cosOmega, long double sinOmega, double p)
  {
    s1 = std::cos((long double) p);
    // this can be compiled using the FSINCOS instruction,
    // and is faster than std::cos(p - omega)
    s2 = (s1 * cosOmega) + (std::sin((long double) p) * sinOmega);
    c = cosOmega * 2.0L;
  }
};

class GaussianWindow
{
 protected:
  long double y;
  long double c1;
  long double c2;
  long double c;
 public:
  // n = window size (N) in samples
  // wMin = minimum value (the first and last sample)
  GaussianWindow(long n, double wMin);
  // this version of the constructor takes a fractional length as N-1,
  // and allows for setting the initial position
  GaussianWindow(long double nM1, double wMin, long double startPos);
  // -((N - 1) / 2) <= n <= ((N - 1) / 2)
  inline void setPosition(long double n)
  {
    y = std::exp(n * n * c);
    c1 = std::exp((2.0L * n + 1.0L) * c);
  }
  inline double sample()
  {
    double  r = double(y);
    y *= c1;
    c1 *= c2;
    return r;
  }
};

// same as GaussianWindow, but the output is normalized to the range 0 to 1

class GaussianWindowNormalized : public GaussianWindow
{
 protected:
  double  offs;
  double  mult;
 public:
  GaussianWindowNormalized(long n, double wMin);
  inline double sample()
  {
    double  r = (double(y) + offs) * mult;
    y *= c1;
    c1 *= c2;
    return r;
  }
};

// ----------------------------------------------------------------------------

class FFTSineGen
{
 protected:
  long double   phase_re, phase_im, omega_re, omega_im;
 public:
  // n = cycle time in samples (omega = 2 * PI / n)
  FFTSineGen(long n);
  // re = cosine
  // im = -sine
  inline void sample(double& re, double& im)
  {
    re = double(phase_re);
    im = double(phase_im);
    long double tmp = (phase_re * omega_re) - (phase_im * omega_im);
    phase_im = (phase_re * omega_im) + (phase_im * omega_re);
    phase_re = tmp;
  }
  inline void reset()
  {
    phase_re = 1.0L;
    phase_im = 0.0L;
  }
};

// 'buf' should contain n * 2 floats (interleaved real/imaginary data)

static void complexFFT(double *buf, size_t n, bool isInverse);

static void realFFT(double *buf, size_t n, bool isInverse);

// ============================================================================

class AudioFilter
{
 public:
  // previous filter in the chain: the first one must be a generator
  AudioFilter   *prv;
 protected:
  double        sampleRate;
  int           channels;
  std::vector< double > outBuf; // one sample frame ('channels' samples)
  // --------
  inline void clearOutputBuffer()
  {
    double  *bufwp = &(outBuf.front());
    int     i = 0;
    do
    {
      bufwp[i] = 0.0;
    }
    while (++i < channels);
  }
  inline void copyOutputBuffer(const std::vector< double >& inBuf)
  {
    const double  *bufrp = &(inBuf.front());
    double  *bufwp = &(outBuf.front());
    int     i = 0;
    do
    {
      bufwp[i] = bufrp[i];
    }
    while (++i < channels);
  }
  inline void copyOutputBuffer(const double *inBuf)
  {
    double  *bufwp = &(outBuf.front());
    int     i = 0;
    do
    {
      bufwp[i] = inBuf[i];
    }
    while (++i < channels);
  }
  AudioFilter(AudioFilter& prv_);
  AudioFilter(double sampleRate_, int channels_);
 public:
  AudioFilter();
  virtual ~AudioFilter();
  // get one sample frame of output signal
  // the return value is true on success
  virtual bool process() = 0;
  inline double getSampleRate() const
  {
    return sampleRate;
  }
  inline int getChannelCount() const
  {
    return channels;
  }
  inline const std::vector< double >& getOutputBuffer() const
  {
    return outBuf;
  }
};

// ----------------------------------------------------------------------------

class AudioFilter_FileInput : public AudioFilter
{
 protected:
  SNDFILE     *f;
  sf_count_t  fileLength;               // in sample frames
  std::vector< double > fileBuf;
  sf_count_t  filePos;
  size_t      bufPos;                   // in samples, not sample frames
  bool        isLooped;
  // --------
  void fillBuffer();
  inline void setSampleRate(double sampleRate_)
  {
    sampleRate = sampleRate_;
  }
  friend class AudioFilter_ResampledFileInput;
 public:
  AudioFilter_FileInput(const char *fname);
  virtual ~AudioFilter_FileInput();
  virtual bool process();
  // set the file read position (sample frames),
  // and if the playback should be looped
  virtual void setPosition(sf_count_t pos, bool looped);
  // return the sound file length in sample frames
  inline sf_count_t getFileLength() const
  {
    return fileLength;
  }
  // return the current file position in sample frames
  inline sf_count_t getPosition() const
  {
    return filePos;
  }
};

// ----------------------------------------------------------------------------

class AudioFilter_Delay : public AudioFilter
{
 protected:
  long    delayFrames;
  size_t  framesRemaining;
 public:
  AudioFilter_Delay(AudioFilter& prv_,
                    long delayFrames_, size_t lengthLimit = 0xFFFFFFFFUL);
  virtual ~AudioFilter_Delay();
  virtual bool process();
};

// ----------------------------------------------------------------------------

class AudioFilter_Upsample : public AudioFilter
{
 protected:
  int     r;                    // upsample ratio
  int     cnt;
  bool    isEOF;
 public:
  AudioFilter_Upsample(AudioFilter& prv_, int upsampleRatio);
  virtual ~AudioFilter_Upsample();
  virtual bool process();
};

// ----------------------------------------------------------------------------

class AudioFilter_Downsample : public AudioFilter
{
 protected:
  int     r;                    // downsample ratio
  bool    isEOF;
 public:
  AudioFilter_Downsample(AudioFilter& prv_, int downsampleRatio);
  virtual ~AudioFilter_Downsample();
  virtual bool process();
};

// ----------------------------------------------------------------------------

static void minimumPhaseIR(std::vector< double >& ir);

// f = cutoff frequency as a fraction (0 to 0.5) of the sample rate;
//     a negative value will create a minimum phase response
// a = gain (1.0 = 0 dB)
// winParam = 0 for a Nuttall window (continuous first derivative), or
//            a negative value for a GaussianWindowNormalized with a 'wMin'
//            parameter of the specified number of dB

static void createLowpassIR(std::vector< double >& ir, size_t irLen,
                            double f, double a = 1.0, int winParam = 0);

class AudioFilter_FIRFilter : public AudioFilter
{
 protected:
  std::vector< std::vector< double > >  irFFTs;
  std::vector< double > fftBuf;
  std::vector< std::vector< double > >  inBufs;
  std::vector< std::vector< double > >  outBufs;
  size_t  outBufPos;        // read position in output mix buffer
  size_t  framesRemaining;  // samples left in output buffer before next FFT
  size_t  irLength;         // impulse response length in sample frames
  size_t  blockLength;      // irLength rounded up to the nearest power of 2
  size_t  fftLength;        // blockLength * 2
  bool    isEOF;
  // --------
  void initFilter(const std::vector< std::vector< double > >& impulseResponses);
 public:
  AudioFilter_FIRFilter(
      AudioFilter& prv_,
      const std::vector< std::vector< double > >& impulseResponses);
  // create linear phase lowpass filter with a cutoff frequency of 'f' (as a
  // fraction of the sample rate in the range 0 to 0.5, it can also be negative
  // for a minimum phase response), gain 'a' (1.0 = 0 dB), and impulse response
  // length of 'irlen' (+samples or -microseconds); the length is rounded up
  // to an even number of samples. If 'winParam' is negative, a
  // GaussianWindowNormalized (see above) will be used with a 'wMin' parameter
  // of the specified number of dB, instead of the default Nuttall window.
  // The low frequency group delay of the filter in samples is returned in 'd'
  // if it is not NULL
  AudioFilter_FIRFilter(AudioFilter& prv_,
                        double f, int irLen, double a = 1.0, int winParam = 0,
                        double *d = (double *) 0);
  virtual ~AudioFilter_FIRFilter();
  virtual bool process();
};

// ----------------------------------------------------------------------------

class AudioFilter_OctaveResample : public AudioFilter
{
 protected:
  std::vector< double > irFFT;
  std::vector< double > fftBuf;
  std::vector< std::vector< double > >  inBufs;
  std::vector< std::vector< double > >  outBufs;
  size_t  outBufPos;        // read position in output mix buffer
  size_t  framesRemaining;  // samples left in output buffer before next FFT
  size_t  irLength;         // impulse response length in sample frames
  size_t  blockLength_In;   // irLength rounded up to the nearest power of 2
  size_t  blockLength_Out;
  size_t  fftLength_In;     // blockLength * 2
  size_t  fftLength_Out;
  bool    isEOF;
  // --------
  void initFilter(const std::vector< double >& impulseResponse, int nOct = 0);
 public:
  // create linear phase lowpass filter with a cutoff frequency of 'f' (as a
  // fraction of the lower sample rate in the range 0 to 0.5, it can also be
  // negative for a minimum phase response), and impulse response length of
  // 'irlen' (+samples or -microseconds); the length is rounded up to an even
  // number of samples. If 'winParam' is negative, a GaussianWindowNormalized
  // (see above) will be used with a 'wMin' parameter of the specified number
  // of dB, instead of the default Nuttall window. The low frequency group
  // delay of the filter in samples is returned in 'd' if it is not NULL.
  //
  // If 'nOct' is greater than zero, the input is upsampled by nOct octaves;
  // if it is less than zero, then the output is downsampled by -nOct octaves
  AudioFilter_OctaveResample(
      AudioFilter& prv_, int nOct, double f, int irLen, int winParam = 0,
      double *d = (double *) 0);
  virtual ~AudioFilter_OctaveResample();
  virtual bool process();
};

// ----------------------------------------------------------------------------

class AudioFilter_Interpolate : public AudioFilter
{
 protected:
  std::vector< double > inBuf;  // buffer size = windowSize * channels
  size_t  framesRemaining;
  bool    isEOF;
  int     windowSize;
  int     writePos;             // in samples (not frames)
  long double readOffs;         // buffer read position relative to writePos+1
  double  r;                    // downsample ratio
  GaussianWindowNormalized  w;
 public:
  // downsampleRatio: the ratio of the input and output sample rate
  //     (<1: upsample, >1: downsample); for downsampling without aliasing,
  //     the input should be lowpass filtered first
  // startPos: initial position in the input stream (in sample frames)
  // winSize: interpolation window size (multiples of 4 in the range 8 to 2048)
  AudioFilter_Interpolate(AudioFilter& prv_,
                          double downsampleRatio, double startPos, int winSize);
  virtual ~AudioFilter_Interpolate();
  virtual bool process();
};

// ----------------------------------------------------------------------------

// faster version that uses a pre-calculated lookup table, but requires
// a simple rational resample ratio

class AudioFilter_InterpolateFast : public AudioFilter
{
 protected:
  std::vector< double > inBuf;  // buffer size = windowSize * channels
  size_t  framesRemaining;
  bool    isEOF;
  int     windowSize;
  int     writePos;             // in samples (not frames)
  int     windowNum;            // windowTable read position / windowSize
  long    readOffs;             // buffer read position relative to writePos+1
  std::vector< double >         windowTable;
  std::vector< unsigned char >  readPosIncTable;
  double  upsampleRatio;
 public:
  // n, d:    the ratio of the output and input sample rate (= n / d, n > d:
  //          upsample, n < d: downsample); for downsampling without aliasing,
  //          the input should be lowpass filtered first. The memory usage of
  //          the tables is n * (winSize * sizeof(double) + 1) bytes
  // startPos: initial position in the input stream (in sample frames)
  // winSize: interpolation window size (multiples of 4 in the range 8 to 2048)
  AudioFilter_InterpolateFast(AudioFilter& prv_,
                              int n, int d, double startPos, int winSize);
  virtual ~AudioFilter_InterpolateFast();
  virtual bool process();
  // automatically choose fast or slow interpolation depending on the
  // resample ratio
  static AudioFilter *createInterpolateFilter(
      AudioFilter& prv_, double downsampleRatio, double startPos, int winSize);
};

// ----------------------------------------------------------------------------

class AudioFilter_ResampledFileInput : public AudioFilter
{
 protected:
  std::vector< AudioFilter * >  filters;
  size_t  framesRemaining;
 public:
  // fname = input file name
  // outputSampleRate = sample rate (in Hz) to resample to; a negative value
  //                    can be used to specify a resample ratio instead (e.g.
  //                    -2.0 will resample 44100 Hz to 88200 Hz)
  // isrAdj = correction factor to be applied to input sample rate (the actual
  //          resample ratio is outputSampleRate / (inputSampleRate * isrAdj))
  // skipTime = start position in input file in seconds (negative means delay)
  // isLooped = loop input file for indefinite duration if true
  // interpLen = sinc interpolation window size
  // lpFilterLen = FIR lowpass filter length (+samples or -microseconds)
  // lpFilterFreq = FIR lowpass filter cutoff frequency as a fraction (0 to
  //                0.5) of the lower sample rate; if a negative value is
  //                specified, then a minimum phase filter will be used
  // lpFilterWindow = FIR lowpass filter window parameter; if it is zero,
  //                  a Nuttall window (continuous first derivative) is used,
  //                  while negative values will use a GaussianWindowNormalized
  //                  (see above) with a 'wMin' parameter of the specified
  //                  number of dB
  AudioFilter_ResampledFileInput(
      const char *fname, double outputSampleRate, double isrAdj = 1.0,
      double skipTime = 0.0, bool isLooped = false,
      int interpLen = 48, int lpFilterLen = 32768, double lpFilterFreq = 0.5,
      int lpFilterWindow = 0);
  virtual ~AudioFilter_ResampledFileInput();
  virtual bool process();
};

// ============================================================================

class RandomGenerator
{
 protected:
  static const int  rndMult = 742938285;
  int     seed;
 public:
  // seed_ = 0 seeds from the current time
  RandomGenerator(int seed_ = 987654321);
  // return a random number in the range 0.0 to 1.0
  double random();
};

// ----------------------------------------------------------------------------

class AudioOutputFile
{
 protected:
  SNDFILE *f;
  std::vector< double > buf;
  size_t  bufPos;
  int     channels;
  double  ditherLevel;
  RandomGenerator ditherRNG;
  std::vector< double > ditherPrvState;
  // --------
  void flushBuffer();
 public:
  // ditherSeed = 0 seeds from the current time, ditherSeed < 0 disables dither
  AudioOutputFile(const char *fname, double sampleRate, int channels_,
                  int sndfile_format = SF_FORMAT_WAV | SF_FORMAT_FLOAT,
                  int ditherSeed = 0);
  virtual ~AudioOutputFile();
  void writeSampleFrame(const double *srcBuf);
};

#endif  // DSPLIB_HPP_INCLUDED

