
#include "headers.hpp"
#include "dsplib.cpp"
#include "options.cpp"
#include "image.cpp"

#include <cctype>

static const int  imageWidth = 1024;
static const int  imageHeight = 752;
static const int  displayWidth = 936;
static const int  displayHeight = 600;
static const int  displayOriginX = 12;
static const int  displayOriginY = 640;

static const float colorTable[14 * 3] =
{
  0.250f, 0.250f, 1.000f,  1.000f, 0.000f, 0.000f,
  0.000f, 0.875f, 0.875f,  1.000f, 0.500f, 0.000f,
  0.000f, 0.500f, 0.000f,  0.500f, 0.000f, 1.000f,
  0.000f, 0.750f, 0.375f,  0.875f, 0.875f, 0.000f,
  0.000f, 0.500f, 1.000f,  1.000f, 0.500f, 0.750f,
  0.500f, 1.000f, 0.000f,  1.000f, 0.000f, 1.000f,
  0.750f, 0.750f, 0.750f,  0.500f, 0.500f, 0.500f
};

enum
{
  ANALYSIS_MODE_IR = 0,         // impulse response (0 dB = 0 dBFS impulse)
  ANALYSIS_MODE_TONE = 1,       // tone (0 dB = 0 dBFS sine wave)
  ANALYSIS_MODE_FREQ = 2,       // frequency response from 'thd_test'
  ANALYSIS_MODE_THDFREQ = 3,    // THD vs. frequency from 'thd_test'
  ANALYSIS_MODE_LEVEL = 4,      // linearity from 'thd_test'
  ANALYSIS_MODE_THDLEVEL = 5,   // THD vs. level from 'thd_test'
  ANALYSIS_MODE_PHASE = 6,      // phase response from 'groupdel'
  ANALYSIS_MODE_PD = 7,         // phase delay vs. frequency from 'groupdel'
  ANALYSIS_MODE_GD = 8          // group delay vs. frequency from 'groupdel'
};

static const char *analysisModeNames[9] =
{
  "Frequency response (from impulse response)",
  "Frequency analysis",
  "Frequency response (from sine sweep)",
  "THD vs. frequency",
  "Linearity",
  "THD vs. level",
  "Phase response",
  "Phase delay vs. frequency",
  "Group delay vs. frequency"
};

enum
{
  WINDOW_TYPE_NUTTALL = 0,
  WINDOW_TYPE_RECTANGULAR = 1,
  WINDOW_TYPE_TRIANGULAR = 2,
  WINDOW_TYPE_HAMMING = 3,
  WINDOW_TYPE_HANN = 4,
  WINDOW_TYPE_BLACKMAN = 5,
  WINDOW_TYPE_BLACKMAN_EXACT = 6,
  WINDOW_TYPE_BLACKMAN_HARRIS = 7,
  WINDOW_TYPE_BLACKMAN_NUTTALL = 8,
  WINDOW_TYPE_FLAT_TOP = 9
};

static const char *windowNames[10] =
{
  "Nuttall (cont. 1st deriv.)",
  "rectangular",
  "triangular",
  "Hamming",
  "Hann",
  "Blackman",
  "Blackman (exact)",
  "Blackman-Harris",
  "Blackman-Nuttall",
  "flat top"
};

struct AnalysisParameters
{
  int     mode;
  int     windowType;
  bool    logXScale;
  bool    logYScale;
  bool    logFreqScale;
  bool    dBScale;
  double  xMin;
  double  xMax;
  double  yMin;
  double  yMax;
  double  offs;
  double  duration;
  double  octBWMult;
  double  freqMin;
  double  freqMax;
  double  level1kHz;    // for ANALYSIS_MODE_LEVEL and ANALYSIS_MODE_THDLEVEL
  double  lvlAdjust;
};

// ----------------------------------------------------------------------------

static double createWindowFunction(std::vector< double >& buf, int windowType,
                                   double windowLength, double windowOffset)
{
  double  winLenD2 = windowLength * 0.5;
  if (windowType == WINDOW_TYPE_RECTANGULAR ||
      windowType == WINDOW_TYPE_HAMMING ||
      windowType >= WINDOW_TYPE_BLACKMAN_EXACT)
  {
    winLenD2 += 0.000001;
  }
  else
  {
    winLenD2 -= 0.000001;
  }
  long    peakPos = long(std::ceil(windowOffset));
  long    startPos = long(std::ceil(windowOffset - winLenD2));
  long    endPos = long(std::floor(windowOffset + winLenD2)) + 1L;
  buf.resize(0);
  buf.resize(size_t(endPos), 0.0);
  double  wSum = 0.0;
  if (windowType < 0 || windowType >= 10)       // Gaussian window
  {
    double  wMin = std::pow(10.0, double(std::abs(windowType)) / -20.0);
    GaussianWindow  w(windowLength, wMin, windowOffset - double(peakPos - 1L));
    double  offs = (windowType < 0 ? (-wMin) : 0.0);
    double  mult = (windowType < 0 ? (1.0 / (1.0 - wMin)) : 1.0);
    for (long i = peakPos; --i >= startPos; )
    {
      double  w_ = (w.sample() + offs) * mult;
      buf[i] = w_;
      wSum += w_;
    }
    w.setPosition(double(peakPos) - windowOffset);
    for (long i = peakPos; i < endPos; i++)
    {
      double  w_ = (w.sample() + offs) * mult;
      buf[i] = w_;
      wSum += w_;
    }
    return wSum;
  }
  switch (windowType)
  {
  case WINDOW_TYPE_RECTANGULAR:
    for (long i = startPos; i < endPos; i++)
      buf[i] = 1.0;
    wSum = double(endPos - startPos);
    break;
  case WINDOW_TYPE_TRIANGULAR:
    {
      double  a = 2.0 / windowLength;
      for (long i = startPos; i < endPos; i++)
      {
        double  w = std::fabs(double(i) - windowOffset) * a;
        buf[i] = w;
        wSum += w;
      }
    }
    break;
  case WINDOW_TYPE_NUTTALL:
  case WINDOW_TYPE_HAMMING:
  case WINDOW_TYPE_HANN:
  case WINDOW_TYPE_BLACKMAN:
  case WINDOW_TYPE_BLACKMAN_EXACT:
  case WINDOW_TYPE_BLACKMAN_HARRIS:
  case WINDOW_TYPE_BLACKMAN_NUTTALL:
  case WINDOW_TYPE_FLAT_TOP:
    {
      double  a0 = 0.0;
      double  a1 = 0.0;
      double  a2 = 0.0;
      double  a3 = 0.0;
      double  a4 = 0.0;
      switch (windowType)
      {
      case WINDOW_TYPE_NUTTALL:
        a0 = 0.211536;
        a1 = -0.449584;
        a2 = 0.288464;
        a3 = -0.050416;
        break;
      case WINDOW_TYPE_HAMMING:
        a0 = 0.53836;
        a1 = -0.46164;
        break;
      case WINDOW_TYPE_HANN:
        a0 = 0.5;
        a1 = -0.5;
        break;
      case WINDOW_TYPE_BLACKMAN:
        a0 = 0.34;
        a1 = -0.5;
        a2 = 0.16;
        break;
      case WINDOW_TYPE_BLACKMAN_EXACT:
        a0 = 0.34974135;
        a1 = -0.4965605;
        a2 = 0.15369815;
        break;
      case WINDOW_TYPE_BLACKMAN_HARRIS:
        a0 = 0.21747;
        a1 = -0.45325;
        a2 = 0.28256;
        a3 = -0.04672;
        break;
      case WINDOW_TYPE_BLACKMAN_NUTTALL:
        a0 = 0.2269824;
        a1 = -0.4572542;
        a2 = 0.273199;
        a3 = -0.0425644;
        break;
      case WINDOW_TYPE_FLAT_TOP:
        a0 = -0.06943231;
        a1 = -0.16724891;
        a2 = 0.51441048;
        a3 = -0.33886463;
        a4 = 0.0489083;
        break;
      }
      {
        CosineGen w(windowLength,
                    (windowOffset + (windowLength * 0.5) - double(peakPos - 1L))
                    * twoPi / windowLength);
        for (long i = peakPos; --i >= startPos; )
        {
          double  c1 = w.sample();
          double  w_ = (((a4 * c1 + a3) * c1 + a2) * c1 + a1) * c1 + a0;
          buf[i] = w_;
          wSum += w_;
        }
      }
      {
        CosineGen w(windowLength,
                    (double(peakPos) - (windowOffset - (windowLength * 0.5)))
                    * twoPi / windowLength);
        for (long i = peakPos; i < endPos; i++)
        {
          double  c1 = w.sample();
          double  w_ = (((a4 * c1 + a3) * c1 + a2) * c1 + a1) * c1 + a0;
          buf[i] = w_;
          wSum += w_;
        }
      }
    }
    break;
  default:
    throw std::runtime_error(std::string("invalid window type"));
  }
  return wSum;
}

static void getMagnitudeResponse(std::vector< double >& fftBuf)
{
  for (size_t i = 0; i < 2 && i < fftBuf.size(); i++)
    fftBuf[i] = std::fabs(fftBuf[i]);
  if (fftBuf.size() < 3)
    return;
  double  tmp = fftBuf[1];              // save magnitude at Nyquist frequency
  for (size_t i = 2; i < fftBuf.size(); i += 2)
  {
    double  re = fftBuf[i];
    double  im = fftBuf[i + 1];
    double  r2 = (re * re) + (im * im);
    if (r2 > 0.0)
      fftBuf[i >> 1] = std::sqrt(r2);
    else
      fftBuf[i >> 1] = 0.0;
  }
  fftBuf[fftBuf.size() >> 1] = tmp;
  fftBuf.resize((fftBuf.size() >> 1) + 1);
}

static void convertPhaseResponse(std::vector< double >& fftBuf,
                                 const AnalysisParameters& params)
{
  double  scale = (params.mode == ANALYSIS_MODE_PHASE ? 100.0 : 10.0);
  double  r_fsD2 = fftBuf[1];           // save magnitude at Nyquist frequency
  fftBuf[1] = 0.0;
  for (size_t i = 0; i <= (fftBuf.size() >> 1); i++)
  {
    double  re = (i < (fftBuf.size() >> 1) ? fftBuf[i << 1] : r_fsD2);
    double  im = (i < (fftBuf.size() >> 1) ? fftBuf[(i << 1) + 1] : 0.0);
    double  r2 = (re * re) + (im * im);
    fftBuf[i] = scale * std::log10(r2 > 1.0e-40 ? r2 : 1.0e-40);
  }
  fftBuf.resize((fftBuf.size() >> 1) + 1);
}

// f = frequency / sample rate

static double octaveBandFilter(const std::vector< double >& buf, double f,
                               const AnalysisParameters& params)
{
  if (!(f >= 0.0 && f <= 0.5))
    return 0.0;
  double  bw = f * params.octBWMult;
  bw = (bw > 1.0e-6 ? (bw < 2.0 ? bw : 2.0) : 1.0e-6);
  double  winLen = 5.0 / bw;
  long    peakPos = long(buf.size() >> 1);
  long    startPos = long(std::floor(double(peakPos) - (winLen * 0.5))) - 1L;
  long    endPos = long(std::ceil(double(peakPos) + (winLen * 0.5))) + 2L;
  startPos = (startPos > 0L ? startPos : 0L);
  endPos = (endPos < long(buf.size()) ? endPos : long(buf.size()));
  GaussianWindow  w(winLen, 2.176413e-10, double(startPos - peakPos));
  long double c = 2.0L * std::cos(twoPiL * (long double) f);
  long double s1 = 0.0L;
  long double s2 = 0.0L;
  for (long i = startPos; i < peakPos; i++)
  {
    double  x = buf[i] * w.sample();
    long double s = (long double) x + (s1 * c) - s2;
    s2 = s1;
    s1 = s;
  }
  w.setPosition(0.0);
  for (long i = peakPos; i < endPos; i++)
  {
    double  x = buf[i] * w.sample();
    long double s = (long double) x + (s1 * c) - s2;
    s2 = s1;
    s1 = s;
  }
  double  p = double((s1 * s1) + (s2 * s2) - (s1 * s2 * c));
  if (p > 0.0)
    return std::sqrt(p);
  return 0.0;
}

static inline double getFreqFromXPos(double x, const AnalysisParameters& params)
{
  if (params.logFreqScale)
  {
    return (params.freqMin * std::pow(params.freqMax / params.freqMin,
                                      x / double(displayWidth - 1)));
  }
  return (params.freqMin + ((params.freqMax - params.freqMin)
                            * x / double(displayWidth - 1)));
}

static double getMagnitudeAtBin(const std::vector< double >& buf, double n)
{
  long    i0 = long(std::floor(n));
  long    i1 = i0 + 1L;
  double  f = n - double(i0);
  if (i0 < 0L)
  {
    if (i1 < 0L || f < 0.99999)
      return 0.0;
    return buf[0];
  }
  if (i1 >= long(buf.size()))
  {
    if (i0 >= long(buf.size()) || f > 0.00001)
      return 0.0;
    return buf[buf.size() - 1];
  }
  return (buf[i0] + ((buf[i1] - buf[i0]) * f));
}

static void analyzeFile(std::vector< double > *graphData,
                        AudioFilter_FileInput& f,
                        const AnalysisParameters& params)
{
  std::vector< std::vector< double > >  bufs((size_t) f.getChannelCount());
  std::vector< double > window;
  double  sampleRate = f.getSampleRate();
  double  windowLength = 65536.0;
  double  windowOffset = 32768.0;
  size_t  windowSize = 65536;
  size_t  fftSize = 65536;
  double  scale = 0.000085147442;
  switch (params.mode)
  {
  case ANALYSIS_MODE_FREQ:
  case ANALYSIS_MODE_LEVEL:
    if (f.getFileLength() != 131072)
      throw std::runtime_error(std::string("invalid input file length"));
    f.setPosition(0, false);
    break;
  case ANALYSIS_MODE_THDFREQ:
  case ANALYSIS_MODE_THDLEVEL:
    if (!params.dBScale)
      scale *= 100.0;
    if (f.getFileLength() != 131072)
      throw std::runtime_error(std::string("invalid input file length"));
    f.setPosition(65536, false);
    break;
  case ANALYSIS_MODE_PHASE:
  case ANALYSIS_MODE_PD:
  case ANALYSIS_MODE_GD:
    if (f.getFileLength() != 65536)
      throw std::runtime_error(std::string("invalid input file length"));
    f.setPosition(0, false);
    break;
  default:
    windowLength = (params.duration < 0.0 ?
                    -(params.duration) : (params.duration * sampleRate));
    windowLength = (windowLength > 8.0 ?
                    (windowLength < 16777216.0 ? windowLength : 16777216.0)
                    : 8.0);
    {
      double  t = (params.offs < 0.0 ?
                   -(params.offs) : (params.offs * sampleRate));
      sf_count_t  filePos = sf_count_t(std::floor(t - (windowLength * 0.5)))
                            - 1L;
      windowOffset = t - double(filePos);
      double  wSum = createWindowFunction(window, params.windowType,
                                          windowLength, windowOffset);
      if (params.mode == ANALYSIS_MODE_IR)
        scale = 1.0;
      else
        scale = 2.0 / wSum;
      size_t  winStartPos = 0;
      size_t  winEndPos = window.size();
      while ((winStartPos + 1) < winEndPos && window[winStartPos] == 0.0)
        winStartPos++;
      while ((winStartPos + 1) < winEndPos && window[winEndPos - 1] == 0.0)
        winEndPos--;
      if (winEndPos < window.size())
        window.resize(winEndPos);
      if (winStartPos > 0)
      {
        windowOffset -= double(long(winStartPos));
        filePos += sf_count_t(winStartPos);
        window.erase(window.begin(), window.begin() + winStartPos);
      }
      f.setPosition(filePos, false);
      windowSize = window.size();
      fftSize = 16;
      while (fftSize < windowSize)
        fftSize = fftSize << 1;
    }
    break;
  }
  for (size_t i = 0; i < bufs.size(); i++)
    bufs[i].resize(fftSize, 0.0);
  if (params.lvlAdjust != 0.0)
    scale = scale * std::pow(10.0, params.lvlAdjust / 20.0);
  for (size_t i = 0; i < windowSize; i++)
  {
    f.process();
    double  w = scale;
    if (i < window.size())
      w = w * window[i];
    for (size_t j = 0; j < bufs.size(); j++)
      bufs[j][i] = f.getOutputBuffer()[j] * w;
  }
  if (!(params.octBWMult > 0.0))
  {
    // FFT analysis
    for (size_t i = 0; i < bufs.size(); i++)
    {
      realFFT(&(bufs[i].front()), fftSize, false);
      if (params.mode != ANALYSIS_MODE_PHASE &&
          params.mode != ANALYSIS_MODE_PD && params.mode != ANALYSIS_MODE_GD)
      {
        getMagnitudeResponse(bufs[i]);
      }
      else
      {
        convertPhaseResponse(bufs[i], params);
      }
      for (int j = 0; j < displayWidth; j++)
      {
        double  f0 = getFreqFromXPos(double(j) - 0.5, params);
        double  f1 = getFreqFromXPos(double(j) + 0.5, params);
        f0 = (f0 < params.freqMin ? params.freqMin : f0) / sampleRate;
        f1 = (f1 > params.freqMax ? params.freqMax : f1) / sampleRate;
        f0 = (f0 < 0.0 ? 0.0 : f0);
        f1 = (f1 > 0.5 ? 0.5 : f1);
        if (!(f1 > f0))
        {
          graphData[i][j << 1] = 0.0;
          graphData[i][(j << 1) + 1] = 0.0;
          continue;
        }
        f0 *= double(long(fftSize));
        f1 *= double(long(fftSize));
        double  yMin = getMagnitudeAtBin(bufs[i], f0);
        double  yMax = yMin;
        double  y = getMagnitudeAtBin(bufs[i], f1);
        yMin = (y < yMin ? y : yMin);
        yMax = (y > yMax ? y : yMax);
        long    n0 = long(std::ceil(f0));
        long    n1 = long(std::floor(f1));
        n0 = (n0 > 0L ? n0 : 0L);
        n1 = (n1 < long(fftSize >> 1) ? n1 : long(fftSize >> 1));
        for (long n = n0; n <= n1; n++)
        {
          y = bufs[i][n];
          yMin = (y < yMin ? y : yMin);
          yMax = (y > yMax ? y : yMax);
        }
        graphData[i][j << 1] = yMin;
        graphData[i][(j << 1) + 1] = yMax;
      }
    }
  }
  else
  {
    // octave band analysis
    for (size_t i = 0; i < bufs.size(); i++)
    {
      // convert input to linear phase
      realFFT(&(bufs[i].front()), fftSize, false);
      bufs[i][0] = std::fabs(bufs[i][0]);
      bufs[i][1] = std::fabs(bufs[i][1]);
      for (size_t j = 2; j < fftSize; j += 2)
      {
        double  re = bufs[i][j];
        double  im = bufs[i][j + 1];
        double  r2 = (re * re) + (im * im);
        if (r2 > 0.0)
          bufs[i][j] = std::sqrt(r2) * ((j & 2) == 0 ? 1.0 : -1.0);
        else
          bufs[i][j] = 0.0;
        bufs[i][j + 1] = 0.0;
      }
      realFFT(&(bufs[i].front()), fftSize, true);
      for (int j = 0; j < displayWidth; j++)
      {
        double  freq = getFreqFromXPos(double(j), params);
        double  y = octaveBandFilter(bufs[i], freq / sampleRate, params);
        graphData[i][j << 1] = y;
        graphData[i][(j << 1) + 1] = y;
      }
      double  prvY = graphData[i][0];
      for (int j = 0; j < displayWidth; j++)
      {
        double  y = graphData[i][j << 1];
        double  nxtY =
            graphData[i][(j < (displayWidth - 1) ? (j + 1) : j) << 1];
        double  y0 = (y + prvY) * 0.5;
        double  y1 = (y + nxtY) * 0.5;
        if (y0 > y1)
        {
          double  tmp = y0;
          y0 = y1;
          y1 = tmp;
        }
        graphData[i][j << 1] = (y < y0 ? y : y0);
        graphData[i][(j << 1) + 1] = (y > y1 ? y : y1);
        prvY = y;
      }
    }
  }
}

// ----------------------------------------------------------------------------

static void drawRGBAPixel(Image& img, int x, int y,
                          float r, float g, float b, float a)
{
  if (a < 0.0001f)
    return;
  if (a < 0.9999f)
  {
    unsigned char r0 = 0;
    unsigned char g0 = 0;
    unsigned char b0 = 0;
    img.getColor(r0, g0, b0, img.getPixel(x, y));
    float   a0 = (1.0f - a) * (1.0f / 255.0f);
    r = (r * a) + (float(r0) * a0);
    g = (g * a) + (float(g0) * a0);
    b = (b * a) + (float(b0) * a0);
  }
  float   d = img.getDitherOffset(x, y);
  r = (r > 0.0f ? (r < 1.0f ? r : 1.0f) : 0.0f) * 5.0f + d;
  g = (g > 0.0f ? (g < 1.0f ? g : 1.0f) : 0.0f) * 6.0f + d;
  b = (b > 0.0f ? (b < 1.0f ? b : 1.0f) : 0.0f) * 5.0f + d;
  img.setPixel(x, y, (unsigned char) ((int(r) * 7 + int(g)) * 6 + int(b) + 4));
}

// each buffer contains displayWidth * 2 floats (interleaved min/max values)

static void drawGraphs(Image& img,
                       const std::vector< std::vector< double > >& bufs,
                       const std::vector< unsigned char >& colors,
                       const AnalysisParameters& params)
{
  int     n = int(bufs.size());
  std::vector< std::vector< std::vector< float > > >  alphaBufs((size_t) n);
  for (int i = 0; i < n; i++)
  {
    alphaBufs[i].resize(3);
    for (int j = 0; j < 3; j++)
      alphaBufs[i][j].resize(size_t(displayHeight), 0.0f);
  }
  for (int x = 0; x <= displayWidth; x++)
  {
    int     aBufIdx = x % 3;
    if (x < displayWidth)
    {
      for (int i = 0; i < n; i++)
      {
        float   *alphaBuf = &(alphaBufs[i][aBufIdx].front());
        for (int y = 0; y < displayHeight; y++)
          alphaBuf[y] = 0.0f;
        double  yMin = bufs[i][x << 1];
        double  yMax = bufs[i][(x << 1) + 1];
        if (yMin > yMax)
        {
          double  tmp = yMin;
          yMin = yMax;
          yMax = tmp;
        }
        yMin = (yMin > params.yMin ? yMin : params.yMin);
        yMax = (yMax < params.yMax ? yMax : params.yMax);
        if (!params.logYScale)
        {
          if (!(yMin < (params.yMax * 2.0 - params.yMin) &&
                yMax > (params.yMin * 2.0 - params.yMax)))
          {
            continue;
          }
          yMin = (yMin - params.yMin) / (params.yMax - params.yMin);
          yMax = (yMax - params.yMin) / (params.yMax - params.yMin);
        }
        else
        {
          if (!(yMin < (params.yMax * params.yMax / params.yMin) &&
                yMax > (params.yMin * params.yMin / params.yMax)))
          {
            continue;
          }
          yMin = std::log(yMin / params.yMin)
                 / std::log(params.yMax / params.yMin);
          yMax = std::log(yMax / params.yMin)
                 / std::log(params.yMax / params.yMin);
        }
        yMin = yMin * double(displayHeight - 1);
        yMax = yMax * double(displayHeight - 1);
        int     y0 = int(std::floor(yMin - 0.3125));
        int     y1 = int(std::ceil(yMax + 0.3125));
        y0 = (y0 > 0 ? y0 : 0);
        y1 = (y1 < (displayHeight - 1) ? y1 : (displayHeight - 1));
        for (int y = y0; y <= y1; y++)
        {
          float   a = float(double(y) - yMin);
          if (a < -0.3125f)
          {
            a = a + 1.3125f;
          }
          else
          {
            a = float(yMax - double(y));
            if (a < -0.3125f)
              a = a + 1.3125f;
            else
              a = 1.0f;
          }
          if (a <= 0.0f)
            continue;
          alphaBuf[y] = 1.0f - ((1.0f - alphaBuf[y]) * (1.0f - a));
        }
      }
    }
    if (x > 0)
    {
      int     aBufIdxM1 = (aBufIdx > 0 ? (aBufIdx - 1) : (aBufIdx + 2));
      int     aBufIdxM2 = (aBufIdx > 1 ? (aBufIdx - 2) : (aBufIdx + 1));
      for (int y = 0; y < displayHeight; y++)
      {
        float   r = 0.0f;
        float   g = 0.0f;
        float   b = 0.0f;
        float   a = 0.0f;
        int     i = int(img.getDitherOffset(x - 1, y) * float(n));
        for (int j = 0; j < n; j++)
        {
          float   a_ = 1.0f - ((1.0f - (alphaBufs[i][aBufIdxM2][y] * 0.5f))
                               * (1.0f - (alphaBufs[i][aBufIdxM1][y]))
                               * (1.0f - (alphaBufs[i][aBufIdx][y] * 0.5f)));
          if (a_ >= 0.0001f)
          {
            a = 1.0f - ((1.0f - a) * (1.0f - a_));
            a_ = a_ / a;
            const float *cp = &(colorTable[int(colors[i]) * 3]);
            r = r + ((cp[0] - r) * a_);
            g = g + ((cp[1] - g) * a_);
            b = b + ((cp[2] - b) * a_);
          }
          i = ((i + 1) < n ? (i + 1) : 0);
        }
        drawRGBAPixel(img, displayOriginX + x - 1, displayOriginY - y,
                      r, g, b, a);
      }
    }
  }
}

static void printValue(Image& img, int x, int y, double n)
{
  char    strBuf[64];
  if (std::fabs(n) < 0.0000000005)
    std::sprintf(strBuf, "0");
  else if (std::fabs(n) < 0.0009995)
    std::sprintf(strBuf, "%gu", n * 1000000.0);
  else if (std::fabs(n) < 0.9995)
    std::sprintf(strBuf, "%gm", n * 1000.0);
  else if (std::fabs(n) < 999.5)
    std::sprintf(strBuf, "%g", n);
  else if (std::fabs(n) < 999500.0)
    std::sprintf(strBuf, "%gk", n * 0.001);
  else
    std::sprintf(strBuf, "%gM", n * 0.000001);
  img.drawText(x, y, 1, strBuf);
}

// ----------------------------------------------------------------------------

int main(int argc, char **argv)
{
  try
  {
    std::vector< std::string >  inFileNames(sizeof(colorTable)
                                            / (sizeof(float) * 3 * 2),
                                            std::string(""));
    std::string outFileName = "";
    std::string titleString = "";
    std::vector< std::string >  descriptionStrings(inFileNames.size(),
                                                   std::string(""));
    std::vector< int >  graphColorTable(inFileNames.size() * 2);
    for (size_t i = 0; i < graphColorTable.size(); i++)
      graphColorTable[i] = int(i + 1);
    std::string analysisMode = "ir";
    AnalysisParameters  params;
    params.mode = ANALYSIS_MODE_IR;
    params.windowType = WINDOW_TYPE_NUTTALL;
    params.logXScale = true;
    params.logYScale = true;
    params.logFreqScale = true;
    params.dBScale = true;
    params.xMin = 0.0;
    params.xMax = 0.0;
    params.yMin = 0.0;
    params.yMax = 0.0;
    params.offs = 0.0;
    params.duration = 1.0;
    params.octBWMult = 0.0;
    params.freqMin = 0.0;
    params.freqMax = 0.0;
    params.level1kHz = -26.0206;        // default to 20 kHz = 0 dBr
    params.lvlAdjust = 0.0;
    bool    blackBackground = false;
    {
      CommandLineParser opts(argv[0]);
      opts.addFileName("OUTFILE", "output (BMP image) file name", outFileName);
      opts.addOption("title", "ts",
                     "title string, defaults to the analysis mode",
                     titleString);
      opts.addOption("input-file-1", "i1", "", inFileNames[0]);
      opts.addOption("input-file-2", "i2", "", inFileNames[1]);
      opts.addOption("input-file-3", "i3", "", inFileNames[2]);
      opts.addOption("input-file-4", "i4", "", inFileNames[3]);
      opts.addOption("input-file-5", "i5", "", inFileNames[4]);
      opts.addOption("input-file-6", "i6", "", inFileNames[5]);
      opts.addOption("input-file-7", "i7",
                     "input (audio) file name(s); at least one must be "
                     "specified",
                     inFileNames[6]);
      opts.addOption("description1", "ds1", "", descriptionStrings[0]);
      opts.addOption("description2", "ds2", "", descriptionStrings[1]);
      opts.addOption("description3", "ds3", "", descriptionStrings[2]);
      opts.addOption("description4", "ds4", "", descriptionStrings[3]);
      opts.addOption("description5", "ds5", "", descriptionStrings[4]);
      opts.addOption("description6", "ds6", "", descriptionStrings[5]);
      opts.addOption("description7", "ds7",
                     "description string for the corresponding input file "
                     "(defaults to\n"
                     "        the file name)",
                     descriptionStrings[6]);
      opts.addOption("color1l", "c1l", "", graphColorTable[0], 1, 14);
      opts.addOption("color1r", "c1r", "", graphColorTable[1], 1, 14);
      opts.addOption("color2l", "c2l", "", graphColorTable[2], 1, 14);
      opts.addOption("color2r", "c2r", "", graphColorTable[3], 1, 14);
      opts.addOption("color3l", "c3l", "", graphColorTable[4], 1, 14);
      opts.addOption("color3r", "c3r", "", graphColorTable[5], 1, 14);
      opts.addOption("color4l", "c4l", "", graphColorTable[6], 1, 14);
      opts.addOption("color4r", "c4r", "", graphColorTable[7], 1, 14);
      opts.addOption("color5l", "c5l", "", graphColorTable[8], 1, 14);
      opts.addOption("color5r", "c5r", "", graphColorTable[9], 1, 14);
      opts.addOption("color6l", "c6l", "", graphColorTable[10], 1, 14);
      opts.addOption("color6r", "c6r", "", graphColorTable[11], 1, 14);
      opts.addOption("color7l", "c7l", "", graphColorTable[12], 1, 14);
      opts.addOption("color7r", "c7r",
                     "graph color for the left or right channel of the "
                     "corresponding\n"
                     "        input file",
                     graphColorTable[13], 1, 14);
      opts.addOption("analysis-mode", "mode",
                     "analysis mode (ir, tone, freq, thdf, lvl, thdl, phs, "
                     "pd, or gd)",
                     analysisMode);
      opts.addOption("black-background", "bk",
                     "use alternate color map with black background",
                     blackBackground);
      opts.addOption("window", "w",
                     "window type:\n"
                     "            0: Nuttall (continuous first derivative)\n"
                     "            1: rectangular\n"
                     "            2: triangular\n"
                     "            3: Hamming\n"
                     "            4: Hann\n"
                     "            5: Blackman\n"
                     "            6: Blackman (exact)\n"
                     "            7: Blackman-Harris\n"
                     "            8: Blackman-Nuttall\n"
                     "            9: flat top\n"
                     "         >=10: Gaussian window with a minimum value of "
                     "the specified\n"
                     "               number in -dB\n"
                     "           <0: Gaussian window normalized to the range "
                     "0 to 1",
                     params.windowType, -240, 240);
      opts.addOption("offset", "t",
                     "center position of window in seconds or -samples",
                     params.offs, -1000000000.0, 10000.0);
      opts.addOption("duration", "d",
                     "window length in seconds or -samples",
                     params.duration, -16777216.0, 300.0);
      opts.addOption("bandwidth-mult", "bwm",
                     "if greater than zero, bandwidth (-6.02 dB) / frequency "
                     "ratio for\n"
                     "        octave band smoothing",
                     params.octBWMult, 0.0, 4.0);
      opts.addOption("minfreq", "fmin", "minimum frequency to display (in Hz)",
                     params.freqMin, 0.0, 500000.0);
      opts.addOption("maxfreq", "fmax",
                     "maximum frequency to display; if the frequency range "
                     "is 0-0 Hz,\n"
                     "        it is set automatically from the sample rate of "
                     "the first input\n"
                     "        file",
                     params.freqMax, 0.0, 500000.0);
      opts.addOption("minvalue", "ymin", "minimum value to display",
                     params.yMin, -1000000.0, 1000000.0);
      opts.addOption("maxvalue", "ymax",
                     "maximum value to display. The range is detected "
                     "automatically if\n"
                     "        it is set to 0 to 0",
                     params.yMax, -1000000.0, 1000000.0);
      opts.addOption("log-x-scale", "logx",
                     "use logarithmic X (frequency) scale",
                     params.logFreqScale);
      opts.addOption("log-y-scale", "logy",
                     "use logarithmic Y (magnitude) scale if it is not in dB",
                     params.logYScale);
      opts.addOption("db-y-scale", "db",
                     "display magnitude in dB, instead of raw linear value",
                     params.dBScale);
      opts.addOption("level-1khz", "l1k",
                     "input level (<=0: dBr, >0: Vrms) at 1 kHz in level or "
                     "THD vs.\n"
                     "        level mode",
                     params.level1kHz, -140.0, 1000.0);
      opts.addOption("leveladjust", "a",
                     "adjust input level by the specified amount (in dB)",
                     params.lvlAdjust, -120.0, 120.0);
      if (!opts.parseCommandLine(argc, argv))
        return 0;
      for (size_t i = 0; i < analysisMode.length(); i++)
        analysisMode[i] = std::tolower(analysisMode[i]);
      if (analysisMode == "ir")
        params.mode = ANALYSIS_MODE_IR;
      else if (analysisMode == "tone")
        params.mode = ANALYSIS_MODE_TONE;
      else if (analysisMode == "freq")
        params.mode = ANALYSIS_MODE_FREQ;
      else if (analysisMode == "thdf")
        params.mode = ANALYSIS_MODE_THDFREQ;
      else if (analysisMode == "lvl")
        params.mode = ANALYSIS_MODE_LEVEL;
      else if (analysisMode == "thdl")
        params.mode = ANALYSIS_MODE_THDLEVEL;
      else if (analysisMode == "phs")
        params.mode = ANALYSIS_MODE_PHASE;
      else if (analysisMode == "pd")
        params.mode = ANALYSIS_MODE_PD;
      else if (analysisMode == "gd")
        params.mode = ANALYSIS_MODE_GD;
      else
        throw std::runtime_error(std::string("invalid analysis mode"));
      for (size_t i = 0; i < inFileNames.size(); i++)
      {
        if (inFileNames[i].empty())
          descriptionStrings[i].clear();
        else if (descriptionStrings[i].empty())
          descriptionStrings[i] = inFileNames[i];
        // leave 12 characters of space for the sample rate (" (xxxxxx Hz)")
        size_t  maxLen = size_t((imageWidth >> 4)
                                - (params.mode == ANALYSIS_MODE_IR ||
                                   params.mode == ANALYSIS_MODE_TONE ?
                                   22 : 10));
        if (descriptionStrings[i].length() > maxLen)
          descriptionStrings[i].resize(maxLen);
      }
      if (titleString.empty())
        titleString = analysisModeNames[params.mode];
      else if (titleString.length() > size_t((imageWidth >> 3) - 2))
        titleString.resize(size_t((imageWidth >> 3) - 2));
      if (params.mode == ANALYSIS_MODE_PHASE ||
          params.mode == ANALYSIS_MODE_PD || params.mode == ANALYSIS_MODE_GD)
      {
        params.logYScale = false;
        params.dBScale = false;
        params.octBWMult = 0.0;
      }
      if (params.logFreqScale && !(params.freqMin >= 0.001))
        params.freqMin = 10.0;
      if (!(params.freqMax == 0.0) &&
          !((!params.logFreqScale &&
             params.freqMax >= (params.freqMin + 0.5)) ||
            (params.logFreqScale &&
             params.freqMax >= (params.freqMin * 1.001))))
      {
        throw std::runtime_error(std::string("invalid frequency range"));
      }
      if (!(params.yMin == 0.0 && params.yMax == 0.0))
      {
        if (!(params.yMax >= (params.yMin + 0.0000005)) ||
            (params.dBScale &&
             !(params.yMin >= -400.0 && params.yMax <= 400.0)))
        {
          throw std::runtime_error(std::string("invalid Y range"));
        }
      }
    }

    std::vector< std::vector< double > >  graphData;
    std::vector< size_t >         graphFileNumbers;
    std::vector< unsigned char >  graphColors;
    {
      graphData.reserve(inFileNames.size() << 1);
      for (size_t i = 0; i < inFileNames.size(); i++)
      {
        if (inFileNames[i].empty())
          continue;
        AudioFilter_FileInput f(inFileNames[i].c_str());
        if (f.getChannelCount() > 2)
        {
          int     fileCount = (f.getChannelCount() + 1) >> 1;
          for (size_t j = i + 1; j < (i + size_t(fileCount)); j++)
          {
            if (j >= inFileNames.size() || !inFileNames[j].empty())
            {
              throw std::runtime_error(std::string("could not allocate "
                                                   "graph colors for file '")
                                       + inFileNames[i] + "'");
            }
          }
        }
        if (graphColors.size() == 0 && params.freqMax == 0.0)
          params.freqMax = f.getSampleRate() * 0.5;
        if (params.mode == ANALYSIS_MODE_IR ||
            params.mode == ANALYSIS_MODE_TONE)
        {
          char    strBuf[64];
          std::sprintf(strBuf, " (%g Hz)", f.getSampleRate());
          descriptionStrings[i] += strBuf;
        }
        for (int j = 0; j < f.getChannelCount(); j++)
        {
          graphData.push_back(std::vector< double >());
          graphData[graphData.size() - 1].resize(size_t(displayWidth * 2), 0.0);
          graphFileNumbers.push_back(i);
          graphColors.push_back((unsigned char)
                                (graphColorTable[int(i << 1) + j] - 1));
        }
        analyzeFile(&(graphData.front())
                    + (graphData.size() - size_t(f.getChannelCount())),
                    f, params);
      }
      if (graphColors.size() < 1)
        throw std::runtime_error(std::string("missing input file name"));
    }

    Image   img(imageWidth, imageHeight);
    if (blackBackground)
    {
      img.setColor(0, 0, 0, 0);
      img.setColor(1, 255, 255, 255);
      img.setColor(2, 176, 176, 176);
      img.setColor(3, 96, 96, 96);
    }
    else
    {
      img.setColor(0, 255, 255, 255);
      img.setColor(1, 0, 0, 0);
      img.setColor(2, 152, 152, 152);
      img.setColor(3, 200, 200, 200);
    }
    for (int i = 4; i <= 255; i++)
    {
      double  r = (double((i - 4) / 42) / 5.0) * 255.0 + 0.5;
      double  g = (double(((i - 4) / 6) % 7) / 6.0) * 255.0 + 0.5;
      double  b = (double((i - 4) % 6) / 5.0) * 255.0 + 0.5;
      img.setColor(i, (unsigned char) r, (unsigned char) g, (unsigned char) b);
    }

    params.logXScale = params.logFreqScale;
    if (params.mode != ANALYSIS_MODE_LEVEL &&
        params.mode != ANALYSIS_MODE_THDLEVEL)
    {
      params.xMin = params.freqMin;
      params.xMax = params.freqMax;
      img.drawText(displayOriginX + (displayWidth >> 1) - (7 * 8),
                   displayOriginY + 24, 1, "Frequency (Hz)", true);
    }
    else if (params.level1kHz <= 0.0)
    {
      if (!params.logXScale)
      {
        double  tmp = std::pow(10.0, params.level1kHz / 20.0);
        params.xMin = tmp * params.freqMin / 1000.0;
        params.xMax = tmp * params.freqMax / 1000.0;
        img.drawText(displayOriginX + (displayWidth >> 1) - (2 * 8),
                     displayOriginY + 24, 1, "Level", true);
      }
      else
      {
        params.logXScale = false;
        params.xMin = params.level1kHz
                      + (20.0 * std::log10(params.freqMin / 1000.0));
        params.xMax = params.level1kHz
                      + (20.0 * std::log10(params.freqMax / 1000.0));
        img.drawText(displayOriginX + (displayWidth >> 1) - (5 * 8),
                     displayOriginY + 24, 1, "Level (dBr)", true);
      }
    }
    else
    {
      params.xMin = params.level1kHz * params.freqMin / 1000.0;
      params.xMax = params.level1kHz * params.freqMax / 1000.0;
      img.drawText(displayOriginX + (displayWidth >> 1) - (6 * 8),
                   displayOriginY + 24, 1, "Level (Vrms)", true);
    }
    if (params.dBScale)
    {
      params.logYScale = false;
      img.drawText(displayOriginX + displayWidth + 24,
                   displayOriginY + 24, 1,
                   (params.mode != ANALYSIS_MODE_TONE ||
                    params.lvlAdjust != 0.0 ? "dBr" : "dBFS"), true);
      double  maxVal = -400.0;
      for (size_t i = 0; i < graphData.size(); i++)
      {
        for (int j = 0; j < (displayWidth * 2); j++)
        {
          double  y = graphData[i][j];
          y = 20.0 * std::log10(y > 1.0e-20 ? y : 1.0e-20);
          graphData[i][j] = y;
          maxVal = (y > maxVal ? y : maxVal);
        }
      }
      if (params.yMin == 0.0 && params.yMax == 0.0)
      {
        params.yMax = std::ceil(maxVal + 1.0);
        params.yMin = params.yMax - 50.0;
      }
    }
    else
    {
      const char  *yLabel = (char *) 0;
      switch (params.mode)
      {
      case ANALYSIS_MODE_THDFREQ:
      case ANALYSIS_MODE_THDLEVEL:
        yLabel = "% THD";
        break;
      case ANALYSIS_MODE_PHASE:
        yLabel = "degrees";
        break;
      case ANALYSIS_MODE_PD:
      case ANALYSIS_MODE_GD:
        yLabel = "ms delay";
        break;
      }
      if (yLabel)
      {
        img.drawText(displayOriginX + displayWidth + 40
                     - (int(std::strlen(yLabel) >> 1) << 3),
                     displayOriginY + 24, 1, yLabel, true);
      }
      if (params.yMax == 0.0 && (params.logYScale || params.yMin == 0.0))
      {
        double  minVal = graphData[0][0];
        double  maxVal = graphData[0][1];
        for (size_t i = 0; i < graphData.size(); i++)
        {
          for (int j = 0; j < (displayWidth * 2); j++)
          {
            double  y = graphData[i][j];
            minVal = (y < minVal ? y : minVal);
            maxVal = (y > maxVal ? y : maxVal);
          }
        }
        if (!(maxVal >= (minVal + 1.0e-10)))
        {
          minVal = ((minVal + maxVal) - 1.0e-10) * 0.5;
          maxVal = minVal + 1.0e-10;
        }
        if (!params.logYScale)
        {
          double  offs = (maxVal - minVal) / 10.0;
          minVal -= offs;
          maxVal += offs;
          if (minVal < 0.0 && params.mode != ANALYSIS_MODE_PHASE &&
              params.mode != ANALYSIS_MODE_PD &&
              params.mode != ANALYSIS_MODE_GD)
          {
            minVal = 0.0;
          }
        }
        else
        {
          minVal = (minVal > (maxVal / 316.228) ? minVal : (maxVal / 316.228))
                   / 1.125;
          maxVal = maxVal * 1.125;
        }
        params.yMin =
            (params.yMin == 0.0 || minVal < params.yMin ? minVal : params.yMin);
        params.yMax = maxVal;
      }
      if (params.logYScale && !(params.yMin >= (params.yMax * 1.0e-20)))
        params.yMin = params.yMax * 0.00316227766;
    }

    {
      char    strBuf[256];
      std::vector< int >  xTickPos1;
      std::vector< int >  xTickPos2;
      std::vector< int >  xTickPos3;
      std::vector< int >  yTickPos1;
      std::vector< int >  yTickPos2;
      std::vector< int >  yTickPos3;
      if (!params.logXScale)
      {
        int     logXRange =
            int(std::floor(std::log10(params.xMax - params.xMin) + 0.9995));
        bool    lowRes = (std::fabs(std::log10(params.xMax - params.xMin)
                                    - double(logXRange)) >= 0.5);
        double  xRange = std::pow(10.0, double(logXRange));
        if (logXRange >= 0)
          xRange = std::floor(xRange + 0.5);
        double  x0 = std::floor(params.xMin / xRange) * xRange;
        for (int i = 0; i <= 400; i += (lowRes ? 1 : 2))
        {
          double  f = x0 + (double(i) * xRange / 200.0);
          int     x = int(std::floor(double(displayWidth - 1)
                                     * (f - params.xMin)
                                     / (params.xMax - params.xMin)
                                     + 0.5)) + displayOriginX;
          if (x < displayOriginX)
            continue;
          if (x >= (displayOriginX + displayWidth))
            break;
          if ((i & 1) || (!lowRes && (i % 10)))
            xTickPos3.push_back(x);
          else if (i % 20)
            xTickPos2.push_back(x);
          else
            xTickPos1.push_back(x);
          if (!((i % 10) == 0 || ((i & 1) == 0 && lowRes)))
            continue;
          if (lowRes || (i % 20) == 0)
            printValue(img, x - 8, displayOriginY + 8, f);
          img.drawLine(x, displayOriginY + 1, x, displayOriginY + 5, 1);
          img.drawLine(x, displayOriginY - displayHeight,
                       x, displayOriginY - displayHeight - 4, 1);
        }
      }
      else
      {
        bool    lowRes = (params.xMax <= (params.xMin * 10.0));
        for (long long i = 1LL; i <= 1000000000000LL; i *= 10LL)
        {
          for (int j = 2; j <= 20; j += (j < 4 ? 1 : 2))
          {
            double  f = double(i) * double(j) / 2000000.0;
            int     x = int(std::floor(double(displayWidth - 1)
                                       * std::log(f / params.xMin)
                                       / std::log(params.xMax / params.xMin)
                                       + 0.5)) + displayOriginX;
            if (x < displayOriginX)
              continue;
            if (x >= (displayOriginX + displayWidth))
              break;
            if (j == 2)
              xTickPos1.push_back(x);
            else if (j == 4 || j == 10)
              xTickPos2.push_back(x);
            else
              xTickPos3.push_back(x);
            if (j == 2 || j == 4 || j == 10 || lowRes)
              printValue(img, x - 8, displayOriginY + 8, f);
            img.drawLine(x, displayOriginY + 1, x, displayOriginY + 5, 1);
            img.drawLine(x, displayOriginY - displayHeight,
                         x, displayOriginY - displayHeight - 4, 1);
          }
        }
      }
      if (!params.logYScale)
      {
        int     logYRange =
            int(std::floor(std::log10(params.yMax - params.yMin) + 0.9995));
        bool    lowRes = (std::fabs(std::log10(params.yMax - params.yMin)
                                    - double(logYRange)) >= 0.5);
        double  yRange = std::pow(10.0, double(logYRange));
        if (logYRange >= 0)
          yRange = std::floor(yRange + 0.5);
        double  y0 = std::floor(params.yMin / yRange) * yRange;
        for (int i = 0; i <= 400; i += (lowRes ? 1 : 2))
        {
          double  f = y0 + (double(i) * yRange / 200.0);
          int     y = displayOriginY
                      - int(std::floor(double(displayHeight - 1)
                                       * (f - params.yMin)
                                       / (params.yMax - params.yMin) + 0.5));
          if (y > displayOriginY)
            continue;
          if (y <= (displayOriginY - displayHeight))
            break;
          if ((i & 1) || (!lowRes && (i % 10)))
            yTickPos3.push_back(y);
          else if (i % 20)
            yTickPos2.push_back(y);
          else
            yTickPos1.push_back(y);
          if (!((i % 10) == 0 || ((i & 1) == 0 && lowRes)))
            continue;
          std::sprintf(strBuf, "%g", f);
          img.drawText(displayOriginX + displayWidth + 8, y - 6, 1, strBuf);
          img.drawLine(displayOriginX - 1, y, displayOriginX - 5, y, 1);
          img.drawLine(displayOriginX + displayWidth, y,
                       displayOriginX + displayWidth + 4, y, 1);
        }
      }
      else
      {
        bool    lowRes = (params.yMax <= (params.yMin * 10.0));
        for (long long i = 1LL; i <= 1000000000000LL; i *= 10LL)
        {
          for (int j = 2; j <= 20; j += (j < 4 ? 1 : 2))
          {
            double  f = double(i) * double(j) / 2000000.0;
            int     y = displayOriginY
                        - int(std::floor(double(displayHeight - 1)
                                         * std::log(f / params.yMin)
                                         / std::log(params.yMax / params.yMin)
                                         + 0.5));
            if (y > displayOriginY)
              continue;
            if (y <= (displayOriginY - displayHeight))
              break;
            if (j == 2)
              yTickPos1.push_back(y);
            else if (j == 4 || j == 10)
              yTickPos2.push_back(y);
            else
              yTickPos3.push_back(y);
            if (j == 2 || j == 4 || j == 10 || lowRes)
            {
              std::sprintf(strBuf, "%g", f);
              img.drawText(displayOriginX + displayWidth + 8, y - 6, 1, strBuf);
            }
            img.drawLine(displayOriginX - 1, y, displayOriginX - 5, y, 1);
            img.drawLine(displayOriginX + displayWidth, y,
                         displayOriginX + displayWidth + 4, y, 1);
          }
        }
      }
      for (size_t i = 0; i < xTickPos3.size(); i++)
      {
        for (int y = displayOriginY - (displayHeight - 1);
             y <= displayOriginY;
             y++)
        {
          if ((xTickPos3[i] ^ y) & 1)
            img.setPixel(xTickPos3[i], y, 3);
        }
      }
      for (size_t i = 0; i < yTickPos3.size(); i++)
      {
        for (int x = displayOriginX;
             x <= displayOriginX + (displayWidth - 1);
             x++)
        {
          if ((yTickPos3[i] ^ x) & 1)
            img.setPixel(x, yTickPos3[i], 3);
        }
      }
      for (size_t i = 0; i < xTickPos2.size(); i++)
      {
        img.drawLine(xTickPos2[i], displayOriginY,
                     xTickPos2[i], displayOriginY - (displayHeight - 1), 3);
      }
      for (size_t i = 0; i < yTickPos2.size(); i++)
      {
        img.drawLine(displayOriginX, yTickPos2[i],
                     displayOriginX + (displayWidth - 1), yTickPos2[i], 3);
      }
      for (size_t i = 0; i < xTickPos1.size(); i++)
      {
        img.drawLine(xTickPos1[i], displayOriginY,
                     xTickPos1[i], displayOriginY - (displayHeight - 1), 2);
      }
      for (size_t i = 0; i < yTickPos1.size(); i++)
      {
        img.drawLine(displayOriginX, yTickPos1[i],
                     displayOriginX + (displayWidth - 1), yTickPos1[i], 2);
      }
    }
    img.drawRectangle(displayOriginX - 1, displayOriginY + 1,
                      displayOriginX + displayWidth,
                      displayOriginY - displayHeight, 1);
    img.drawText(8, 4, 1, titleString.c_str(), true);
    {
      char    strBuf[256];
      std::string paramsString;
      if (params.mode == ANALYSIS_MODE_IR || params.mode == ANALYSIS_MODE_TONE)
      {
        if (params.windowType >= 0 && params.windowType < 10)
        {
          paramsString = "Window: ";
          paramsString += windowNames[params.windowType];
        }
        else
        {
          std::sprintf(strBuf, "Window: Gaussian (%d dB%s)",
                       std::abs(params.windowType),
                       (params.windowType < 0 ? " normalized" : ""));
          paramsString = strBuf;
        }
        if (params.offs < 0.0)
          std::sprintf(strBuf, ", centered at %.2f smps", -(params.offs));
        else
          std::sprintf(strBuf, ", centered at %.6f secs", params.offs);
        paramsString += strBuf;
        if (params.duration < 0.0)
          std::sprintf(strBuf, ", length = %.2f smps", -(params.duration));
        else
          std::sprintf(strBuf, ", length = %.6f secs", params.duration);
        paramsString += strBuf;
      }
      else
      {
        paramsString = "FFT size = 65536 samples";
      }
      if (params.octBWMult > 0.0)
      {
        std::sprintf(strBuf, "    BWMult: %.4f", params.octBWMult);
        paramsString += strBuf;
      }
      if (paramsString.length() > size_t((imageWidth >> 3) - 2))
        paramsString.resize(size_t((imageWidth >> 3) - 2));
      img.drawText(8, 18, 2, paramsString.c_str());
    }
    {
      int     n = -1;
      int     c = 0;
      for (size_t i = 0; i < graphColors.size(); i++)
      {
        bool    printDesc = false;
        if (i == 0 || graphFileNumbers[i] != graphFileNumbers[i - 1])
        {
          printDesc = true;
          n++;
          c = 0;
        }
        else if (++c >= 2)
        {
          n++;
          c = 0;
        }
        int     x = (imageWidth >> 1) * (n / 4) + 8;
        int     y = displayOriginY + 44 + ((n % 4) * 16);
        img.drawRectangle(x + (c * 32), y + 1, x + (c * 32) + 24, y + 11, 1);
        for (int j = 2; j < 11; j++)
        {
          for (int k = 1; k < 24; k++)
          {
            drawRGBAPixel(img, x + (c * 32) + k, y + j,
                          colorTable[graphColors[i] * 3],
                          colorTable[graphColors[i] * 3 + 1],
                          colorTable[graphColors[i] * 3 + 2], 1.0f);
          }
        }
        if (printDesc)
        {
          img.drawText(x + 64, y, 1,
                       descriptionStrings[graphFileNumbers[i]].c_str());
        }
      }
    }

    drawGraphs(img, graphData, graphColors, params);

    img.writeBMPFile(outFileName.c_str());
  }
  catch (std::exception& e)
  {
    std::fprintf(stderr, "%s: %s\n", argv[0], e.what());
    return -1;
  }
  return 0;
}

