
#include "headers.hpp"
#include "dsplib.cpp"
#include "options.cpp"

#define LEVEL_ADJUST_BASE       -8.9124308
#define FFT_SIZE                65536

struct AnalysisParameters
{
  double  bandWidth;
  int     mode;                 // 0: group delay, 1: phase delay, 2: phase
  int     winParam;
  bool    normalizePhase;
  double  duration;
  double  offs;
  double  scale;
  double  lvlAdjust;            // in dB
  std::vector< double > phaseOffsets;
};

static void analyzePhase(std::vector< std::vector< double > >& fftBufs,
                         AudioFilter_FileInput& sf,
                         const AnalysisParameters& params)
{
  double  sampleRate = sf.getSampleRate();
  size_t  channels = size_t(sf.getChannelCount());
  double  winStart = (params.offs - (params.duration * 0.5)) * sampleRate;
  double  winEnd = (params.offs + (params.duration * 0.5)) * sampleRate;
  double  windowSamples = params.duration * sampleRate;
  long    winStart_i = long(std::ceil(winStart));
  long    winEnd_i = long(std::floor(winEnd)) + 1L;
  long    windowSamples_i = winEnd_i - winStart_i;
  size_t  fftSize = FFT_SIZE;
  while (fftSize < size_t(windowSamples_i))
    fftSize = fftSize << 1;
  fftBufs.clear();
  fftBufs.resize(channels);
  for (size_t i = 0; i < channels; i++)
    fftBufs[i].resize(fftSize, 0.0);
  sf.setPosition(sf_count_t(winStart_i), false);
  if (params.winParam < 0)
  {
    GaussianWindowNormalized  w(long(windowSamples + 0.5) + 1L,
                                std::pow(10.0, double(params.winParam) * 0.05));
    w.setPosition(double(winStart_i) - (params.offs * sampleRate));
    for (long i = 0L; i < windowSamples_i; i++)
    {
      sf.process();
      double  w_ = w.sample();
      for (size_t j = 0; j < channels; j++)
        fftBufs[j][i] = sf.getOutputBuffer()[j] * w_;
    }
  }
  else
  {
    CosineGen w(windowSamples,
                (double(winStart_i) - winStart) * twoPi / windowSamples);
    for (long i = 0L; i < windowSamples_i; i++)
    {
      sf.process();
      double  w_ = cosineToNuttallWindow(w.sample());
      for (size_t j = 0; j < channels; j++)
        fftBufs[j][i] = sf.getOutputBuffer()[j] * w_;
    }
  }
  for (size_t i = 0; i < channels; i++)
  {
    long double phaseOffsInc = ((long double) params.offs * sampleRate
                                - (long double) winStart_i)
                               * twoPiL / (long double) long(fftSize);
    realFFT(&(fftBufs[i].front()), fftSize, false);
    if (params.normalizePhase)
    {
      // find phase delay at the maximum frequency
      long double phaseOffs = 0.0L;
      long double prvPhase = 0.0L;
      if (i < params.phaseOffsets.size())
        prvPhase = phaseOffs = piL * params.phaseOffsets[i] / 180.0L;
      double  dummy_im = 0.0;
      for (size_t j = 0; true; j++)
      {
        double  f = double(long(j)) * sampleRate / double(long(fftSize));
        if (j > (fftSize >> 1) || f > params.bandWidth)
        {
          if (j-- > 1)
          {
            std::printf("Channel #%d: adjusted time offset by %.6f us\n",
                        int(i + 1),
                        1000000.0 * double(prvPhase) * double(long(fftSize))
                        / (twoPi * double(long(j)) * sampleRate));
            phaseOffsInc -= (prvPhase / (long double) long(j));
          }
          break;
        }
        double& re = fftBufs[i][(j < (fftSize >> 1) ? (j << 1) : 1)];
        double& im =
            (j > 0 && j < (fftSize >> 1) ? fftBufs[i][(j << 1) + 1] : dummy_im);
        // calculate phase and add offset
        long double phase = std::atan2((long double) im, (long double) re)
                            + phaseOffs;
        phaseOffs += phaseOffsInc;
        // unwrap phase
        if (j > 0)
        {
          long double tmp = 0.0L;
          (void) std::modf((prvPhase - phase) / twoPiL, &tmp);
          phase += (tmp * twoPiL);
          if ((phase - prvPhase) > (piL * 1.000000000001L))
            phase -= twoPiL;
          else if ((phase - prvPhase) < (piL * -1.000000000001L))
            phase += twoPiL;
        }
        prvPhase = phase;
      }
    }
    long double phaseOffs = 0.0L;
    long double prvPhase = 0.0L;
    if (i < params.phaseOffsets.size())
      prvPhase = phaseOffs = piL * params.phaseOffsets[i] / 180.0L;
    double  dummy_im = 0.0;
    for (size_t j = 0; j <= (fftSize >> 1); j++)
    {
      double& re = fftBufs[i][(j < (fftSize >> 1) ? (j << 1) : 1)];
      double& im =
          (j > 0 && j < (fftSize >> 1) ? fftBufs[i][(j << 1) + 1] : dummy_im);
      double  f = double(long(j)) * sampleRate / double(long(fftSize));
      double  result = 0.0;
      if (f <= params.bandWidth)
      {
        // calculate phase and add offset
        long double phase = std::atan2((long double) im, (long double) re)
                            + phaseOffs;
        phaseOffs += phaseOffsInc;
        // unwrap phase
        if (j > 0)
        {
          long double tmp = 0.0L;
          (void) std::modf((prvPhase - phase) / twoPiL, &tmp);
          phase += (tmp * twoPiL);
          if ((phase - prvPhase) > (piL * 1.000000000001L))
            phase -= twoPiL;
          else if ((phase - prvPhase) < (piL * -1.000000000001L))
            phase += twoPiL;
        }
        if (params.mode == 0)           // group delay
        {
          result = -1000.0 * double(phase - prvPhase)
                   / (twoPi * sampleRate / double(long(fftSize)));
        }
        else if (params.mode == 1)      // phase delay
        {
          result = -1000.0 * double(phase) / (twoPi * f);
        }
        else                            // phase
        {
          result = 18.0 * double(phase) / pi;
        }
        result *= params.scale;
        result = (result > -72.0 ? (result < 72.0 ? result : 72.0) : -72.0);
        prvPhase = phase;
      }
      re = std::pow(10.0,
                    (result + params.lvlAdjust + LEVEL_ADJUST_BASE) * 0.05)
           * double(FFT_SIZE >> 1);
      im = 0.0;
      if (j & 1)
        re = -re;
    }
    realFFT(&(fftBufs[i].front()), fftSize, true);
  }
}

// ----------------------------------------------------------------------------

int main(int argc, char **argv)
{
  try
  {
    AnalysisParameters  params;
    params.bandWidth = 500000.0;
    params.mode = 0;
    params.winParam = 0;
    params.normalizePhase = false;
    params.duration = 1.0;
    params.offs = 0.0;
    params.scale = 1.0;
    params.lvlAdjust = 0.0;
    params.phaseOffsets.resize(8, 0.0);
    std::string inFileName;
    std::string outFileName;
    {
      CommandLineParser opts(argv[0]);
      opts.addFileName("INFILE", "input file name", inFileName);
      opts.addFileName("OUTFILE", "output file name", outFileName);
      opts.addOption("infile", "i",
                     "input file name",
                     inFileName);
      opts.addOption("outfile", "o",
                     "output file name",
                     outFileName);
      opts.addOption("mode", "m",
                     "analysis mode: 0: group delay, 1: phase delay, 2: phase",
                     params.mode, 0, 2);
      opts.addOption("bandwidth", "bw", "bandwidth limit in Hz",
                     params.bandWidth, 500.0, 500000.0);
      opts.addOption("offset", "t",
                     "assumed delay (in seconds or -samples) of impulse",
                     params.offs, -1000000000.0, 10000.0);
      opts.addOption("normalize", "n",
                     "automatically adjust delay for zero phase delay at "
                     "maximum freq.",
                     params.normalizePhase);
      opts.addOption("duration", "d", "window length in seconds or -samples",
                     params.duration, -4194304.0, 30.0);
      opts.addOption("window", "w",
                     "window type (0: Nuttall window, <0: Gaussian window "
                     "with a\n"
                     "        minimum value of the specified number of dB, "
                     "normalized to the\n"
                     "        range 0 to 1)",
                     params.winParam, -240, 0);
      opts.addOption("phase-offset-1", "p1", "",
                     params.phaseOffsets[0], -1440.0, 1440.0);
      opts.addOption("phase-offset-2", "p2", "",
                     params.phaseOffsets[1], -1440.0, 1440.0);
      opts.addOption("phase-offset-3", "p3", "",
                     params.phaseOffsets[2], -1440.0, 1440.0);
      opts.addOption("phase-offset-4", "p4", "",
                     params.phaseOffsets[3], -1440.0, 1440.0);
      opts.addOption("phase-offset-5", "p5", "",
                     params.phaseOffsets[4], -1440.0, 1440.0);
      opts.addOption("phase-offset-6", "p6", "",
                     params.phaseOffsets[5], -1440.0, 1440.0);
      opts.addOption("phase-offset-7", "p7", "",
                     params.phaseOffsets[6], -1440.0, 1440.0);
      opts.addOption("phase-offset-8", "p8",
                     "phase offset (in degrees) for channel 1 to 8",
                     params.phaseOffsets[7], -1440.0, 1440.0);
      opts.addOption("scale", "s", "scale dB output by this factor",
                     params.scale, 0.001, 1000.0);
      opts.addOption("leveladjust", "a",
                     "adjust output level by the specified amount (in dB)",
                     params.lvlAdjust, -80.0, 80.0);
      if (!opts.parseCommandLine(argc, argv))
        return 0;
    }

    // analyze input file
    std::vector< std::vector< double > >  fftBufs;
    AudioFilter_FileInput  inFile(inFileName.c_str());
    if (params.duration < 0.0)
      params.duration /= -(inFile.getSampleRate());
    if (!(params.duration >= 0.0025))
      params.duration = 0.0025;
    if (params.offs < 0.0)
      params.offs /= -(inFile.getSampleRate());
    analyzePhase(fftBufs, inFile, params);

    // write output file
    {
      AudioOutputFile   outFile(outFileName.c_str(), inFile.getSampleRate(),
                                inFile.getChannelCount());
      std::vector< double > buf(size_t(inFile.getChannelCount()), 0.0);
      for (size_t i = 0; i < FFT_SIZE; i++)
      {
        for (size_t j = 0; j < buf.size(); j++)
          buf[j] = fftBufs[j][((fftBufs[j].size() - FFT_SIZE) >> 1) + i];
        outFile.writeSampleFrame(&(buf.front()));
      }
    }
  }
  catch (std::exception& e)
  {
    std::fprintf(stderr, "%s: %s\n", argv[0], e.what());
    return -1;
  }
  return 0;
}

