
#ifndef IMAGE_CPP_INCLUDED
#define IMAGE_CPP_INCLUDED

#include "headers.hpp"

class Image
{
 protected:
  static const unsigned char    fontData[1140];
  static const unsigned char    fontDataBold[1140];
  static const int              ditherSize = 16;
  std::vector< unsigned char >  buf;
  std::vector< unsigned char >  colormap;
  int     width;
  int     height;
  std::vector< float >          ditherTable;
  // --------
  static inline void writeWord(std::vector< unsigned char >& buf_, size_t pos,
                               unsigned short n)
  {
    buf_[pos] = (unsigned char) (n & 0xFF);
    buf_[pos + 1] = (unsigned char) ((n >> 8) & 0xFF);
  }
  static inline void writeDWord(std::vector< unsigned char >& buf_, size_t pos,
                                unsigned int n)
  {
    buf_[pos] = (unsigned char) (n & 0xFFU);
    buf_[pos + 1] = (unsigned char) ((n >> 8) & 0xFFU);
    buf_[pos + 2] = (unsigned char) ((n >> 16) & 0xFFU);
    buf_[pos + 3] = (unsigned char) ((n >> 24) & 0xFFU);
  }
  void createDitherTable(int n = 0, int x = 0, int y = 0, float offs = 0.0f);
 public:
  Image(int w, int h);
  virtual ~Image();
  void setColor(int c, unsigned char r, unsigned char g, unsigned char b);
  inline void getColor(unsigned char& r, unsigned char& g, unsigned char& b,
                       int c) const
  {
    const unsigned char *cp = &(colormap.front()) + ((c & 0xFF) * 3);
    r = cp[0];
    g = cp[1];
    b = cp[2];
  }
  inline void setPixel(int x, int y, unsigned char c)
  {
    if (x < 0 || x >= width || y < 0 || y >= height)
      return;
    buf[(y * width) + x] = c;
  }
  inline int getPixel(int x, int y) const
  {
    if (x < 0 || x >= width || y < 0 || y >= height)
      return -1;
    return buf[(y * width) + x];
  }
  void clear(unsigned char c = 0x00);
  void drawLine(int x0, int y0, int x1, int y1, unsigned char c);
  void drawRectangle(int x0, int y0, int x1, int y1, unsigned char c,
                     bool isFilled = false);
  void drawText(int x0, int y0, unsigned char c, const char *s,
                bool isBold = false, bool doubleSize = false);
  void writeBMPFile(const char *fileName) const;
  inline float getDitherOffset(int x, int y) const
  {
    return ditherTable[((y & (ditherSize - 1)) * ditherSize)
                       + (x & (ditherSize - 1))];
  }
};

#include "fontdata.cpp"

void Image::createDitherTable(int n, int x, int y, float offs)
{
  if (n == 1)
  {
    ditherTable[y * ditherSize + x] = offs;
    return;
  }
  if (n < 1)
  {
    ditherTable.resize(size_t(ditherSize * ditherSize), 0.0f);
    n = ditherSize;
  }
  int     nD2 = n >> 1;
  float   offs_ = 1.0f / float(n * n);
  createDitherTable(nD2, x, y, offs);
  createDitherTable(nD2, x + nD2, y + nD2, offs + offs_);
  createDitherTable(nD2, x, y + nD2, offs + (offs_ * 2.0f));
  createDitherTable(nD2, x + nD2, y, offs + (offs_ * 3.0f));
}

Image::Image(int w, int h)
{
  if (w < 1 || w >= 32768 || h < 1 || h >= 32768)
    throw std::runtime_error(std::string("invalid image dimensions"));
  width = w;
  height = h;
  buf.resize(size_t(w * h), 0x00);
  colormap.resize(256 * 3, 0x00);
  // default colormap is greyscale
  for (int i = 0; i < 256; i++)
  {
    colormap[i * 3] = (unsigned char) i;
    colormap[(i * 3) + 1] = (unsigned char) i;
    colormap[(i * 3) + 2] = (unsigned char) i;
  }
  createDitherTable();
}

Image::~Image()
{
}

void Image::setColor(int c, unsigned char r, unsigned char g, unsigned char b)
{
  if (c < 0 || c >= 256)
    throw std::runtime_error(std::string("invalid colormap index"));
  colormap[c * 3] = r;
  colormap[(c * 3) + 1] = g;
  colormap[(c * 3) + 2] = b;
}

void Image::clear(unsigned char c)
{
  std::memset(&(buf.front()), c, buf.size());
}

void Image::drawLine(int x0, int y0, int x1, int y1, unsigned char c)
{
  int     d1 = std::abs(x1 - x0);
  int     d2 = std::abs(y1 - y0);
  int     x = x0;
  int     y = y0;
  int&    c1 = (d1 >= d2 ? x : y);
  int&    c2 = (d1 >= d2 ? y : x);
  int     c1i = (x1 > x0 ? 1 : -1);
  int     c2i = (y1 > y0 ? 1 : -1);
  if (d1 < d2)
  {
    int     tmp = d1;
    d1 = d2;
    d2 = tmp;
    tmp = c1i;
    c1i = c2i;
    c2i = tmp;
  }
  int     d = d1 >> 1;
  int     n = d1 + 1;
  do
  {
    setPixel(x, y, c);
    c1 += c1i;
    d += d2;
    if (d >= d1)
    {
      d -= d1;
      c2 += c2i;
    }
  }
  while (--n);
}

void Image::drawRectangle(int x0, int y0, int x1, int y1, unsigned char c,
                          bool isFilled)
{
  if (x0 > x1)
  {
    int     tmp = x0;
    x0 = x1;
    x1 = tmp;
  }
  if (y0 > y1)
  {
    int     tmp = y0;
    y0 = y1;
    y1 = tmp;
  }
  if (x1 < 0 || x0 >= width || y1 < 0 || y0 >= height)
    return;
  if (isFilled || x0 == x1 || y0 == y1)
  {
    x0 = (x0 > 0 ? x0 : 0);
    x1 = (x1 < (width - 1) ? x1 : (width - 1));
    y0 = (y0 > 0 ? y0 : 0);
    y1 = (y1 < (height - 1) ? y1 : (height - 1));
    int     y = y0;
    do
    {
      std::memset(&(buf.front()) + ((y * width) + x0), c, size_t(x1 + 1 - x0));
    }
    while (++y <= y1);
  }
  else
  {
    for (int x = x0; x <= x1; x++)
    {
      setPixel(x, y0, c);
      setPixel(x, y1, c);
    }
    for (int y = y0 + 1; y < y1; y++)
    {
      setPixel(x0, y, c);
      setPixel(x1, y, c);
    }
  }
}

void Image::drawText(int x0, int y0, unsigned char c, const char *s,
                     bool isBold, bool doubleSize)
{
  y0++;
  if (!s || y0 >= height)
    return;
  while (x0 < width && *s != '\0')
  {
    int     n = (*(s++) & 0x7F) - 0x20;
    n = (n >= 0 && n < 95 ? n : 0) * 12;
    const unsigned char *fontp = (isBold ? &(fontDataBold[n]) : &(fontData[n]));
    int     y = y0;
    for (int i = 0; i < 12 && y < height; i++)
    {
      unsigned char b = fontp[i];
      if (b)
      {
        int     x = x0;
        for (int j = 0; j < 8 && x < width; j++)
        {
          if (b & 0x80)
          {
            setPixel(x, y, c);
            if (doubleSize)
            {
              setPixel(x + 1, y, c);
              setPixel(x, y + 1, c);
              setPixel(x + 1, y + 1, c);
            }
          }
          x += (doubleSize ? 2 : 1);
          b = (b & 0x7F) << 1;
        }
      }
      y += (doubleSize ? 2 : 1);
    }
    x0 += (doubleSize ? 16 : 8);
  }
}

void Image::writeBMPFile(const char *fileName) const
{
  if (!fileName || !fileName[0])
    throw std::runtime_error(std::string("invalid image file name"));
  std::vector< unsigned char >  hdr(0x0436, 0x00);
  hdr[0] = (unsigned char) 'B';
  hdr[1] = (unsigned char) 'M';
  writeDWord(hdr, 0x000A, (unsigned int) hdr.size());
  writeDWord(hdr, 0x000E, 40U);                 // bitmap header size
  writeDWord(hdr, 0x0012, (unsigned int) width);
  writeDWord(hdr, 0x0016, (unsigned int) height);
  writeWord(hdr, 0x001A, 1);                    // number of planes
  writeWord(hdr, 0x001C, 8);                    // bits per pixel
  writeDWord(hdr, 0x001E, 1U);                  // compression (RLE8)
  writeDWord(hdr, 0x0022, (unsigned int) (width * height));     // bitmap size
  writeDWord(hdr, 0x0026, 3780U);               // horizontal pixels/m (96 dpi)
  writeDWord(hdr, 0x002A, 3780U);               // vertical pixels/m (96 dpi)
  writeDWord(hdr, 0x002E, 256U);                // colors used
  writeDWord(hdr, 0x0032, 0U);                  // number of important colors
  for (int i = 0; i < 256; i++)
  {
    hdr[(i << 2) + 0x0036] = colormap[(i * 3) + 2];
    hdr[(i << 2) + 0x0037] = colormap[(i * 3) + 1];
    hdr[(i << 2) + 0x0038] = colormap[i * 3];
  }
  std::vector< unsigned char >  rleBuf;
  rleBuf.reserve(size_t((((width + 1) & ~1) + 4) * height + 2));
  std::vector< int >    lengths(size_t(width + 1));
  std::vector< size_t > dataSizes(size_t(width + 1));
  for (int y = height; y-- > 0; )
  {
    lengths[width] = 0;
    dataSizes[width] = 0;
    const unsigned char *lp = &(buf.front()) + (y * width);
    for (int x = width; x-- > 0; )
    {
      int     len = 1;
      size_t  size = dataSizes[x + 1] + 2;
      if ((x + 1) < width)
      {
        // test encoded runs
        if (lp[x] == lp[x + 1])
        {
          if (lengths[x + 1] > 0 && lengths[x + 1] < 255)
          {
            len = lengths[x + 1] + 1;
            size = dataSizes[x + 1];
          }
          else
          {
            len = 2;
            size = dataSizes[x + 2] + 2;
          }
        }
        if ((x + 2) < width && len < 2)
        {
          // test unencoded runs
          int     l = 3;
          if (lengths[x + 1] < 0 && lengths[x + 1] > -255)
            l = 1 - lengths[x + 1];
          do
          {
            size_t  size_ = dataSizes[x + l] + size_t(l + (l & 1) + 2);
            if (size_ < size)
            {
              len = -l;
              size = size_;
            }
            l++;
          }
          while (l <= 6 && (x + l) <= width);
        }
      }
      lengths[x] = len;
      dataSizes[x] = size;
    }
    for (int x = 0; x < width; )
    {
      if (lengths[x] > 0)
      {
        rleBuf.push_back((unsigned char) lengths[x]);
        rleBuf.push_back(lp[x]);
        x = x + lengths[x];
      }
      else
      {
        rleBuf.push_back(0x00);
        int     n = -(lengths[x]);
        rleBuf.push_back((unsigned char) n);
        do
        {
          rleBuf.push_back(lp[x]);
          if (n < 2)
          {
            rleBuf.push_back(0x00);
            x++;
            break;
          }
          rleBuf.push_back(lp[x + 1]);
          x += 2;
          n -= 2;
        }
        while (n > 0);
      }
    }
    rleBuf.push_back(0x00);                     // end of scan line
    rleBuf.push_back(0x00);
  }
  rleBuf.push_back(0x00);                       // end of bitmap data
  rleBuf.push_back(0x01);
  writeDWord(hdr, 0x0002, (unsigned int) (hdr.size() + rleBuf.size()));
  std::FILE *f = std::fopen(fileName, "wb");
  if (!f)
  {
    throw std::runtime_error(std::string("error opening image file '")
                             + fileName + "'");
  }
  try
  {
    if (std::fwrite(&(hdr.front()), sizeof(unsigned char), hdr.size(), f)
        != hdr.size() ||
        std::fwrite(&(rleBuf.front()), sizeof(unsigned char), rleBuf.size(), f)
        != rleBuf.size() ||
        std::fflush(f) != 0)
    {
      throw std::runtime_error(std::string("error writing image file '")
                               + fileName + "'");
    }
  }
  catch (...)
  {
    std::fclose(f);
    std::remove(fileName);
    throw;
  }
  std::fclose(f);
}

#endif  // IMAGE_CPP_INCLUDED

