
#include "headers.hpp"
#include "dsplib.cpp"
#include "options.cpp"

struct AnalysisParameters
{
  double  sampleRateMult;
  double  noiseFreqMin;
  double  noiseFreqMax;
  double  signalBandwidthBins;
  double  lvlAdjust;
  double  fullScaleVoltage;
  int     winParam;
  std::vector< double > signalFrequencies;
};

struct AnalysisResults
{
  double  dcLevel;
  double  signalLevel;
  double  signalAWeighted;
  double  signalR468Weighted;
  double  noiseLevel;
  double  noiseAWeighted;
  double  noiseR468Weighted;
  double  peakMin;
  double  peakMax;
  // --------
  void clear()
  {
    dcLevel = 0.0;
    signalLevel = 0.0;
    signalAWeighted = 0.0;
    signalR468Weighted = 0.0;
    noiseLevel = 0.0;
    noiseAWeighted = 0.0;
    noiseR468Weighted = 0.0;
    peakMin = 0.0;
    peakMax = 0.0;
  }
  AnalysisResults()
  {
    clear();
  }
};

// ----------------------------------------------------------------------------

struct Complex
{
  double  re;
  double  im;
  // --------
  Complex()
    : re(0.0),
      im(0.0)
  {
  }
  Complex(double re_, double im_)
    : re(re_),
      im(im_)
  {
  }
  Complex(const Complex& x)
    : re(x.re),
      im(x.im)
  {
  }
  Complex& operator=(const Complex& x)
  {
    re = x.re;
    im = x.im;
    return *this;
  }
  Complex& operator+=(const Complex& x)
  {
    re += x.re;
    im += x.im;
    return *this;
  }
  Complex& operator-=(const Complex& x)
  {
    re -= x.re;
    im -= x.im;
    return *this;
  }
  Complex& operator*=(const Complex& x)
  {
    double  tmp = (re * x.re) - (im * x.im);
    im = (re * x.im) + (im * x.re);
    re = tmp;
    return *this;
  }
  Complex& operator/=(const Complex& x)
  {
    double  r2 = (x.re * x.re) + (x.im * x.im);
    Complex tmp(x.re / r2, -x.im / r2);
    *this *= tmp;
    return *this;
  }
  operator double() const
  {
    return std::sqrt((re * re) + (im * im));
  }
  double phase() const
  {
    return std::atan2(im, re);
  }
  Complex& resistor(double r)
  {
    re = r;
    im = 0.0;
    return *this;
  }
  Complex& capacitor(double c, double f)
  {
    re = 0.0;
    im = -1.0 / (twoPi * f * c);
    return *this;
  }
  Complex& inductor(double l, double f)
  {
    re = 0.0;
    im = twoPi * f * l;
    return *this;
  }
  Complex& parallel(const Complex& x)
  {
    Complex tmp(*this);
    tmp += x;
    *this *= x;
    *this /= tmp;
    return *this;
  }
  Complex& voltageDividerGain(const Complex& zLoad, const Complex& zSrc)
  {
    Complex tmp(zLoad);
    tmp += zSrc;
    *this = zLoad;
    *this /= tmp;
    return *this;
  }
  Complex& voltageDivide(const Complex& zLoad, const Complex& zSrc)
  {
    Complex tmp(zLoad);
    tmp += zSrc;
    *this *= zLoad;
    *this /= tmp;
    return *this;
  }
};

static double ituR468Weight(double f)
{
  Complex g(8.164578, 0.0);
  Complex zSrc(600.0, 0.0);
  Complex c1;
  c1.capacitor(0.00000001385, f);
  g.voltageDivide(c1, zSrc);
  zSrc.parallel(c1);
  Complex l1;
  l1.inductor(0.01288, f);
  zSrc += l1;
  Complex c2;
  c2.capacitor(0.00000002682, f);
  g.voltageDivide(c2, zSrc);
  zSrc.parallel(c2);
  Complex c3;
  c3.capacitor(0.00000003306, f);
  zSrc += c3;
  Complex c4;
  c4.capacitor(0.00000000921, f);
  g.voltageDivide(c4, zSrc);
  zSrc.parallel(c4);
  Complex l2;
  l2.inductor(0.02649, f);
  zSrc += l2;
  Complex c5;
  c5.capacitor(0.00000003147, f);
  g.voltageDivide(c5, zSrc);
  zSrc.parallel(c5);
  Complex r1(600.0, 0.0);
  g.voltageDivide(r1, zSrc);
  return double(g);
}

static double aWeight(double f)
{
  double  n = 12200.0 * f * f;
  n = n * n;
  double  d = (f * f) + (107.7 * 107.7);
  d *= ((f * f) + (737.9 * 737.9));
  d = std::sqrt(d) * ((f * f) + (20.6 * 20.6));
  d *= ((f * f) + (12200.0 * 12200.0));
  return ((n / d) * 1.258897);
}

// ----------------------------------------------------------------------------

static void analyzeSignal(AnalysisResults& results,
                          const std::vector< double >& buf, double sampleRate,
                          const AnalysisParameters& params)
{
  results.clear();
  if (buf.size() < 4)
    return;
  sampleRate *= params.sampleRateMult;
  size_t  fftSize = 16;
  while (fftSize < buf.size())
    fftSize = fftSize << 1;
  std::vector< double > fftBuf(fftSize, 0.0);
  results.peakMin = buf[0];
  results.peakMax = buf[0];
  double  wsum = 0.0;
  double  w2sum = 0.0;
  if (params.winParam < 0)
  {
    GaussianWindowNormalized  w(long(buf.size() + 1),
                                std::pow(10.0, double(params.winParam) * 0.05));
    (void) w.sample();
    for (size_t i = 1; true; i++)
    {
      double  w_ = w.sample();
      wsum += w_;
      w2sum += (w_ * w_);
      fftBuf[i] = buf[i] * w_;
      if (buf[i] < results.peakMin)
        results.peakMin = buf[i];
      else if (buf[i] > results.peakMax)
        results.peakMax = buf[i];
      size_t  j = buf.size() - i;
      if (j <= i)
        break;
      wsum += w_;
      w2sum += (w_ * w_);
      fftBuf[j] = buf[j] * w_;
      if (buf[j] < results.peakMin)
        results.peakMin = buf[j];
      else if (buf[j] > results.peakMax)
        results.peakMax = buf[j];
    }
  }
  else
  {
    CosineGen w(double(long(buf.size())));
    (void) w.sample();
    for (size_t i = 1; true; i++)
    {
      double  w_ = cosineToNuttallWindow(w.sample());
      wsum += w_;
      w2sum += (w_ * w_);
      fftBuf[i] = buf[i] * w_;
      if (buf[i] < results.peakMin)
        results.peakMin = buf[i];
      else if (buf[i] > results.peakMax)
        results.peakMax = buf[i];
      size_t  j = buf.size() - i;
      if (j <= i)
        break;
      wsum += w_;
      w2sum += (w_ * w_);
      fftBuf[j] = buf[j] * w_;
      if (buf[j] < results.peakMin)
        results.peakMin = buf[j];
      else if (buf[j] > results.peakMax)
        results.peakMax = buf[j];
    }
  }
  realFFT(&(fftBuf.front()), fftSize, false);
  double  bw = (params.signalBandwidthBins + 0.25)
               * sampleRate / double(long(buf.size()));
  for (size_t i = 0; i <= (fftSize >> 1); i++)
  {
    double  f = double(long(i)) * sampleRate / double(long(fftSize));
    double  re = fftBuf[(i < (fftSize >> 1) ? (i << 1) : 1)];
    double  im = (i > 0 && i < (fftSize >> 1) ? fftBuf[(i << 1) + 1] : 0.0);
    double  r2 = (re * re) + (im * im);
    bool    isSignal = false;
    for (size_t j = 0; j < params.signalFrequencies.size(); j++)
    {
      if (std::fabs(f - params.signalFrequencies[j]) < bw)
      {
        isSignal = true;
        break;
      }
    }
    double  aWt = 0.0;
    double  r468Wt = 0.0;
    if (i > 0)
    {
      aWt = aWeight(f);
      r468Wt = ituR468Weight(f);
    }
    else
    {
      results.dcLevel = re;
    }
    if (isSignal)
    {
      results.signalLevel += r2;
      results.signalAWeighted += (r2 * aWt * aWt);
      results.signalR468Weighted += (r2 * r468Wt * r468Wt);
    }
    else if (f >= params.noiseFreqMin && f <= params.noiseFreqMax)
    {
      results.noiseLevel += r2;
      results.noiseAWeighted += (r2 * aWt * aWt);
      results.noiseR468Weighted += (r2 * r468Wt * r468Wt);
    }
  }
  double  scale = 4.0 / (w2sum * double(long(fftSize)));
  results.dcLevel = results.dcLevel / wsum;
  results.signalLevel = std::sqrt(results.signalLevel * scale);
  results.signalAWeighted = std::sqrt(results.signalAWeighted * scale);
  results.signalR468Weighted = std::sqrt(results.signalR468Weighted * scale);
  results.noiseLevel = std::sqrt(results.noiseLevel * scale);
  results.noiseAWeighted = std::sqrt(results.noiseAWeighted * scale);
  results.noiseR468Weighted = std::sqrt(results.noiseR468Weighted * scale);
}

static void compensateLevel(double& x, const double& noise)
{
  x = (x * x) - (noise * noise);
  if (x > 0.0)
    x = std::sqrt(x);
  else
    x = 0.0;
}

static void printLevel(const char *s, double x,
                       const AnalysisParameters& params, bool isPeak = false)
{
  double  v = x * params.fullScaleVoltage;
  if (params.lvlAdjust != 0.0)
    x = x * std::pow(10.0, params.lvlAdjust / 20.0);
  std::printf("    %s%11.7f", s, x);
  if (x <= -1.0e-10 || x >= 1.0e-10)
  {
    std::printf(" (%8.3f dBr", 20.0 * std::log10(std::fabs(x)));
    if (v <= -1.0e-12 || v >= 1.0e-12)
    {
      if (!isPeak)
      {
        std::printf(", %8.3f dBu",
                    20.0 * std::log10(std::fabs(v) / 0.77459667));
      }
      else
      {
        v *= std::sqrt(2.0);
      }
      if (v < -0.999 || v > 0.999)
        std::printf(", %9.4f V", v);
      else if (v < -0.000999 || v > 0.000999)
        std::printf(", %8.3f mV", v * 1000.0);
      else if (v < -0.000000999 || v > 0.000000999)
        std::printf(", %8.3f uV", v * 1.0e6);
      else
        std::printf(", %8.3f nV", v * 1.0e9);
      if (!isPeak)
        std::printf("rms");
    }
    std::printf(")\n");
  }
  else
  {
    std::printf("\n");
  }
}

int main(int argc, char **argv)
{
  try
  {
    AnalysisParameters  params;
    params.sampleRateMult = 1.0;
    params.noiseFreqMin = 10.0;
    params.noiseFreqMax = 22000.0;
    params.signalBandwidthBins = 10.0;
    params.lvlAdjust = 0.0;
    params.fullScaleVoltage = 0.0;
    params.winParam = -120;
    params.signalFrequencies.resize(10, -1.0);
    std::string fileName;
    std::string noiseFileName;
    double      startTime = 0.0;
    double      duration = 10.0;
    int         channel = 0;
    {
      CommandLineParser opts(argv[0]);
      opts.addFileName("FILENAME", "input file name", fileName);
      opts.addOption("noise-file", "nf",
                     "compensate the results for the noise in this file",
                     noiseFileName);
      opts.addOption("isrm", "m",
                     "factor to multiply assumed input sample rate by; "
                     "it only affects\n"
                     "        the duration and signal frequencies, "
                     "but not the conversion of\n"
                     "        skip time to samples",
                     params.sampleRateMult, 0.1, 10.0);
      opts.addOption("skip", "k",
                     "start time in seconds or -samples (input file only)",
                     startTime);
      opts.addOption("duration", "d",
                     "analysis window length in seconds or -samples "
                     "(for both files)",
                     duration);
      opts.addOption("channel", "c", "channel to test (0 = all)",
                     channel, 0, 16);
      opts.addOption("window", "w",
                     "window type (0: Nuttall window, <0: Gaussian window "
                     "with a\n"
                     "        minimum value of the specified number of dB, "
                     "normalized to the\n"
                     "        range 0 to 1)",
                     params.winParam, -200, 0);
      opts.addOption("noise-freq-min", "fmin", "",
                     params.noiseFreqMin, 0.0, 500000.0);
      opts.addOption("noise-freq-max", "fmax",
                     "noise frequency range (in Hz)",
                     params.noiseFreqMax, 0.0, 500000.0);
      opts.addOption("signal-bandwidth", "bw",
                     "signal bandwidth in +/- 1/duration Hz",
                     params.signalBandwidthBins, 0.0, 100.0);
      opts.addOption("signal-frequency-1", "f1", "",
                     params.signalFrequencies[0], -1.0, 500000.0);
      opts.addOption("signal-frequency-2", "f2", "",
                     params.signalFrequencies[1], -1.0, 500000.0);
      opts.addOption("signal-frequency-3", "f3", "",
                     params.signalFrequencies[2], -1.0, 500000.0);
      opts.addOption("signal-frequency-4", "f4", "",
                     params.signalFrequencies[3], -1.0, 500000.0);
      opts.addOption("signal-frequency-5", "f5", "",
                     params.signalFrequencies[4], -1.0, 500000.0);
      opts.addOption("signal-frequency-6", "f6", "",
                     params.signalFrequencies[5], -1.0, 500000.0);
      opts.addOption("signal-frequency-7", "f7", "",
                     params.signalFrequencies[6], -1.0, 500000.0);
      opts.addOption("signal-frequency-8", "f8", "",
                     params.signalFrequencies[7], -1.0, 500000.0);
      opts.addOption("signal-frequency-9", "f9", "",
                     params.signalFrequencies[8], -1.0, 500000.0);
      opts.addOption("signal-frequency-10", "f10",
                     "signal frequencies (ignored if zero or negative)",
                     params.signalFrequencies[9], -1.0, 500000.0);
      opts.addOption("leveladjust", "a",
                     "adjust relative levels by the specified amount (in dB)",
                     params.lvlAdjust, -120.0, 120.0);
      opts.addOption("full-scale-voltage", "fsv",
                     "if non-zero, convert 0 dBFS to the specified voltage "
                     "in Vrms",
                     params.fullScaleVoltage, -1000.0, 1000.0);
      if (!opts.parseCommandLine(argc, argv))
        return 0;
      // remove unused signal frequencies
      size_t  nSignalFreqs = 0;
      for (size_t i = 0; i < params.signalFrequencies.size(); i++)
      {
        if (params.signalFrequencies[i] > 0.0)
        {
          params.signalFrequencies[nSignalFreqs] = params.signalFrequencies[i];
          nSignalFreqs++;
        }
      }
      params.signalFrequencies.resize(nSignalFreqs);
    }
    AudioFilter_FileInput inFile(fileName.c_str());
    double  sampleRate = inFile.getSampleRate();
    size_t  channels = size_t(inFile.getChannelCount());
    if (size_t(channel) > channels)
      throw std::runtime_error(std::string("channel number is out of range"));
    size_t  nFrames =
        size_t(long((duration > 0.0 ?
                     (duration * sampleRate * params.sampleRateMult)
                     : -duration)
                    + 0.5));
    if (nFrames < 4 || nFrames > 0x02000000)
      throw std::runtime_error(std::string("invalid duration"));
    {
      sf_count_t  skipFrames =
          sf_count_t((startTime > 0.0 ? (startTime * sampleRate) : -startTime)
                     + 0.5);
      if ((inFile.getFileLength() - skipFrames) < sf_count_t(nFrames))
      {
        throw std::runtime_error(std::string("input file has insufficient data "
                                             "for the specified duration"));
      }
      inFile.setPosition(skipFrames, false);
    }
    std::vector< AnalysisResults >  inFileResults(channel < 1 ? channels : 1);
    std::vector< AnalysisResults >  noiseFileResults(inFileResults.size());
    std::vector< std::vector< double > >  bufs(inFileResults.size());
    for (size_t i = 0; i < bufs.size(); i++)
      bufs[i].resize(nFrames, 0.0);
    // if specified, load and analyze noise file
    if (!noiseFileName.empty())
    {
      AudioFilter_FileInput noiseFile(noiseFileName.c_str());
      if (size_t(noiseFile.getChannelCount()) != channels)
      {
        throw std::runtime_error(std::string("input files have unmatched "
                                             "channel counts"));
      }
      if (noiseFile.getSampleRate() != sampleRate)
      {
        throw std::runtime_error(std::string("input files have unmatched "
                                             "sample rates"));
      }
      if (noiseFile.getFileLength() < sf_count_t(nFrames))
      {
        throw std::runtime_error(std::string("noise file has insufficient data "
                                             "for the specified duration"));
      }
      for (size_t i = 0; i < nFrames; i++)
      {
        noiseFile.process();
        for (size_t j = 0; j < bufs.size(); j++)
        {
          bufs[j][i] = noiseFile.getOutputBuffer()[(channel < 1 ?
                                                    j : size_t(channel - 1))];
        }
      }
      for (size_t i = 0; i < noiseFileResults.size(); i++)
        analyzeSignal(noiseFileResults[i], bufs[i], sampleRate, params);
    }
    // load input file
    for (size_t i = 0; i < nFrames; i++)
    {
      inFile.process();
      for (size_t j = 0; j < bufs.size(); j++)
      {
        bufs[j][i] = inFile.getOutputBuffer()[(channel < 1 ?
                                               j : size_t(channel - 1))];
      }
    }
    // analyze input, and print results
    for (size_t i = 0; i < inFileResults.size(); i++)
    {
      analyzeSignal(inFileResults[i], bufs[i], sampleRate, params);
      std::printf("Channel #%d:\n", (channel < 1 ? int(i + 1) : channel));
      printLevel("DC offset:         ",
                 inFileResults[i].dcLevel, params, true);
      printLevel("Minimum value:     ",
                 inFileResults[i].peakMin, params, true);
      printLevel("Maximum value:     ",
                 inFileResults[i].peakMax, params, true);
      printLevel("Signal (RMS):      ",
                 inFileResults[i].signalLevel, params);
      printLevel("Signal (A-wtd.):   ",
                 inFileResults[i].signalAWeighted, params);
      printLevel("Signal (R468 wtd.):",
                 inFileResults[i].signalR468Weighted, params);
      printLevel("Noise (RMS):       ",
                 inFileResults[i].noiseLevel, params);
      printLevel("Noise (A-weighted):",
                 inFileResults[i].noiseAWeighted, params);
      printLevel("Noise (R468 wtd.): ",
                 inFileResults[i].noiseR468Weighted, params);
      if (!noiseFileName.empty())
      {
        std::printf("  Noise compensated levels:\n");
        compensateLevel(inFileResults[i].signalLevel,
                        noiseFileResults[i].signalLevel);
        compensateLevel(inFileResults[i].signalAWeighted,
                        noiseFileResults[i].signalAWeighted);
        compensateLevel(inFileResults[i].signalR468Weighted,
                        noiseFileResults[i].signalR468Weighted);
        compensateLevel(inFileResults[i].noiseLevel,
                        noiseFileResults[i].noiseLevel);
        compensateLevel(inFileResults[i].noiseAWeighted,
                        noiseFileResults[i].noiseAWeighted);
        compensateLevel(inFileResults[i].noiseR468Weighted,
                        noiseFileResults[i].noiseR468Weighted);
        printLevel("Signal (RMS):      ",
                   inFileResults[i].signalLevel, params);
        printLevel("Signal (A-wtd.):   ",
                   inFileResults[i].signalAWeighted, params);
        printLevel("Signal (R468 wtd.):",
                   inFileResults[i].signalR468Weighted, params);
        printLevel("Noise (RMS):       ",
                   inFileResults[i].noiseLevel, params);
        printLevel("Noise (A-weighted):",
                   inFileResults[i].noiseAWeighted, params);
        printLevel("Noise (R468 wtd.): ",
                   inFileResults[i].noiseR468Weighted, params);
      }
    }
  }
  catch (std::exception& e)
  {
    std::fprintf(stderr, "%s: %s\n", argv[0], e.what());
    return -1;
  }
  return 0;
}

