
#include "headers.hpp"
#include "dsplib.cpp"
#include "options.cpp"

int main(int argc, char **argv)
{
  try
  {
    std::string inFileName;
    std::string outFileName;
    double      srOut = -1.0;
    double      isrAdj = 1.0;
    double      startPos = 0.0;
    double      duration = 0.0;
    bool        isLooped = false;
    bool        gainFlag = false;
    int         quality = 0;
    int         interpLen = 48;
    int         lpFilterLen = 32768;
    double      lpFilterFreq = 0.4994;
    int         lpFilterWindow = -124;
    int         outFmt = 0;
    int         ditherSeed = 0;
    double      gain = 1.0;
    double      channelGain[16] =
    {
      1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
      1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0
    };
    static const int  formatTable[6] =
    {
      SF_FORMAT_WAV | SF_FORMAT_FLOAT,
      SF_FORMAT_WAV | SF_FORMAT_PCM_U8,
      SF_FORMAT_WAV | SF_FORMAT_PCM_16,
      SF_FORMAT_WAV | SF_FORMAT_PCM_24,
      SF_FORMAT_WAV | SF_FORMAT_PCM_32,
      SF_FORMAT_WAV | SF_FORMAT_DOUBLE
    };
    {
      CommandLineParser opts(argv[0]);
      opts.addFileName("INFILE", "input file name", inFileName);
      opts.addFileName("OUTFILE", "output file name", outFileName);
      opts.addOption("format", "f",
                     "output sample format (0: float, 1/2/3/4: "
                     "8/16/24/32 bit PCM,\n"
                     "        5: double)",
                     outFmt, 0, 5);
      opts.addOption("dither-seed", "ds",
                     "random seed for dithering integer sample formats; zero "
                     "seeds\n"
                     "        from the current time in seconds, -1 disables "
                     "dither",
                     ditherSeed, -1, 0x7FFFFFFE);
      opts.addOption("sample-rate", "r",
                     "output sample rate in Hz, or resample ratio if negative "
                     "(e.g.\n"
                     "        -2.0 will resample 44100 Hz to 88200 Hz)",
                     srOut, -100.0, 1000000.0);
      opts.addOption("isrm", "m",
                     "factor to multiply assumed input sample rate by; "
                     "it only affects\n"
                     "        the resample ratio, but not the conversion of "
                     "skip time to\n"
                     "        samples",
                     isrAdj, 0.01, 100.0);
      opts.addOption("skip", "k",
                     "start time in seconds (a negative value delays the "
                     "output)",
                     startPos);
      opts.addOption("looped", "l", "loop input file (requires -d)", isLooped);
      opts.addOption("duration", "d",
                     "if non-zero, duration limit in seconds, or "
                     "-sample frames of\n"
                     "        output",
                     duration);
      opts.addOption("interp-len", "il",
                     "interpolation window length in samples (must be a "
                     "multiple of 4)",
                     interpLen, 8, 2048);
      opts.addOption("filter-len", "fl",
                     "FIR lowpass filter length in samples or -microseconds",
                     lpFilterLen, -10000000, 4194304);
      opts.addOption("filter-freq", "ff",
                     "FIR lowpass filter cutoff frequency as a fraction "
                     "of the lower\n"
                     "        sample rate; if negative, a minimum phase "
                     "filter will be used",
                     lpFilterFreq, -1.0, 1.0);
      opts.addOption("filter-window", "fw",
                     "FIR lowpass filter window parameter (0: Nuttall window, "
                     "<0:\n"
                     "        Gaussian window with a minimum value of the "
                     "specified number of\n"
                     "        dB, normalized to the range 0 to 1)",
                     lpFilterWindow, -240, 0);
      opts.addOption("quality", "q",
                     "if non-zero, automatically set -il, -fl, and -fw to "
                     "avoid\n"
                     "        imaging and aliasing, with a stopband rejection "
                     "of about 70 to\n"
                     "        200 dB, depending on the quality value "
                     "specified (higher is\n"
                     "        better, but slower)",
                     quality, 0, 9);
      opts.addOption("gain", "g", "overall gain (1.0 = 0 dB) for all channels",
                     gain, -1000.0, 1000.0);
      opts.addOption("channel1-gain", "g1", "",
                     channelGain[0], -1000.0, 1000.0);
      opts.addOption("channel2-gain", "g2", "",
                     channelGain[1], -1000.0, 1000.0);
      opts.addOption("channel3-gain", "g3", "",
                     channelGain[2], -1000.0, 1000.0);
      opts.addOption("channel4-gain", "g4", "",
                     channelGain[3], -1000.0, 1000.0);
      opts.addOption("channel5-gain", "g5", "",
                     channelGain[4], -1000.0, 1000.0);
      opts.addOption("channel6-gain", "g6", "",
                     channelGain[5], -1000.0, 1000.0);
      opts.addOption("channel7-gain", "g7", "",
                     channelGain[6], -1000.0, 1000.0);
      opts.addOption("channel8-gain", "g8", "gain for channel 1 to 8",
                     channelGain[7], -1000.0, 1000.0);
      if (!opts.parseCommandLine(argc, argv))
        return 0;
    }
    if (isLooped && duration == 0.0)
    {
      throw std::runtime_error(std::string("no duration specified for looped "
                                           "input"));
    }
    if (quality)
    {
      interpLen = 12 + (quality << 2);
      double  filterFreq = std::fabs(lpFilterFreq);
      bool    isMinimumPhase = (lpFilterFreq < 0.0);
      filterFreq = (filterFreq > 0.375 ?
                    (filterFreq < 0.4995 ? filterFreq : 0.4995) : 0.375);
      lpFilterFreq = (isMinimumPhase ? -filterFreq : filterFreq);
      lpFilterWindow = -(34 + (quality * 14));
      double  sampleRateIn = 0.0;
      {
        AudioFilter_FileInput f(inFileName.c_str());
        sampleRateIn = f.getSampleRate();
      }
      double  sampleRateOut = (srOut < 0.0 ? (-srOut * sampleRateIn) : srOut);
      sampleRateIn = sampleRateIn * isrAdj;
      double  bw = (sampleRateIn < sampleRateOut ? sampleRateIn : sampleRateOut)
                   * (0.5 - filterFreq);
      bw = (bw > 5.0 ? bw : 5.0);
      lpFilterLen =
          int((-2113281.25 - (1136718.75 * double(quality))) / bw - 0.5);
    }
    AudioFilter_ResampledFileInput  inFile(inFileName.c_str(), srOut, isrAdj,
                                           startPos, isLooped, interpLen,
                                           lpFilterLen, lpFilterFreq,
                                           lpFilterWindow);
    double  sampleRate = inFile.getSampleRate();
    int     channels = inFile.getChannelCount();
    if (channels > int(sizeof(channelGain) / sizeof(double)))
    {
      throw std::runtime_error(std::string("input file has unsupported "
                                           "number of channels"));
    }
    sf_count_t  maxFrames = -1;
    if (duration > 0.0)
      maxFrames = sf_count_t(duration * sampleRate + 0.5);
    else if (duration < 0.0)
      maxFrames = sf_count_t(-duration + 0.5);
    for (int i = 0; i < channels; i++)
    {
      if (gain != 1.0 || channelGain[i] != 1.0)
      {
        gainFlag = true;
        channelGain[i] = channelGain[i] * gain;
      }
    }
    AudioOutputFile outFile(outFileName.c_str(), sampleRate, channels,
                            formatTable[outFmt], ditherSeed);
    if (!gainFlag)
    {
      if (maxFrames >= 0)
      {
        while (maxFrames > 0 && inFile.process())
        {
          maxFrames--;
          outFile.writeSampleFrame(&(inFile.getOutputBuffer().front()));
        }
      }
      else
      {
        while (inFile.process())
        {
          outFile.writeSampleFrame(&(inFile.getOutputBuffer().front()));
        }
      }
    }
    else
    {
      double  tmpBuf[16];
      const double  *bufrp = &(inFile.getOutputBuffer().front());
      if (maxFrames >= 0)
      {
        while (maxFrames > 0 && inFile.process())
        {
          maxFrames--;
          int     i = 0;
          do
          {
            tmpBuf[i] = bufrp[i] * channelGain[i];
          }
          while (++i < channels);
          outFile.writeSampleFrame(&(tmpBuf[0]));
        }
      }
      else
      {
        while (inFile.process())
        {
          int     i = 0;
          do
          {
            tmpBuf[i] = bufrp[i] * channelGain[i];
          }
          while (++i < channels);
          outFile.writeSampleFrame(&(tmpBuf[0]));
        }
      }
    }
  }
  catch (std::exception& e)
  {
    std::fprintf(stderr, "%s: %s\n", argv[0], e.what());
    return -1;
  }
  return 0;
}

