
#include "headers.hpp"
#include "dsplib.cpp"

#include <cctype>
#include <list>
#include <typeinfo>
#include <algorithm>

// ----------------

class Generator
{
 public:
  Generator()
  {
  }
  virtual ~Generator()
  {
  }
  virtual bool process(double& l, double& r) = 0;
};

// ----------------

static inline double wrapPhase(double phs)
{
  if (phs < 0.0)
  {
    phs += twoPi;
    if (phs < 0.0)
      phs += twoPi;
    if (!(phs >= 0.0))
      phs = 0.0;
  }
  else if (phs >= twoPi)
  {
    phs -= twoPi;
    if (phs >= twoPi)
      phs -= twoPi;
    if (!(phs < twoPi))
      phs = 0.0;
  }
  return phs;
}

class SineGenerator : public Generator
{
 protected:
  double  phase;
  double  omega;
  double  ampL;
  double  ampR;
  long    samplesRemaining;
 public:
  // startPhase is in degrees
  SineGenerator(double sampleRate, double freq, double ampL_, double ampR_,
                double startPhase, double duration);
  virtual bool process(double& l, double& r);
};

SineGenerator::SineGenerator(double sampleRate, double freq,
                             double ampL_, double ampR_,
                             double startPhase, double duration)
{
  samplesRemaining = long(duration * sampleRate + 0.5);
  phase = wrapPhase(startPhase * pi / 180.0);
  omega = freq * twoPi / sampleRate;
  ampL = ampL_;
  ampR = ampR_;
  if (omega < -pi || omega > pi)
  {
    omega = 0.0;
    ampL = 0.0;
    ampR = 0.0;
  }
}

bool SineGenerator::process(double& l, double& r)
{
  if (samplesRemaining < 1L)
    return false;
  samplesRemaining--;
  double  y = std::sin(phase);
  l += (y * ampL);
  r += (y * ampR);
  phase = wrapPhase(phase + omega);
  return true;
}

// ----------------

class SineSweepGenerator : public Generator
{
 protected:
  double  phase;
  double  omega;
  double  omegaMult;
  double  ampL;
  double  ampR;
  double  ampMult;
  long    samplesRemaining;
 public:
  // startPhase is in degrees
  SineSweepGenerator(double sampleRate, double startFreq, double endFreq,
                     double startAmp, double endAmp, double ampL_, double ampR_,
                     double startPhase, double duration);
  virtual bool process(double& l, double& r);
};

SineSweepGenerator::SineSweepGenerator(double sampleRate,
                                       double startFreq, double endFreq,
                                       double startAmp, double endAmp,
                                       double ampL_, double ampR_,
                                       double startPhase, double duration)
{
  samplesRemaining = long(duration * sampleRate + 0.5);
  phase = wrapPhase(startPhase * pi / 180.0);
  omegaMult = std::pow(endFreq / startFreq, 1.0 / double(samplesRemaining));
  omega = std::sqrt(omegaMult) * startFreq * twoPi / sampleRate;
  ampMult = std::pow(endAmp / startAmp, 1.0 / double(samplesRemaining));
  ampL = std::sqrt(ampMult) * startAmp;
  ampR = ampR_ * ampL;
  ampL = ampL_ * ampL;
}

bool SineSweepGenerator::process(double& l, double& r)
{
  if (samplesRemaining < 1L)
    return false;
  samplesRemaining--;
  double  y = std::sin(phase);
  if (omega >= -pi && omega <= pi)
  {
    l += (y * ampL);
    r += (y * ampR);
  }
  phase = wrapPhase(phase + omega);
  omega *= omegaMult;
  ampL *= ampMult;
  ampR *= ampMult;
  return true;
}

// ----------------

class JTestGenerator : public Generator
{
 protected:
  long    samplesRemaining;
  int     sampleCnt;
  int     toggleInterval;
  int     nBits;
 public:
  JTestGenerator(double sampleRate, int nBits_, double toggleInterval_,
                 double duration);
  virtual bool process(double& l, double& r);
};

JTestGenerator::JTestGenerator(double sampleRate, int nBits_,
                               double toggleInterval_, double duration)
{
  samplesRemaining = long(duration * sampleRate + 0.5);
  sampleCnt = 0;
  toggleInterval = int(toggleInterval_ * sampleRate + 0.5);
  toggleInterval = std::min(std::max(toggleInterval, 1), 1000000000);
  nBits = std::min(std::max(nBits_, 8), 32);
}

bool JTestGenerator::process(double& l, double& r)
{
  if (samplesRemaining < 1L)
    return false;
  samplesRemaining--;
  int     tmp = 0x40000000;
  if ((sampleCnt & 2) != 0)
    tmp = -tmp;
  if (sampleCnt >= 0)
    tmp--;
  tmp &= (int(0xFFFFFFFFU) << (32 - nBits));
  l = double(tmp) / double(0x80000000U);
  r = l;
  sampleCnt = (sampleCnt < (toggleInterval - 1) ?
               (sampleCnt + 1) : -toggleInterval);
  return true;
}

// ----------------

class MLSGenerator : public Generator
{
 protected:
  unsigned int  curState;
  unsigned int  poly;
  double  ampL;
  double  ampR;
  long    samplesRemaining;
 public:
  MLSGenerator(double sampleRate, unsigned int poly_,
               double ampL_, double ampR_, double duration);
  virtual bool process(double& l, double& r);
};

MLSGenerator::MLSGenerator(double sampleRate, unsigned int poly_,
                           double ampL_, double ampR_, double duration)
{
  samplesRemaining = long(duration * sampleRate + 0.5);
  curState = 0xFFFFFFFFU;
  poly = poly_;
  ampL = ampL_;
  ampR = ampR_;
}

bool MLSGenerator::process(double& l, double& r)
{
  if (samplesRemaining < 1L)
    return false;
  samplesRemaining--;
  l = double((int(curState & 1U) << 1) - 1);
  r = l * ampR;
  l = l * ampL;
  unsigned int  tmp = curState & poly;
  tmp = tmp ^ (tmp >> 16);
  tmp = tmp ^ (tmp >> 8);
  tmp = tmp ^ (tmp >> 4);
  tmp = tmp ^ (tmp >> 2);
  tmp = tmp ^ (tmp >> 1);
  curState = (curState << 1) | (tmp & 1U);
  return true;
}

// ----------------

class WaveGenerator : public Generator
{
 public:
  enum WaveForm
  {
    SQUARE = 1,
    SAWTOOTH = 2,
    TRIANGLE = 3,
    PULSE = 4
  };
 protected:
  int     phase;
  int     cycleTime;
  WaveForm  waveForm;
  double  ampL;
  double  ampR;
  long    samplesRemaining;
 public:
  // startPhase is in degrees
  WaveGenerator(double sampleRate, WaveForm waveForm_, double freq,
                double ampL_, double ampR_, double startPhase, double duration);
  virtual bool process(double& l, double& r);
};

WaveGenerator::WaveGenerator(double sampleRate, WaveForm waveForm_,
                             double freq, double ampL_, double ampR_,
                             double startPhase, double duration)
{
  samplesRemaining = long(duration * sampleRate + 0.5);
  waveForm = waveForm_;
  if (waveForm_ == SQUARE || waveForm_ == TRIANGLE)
    cycleTime = int(sampleRate * 0.5 / freq + 0.5) << 1;
  else
    cycleTime = int(sampleRate / freq + 0.5);
  if (cycleTime < 2)
    cycleTime = 2;
  if (waveForm_ == TRIANGLE)
    startPhase -= 90.0;
  phase = int(startPhase * cycleTime / 360.0 + (startPhase < 0.0 ? -0.5 : 0.5));
  phase = phase % cycleTime;
  if (phase < 0)
    phase += cycleTime;
  ampL = ampL_;
  ampR = ampR_;
}

bool WaveGenerator::process(double& l, double& r)
{
  if (samplesRemaining < 1L)
    return false;
  samplesRemaining--;
  double  y = 0.0;
  switch (waveForm)
  {
    case SQUARE:
      y = (phase < (cycleTime >> 1) ? 1.0 : -1.0);
      break;
    case SAWTOOTH:
      y = 1.0 - ((double(phase) * 2.0 + 1.0) / double(cycleTime));
      break;
    case TRIANGLE:
      if (phase < (cycleTime >> 1))
        y = 1.0 - (double(phase) * 4.0 / double(cycleTime));
      else
        y = double(phase) * 4.0 / double(cycleTime) - 3.0;
      break;
    case PULSE:
      y = (phase == 0 ? 1.0 : 0.0) - (1.0 / double(cycleTime));
      break;
  }
  l += (y * ampL);
  r += (y * ampR);
  if (++phase >= cycleTime)
    phase = 0;
  return true;
}

// ----------------

class DitherGenerator : public Generator
{
 protected:
  double  ampScale;
  long    samplesRemaining;
  RandomGenerator rng;
 public:
  DitherGenerator(double sampleRate, int nBits_, double duration);
  virtual bool process(double& l, double& r);
};

DitherGenerator::DitherGenerator(double sampleRate, int nBits_, double duration)
  : Generator(),
    ampScale(0.5 / double(1 << (std::min(std::max(nBits_ - 2, 2), 30)))),
    samplesRemaining(long(duration * sampleRate + 0.5)),
    rng(1)
{
}

bool DitherGenerator::process(double& l, double& r)
{
  if (samplesRemaining < 1L)
    return false;
  samplesRemaining--;
  l = (rng.random() + rng.random() - 1.0) * ampScale;
  r = (rng.random() + rng.random() - 1.0) * ampScale;
  return true;
}

// ----------------

class NoiseGenerator : public Generator, public AudioFilter
{
 protected:
  double  ampL;
  double  ampR;
  long    samplesRemaining;
  RandomGenerator rng;
  AudioFilter_FIRFilter *filter;
 public:
  NoiseGenerator(double sampleRate_, double slope,
                 double ampL_, double ampR_, double duration);
  virtual ~NoiseGenerator();
  virtual bool process();
  virtual bool process(double& l, double& r);
};

NoiseGenerator::NoiseGenerator(double sampleRate_, double slope,
                               double ampL_, double ampR_, double duration)
  : Generator(),
    AudioFilter(sampleRate_, 2),
    ampL(ampL_),
    ampR(ampR_),
    samplesRemaining(long(duration * sampleRate_ + 0.5)),
    rng(1237654),
    filter((AudioFilter_FIRFilter *) 0)
{
  std::vector< std::vector< double > >  irBufs(1);
  std::vector< double >&  fftBuf = irBufs[0];
  if (!(slope < -1.0e-6 || slope > 1.0e-6))
  {
    fftBuf.resize(2, 0.0);
    fftBuf[0] = 1.0;
  }
  else
  {
    fftBuf.resize(131072, 0.0);
    double  maxMagSq = 100000.0;        // limit magnitude to 50 dB
    if (slope < 0.0)
    {
      // white noise below 5 Hz
      double  tmp = std::pow(5.0 / 1000.0, slope * 2.0);
      maxMagSq = (tmp < maxMagSq ? tmp : maxMagSq);
    }
    for (size_t i = 0; i <= (fftBuf.size() >> 1); i++)
    {
      double  a = 0.0;
      if (i > 0)
      {
        // 0 dB gain at 1 kHz
        a = double(long(i)) * sampleRate
            / (double(long(fftBuf.size())) * 1000.0);
        a = 1.0 / (std::pow(a, slope * -2.0) + 0.00001);
        a = std::sqrt(a < maxMagSq ? a : maxMagSq);
      }
      else if (slope < 0.0)
      {
        a = std::sqrt(maxMagSq);
      }
      fftBuf[(i < (fftBuf.size() >> 1) ? (i << 1) : 1)] = (!(i & 1) ? a : -a);
    }
    realFFT(&(fftBuf.front()), fftBuf.size(), true);
    CosineGen w(double(long(fftBuf.size())));
    for (size_t i = 0; i < fftBuf.size(); i++)
      fftBuf[i] *= cosineToNuttallWindow(w.sample());
    minimumPhaseIR(fftBuf);
  }
  filter = new AudioFilter_FIRFilter(*this, irBufs);
}

NoiseGenerator::~NoiseGenerator()
{
  delete filter;
}

bool NoiseGenerator::process()
{
  outBuf[0] = rng.random() * 2.0 - 1.0;
  outBuf[1] = rng.random() * 2.0 - 1.0;
  return true;
}

bool NoiseGenerator::process(double& l, double& r)
{
  if (samplesRemaining < 1L)
    return false;
  samplesRemaining--;
  filter->process();
  l += filter->getOutputBuffer()[0] * ampL;
  r += filter->getOutputBuffer()[1] * ampR;
  return true;
}

// ----------------

class InputFile
{
 protected:
  std::FILE *f;
  int       prvChar;
  size_t    lineNum;
 public:
  InputFile(const char *fileName);
  virtual ~InputFile();
  int getCharacter();
  bool getLine(std::vector< std::string >& args);
  std::string getLineNumberString() const;
};

InputFile::InputFile(const char *fileName)
{
  if (!fileName || !fileName[0] || !(f = std::fopen(fileName, "rb")))
    throw std::runtime_error(std::string("error opening input file"));
  prvChar = '\n';
  lineNum = 0;
}

InputFile::~InputFile()
{
  std::fclose(f);
}

int InputFile::getCharacter()
{
  if (prvChar == EOF)
    return EOF;
  int     c = std::fgetc(f);
  if (c == EOF)
  {
    prvChar = EOF;
    return EOF;
  }
  c &= 0xFF;
  if (c == '\0')
    c = ' ';
  if (c == '\n' && prvChar == '\r')
  {
    prvChar = c;
    return getCharacter();
  }
  if (prvChar == '\r' || prvChar == '\n')
    lineNum++;
  if (c == '\r')
  {
    prvChar = c;
    return '\n';
  }
  prvChar = c;
  return c;
}

bool InputFile::getLine(std::vector< std::string >& args)
{
  args.clear();
  std::string buf;
  while (true)
  {
    int     c = getCharacter();
    if (c == '#' || c == ';')
    {
      do
      {
        c = getCharacter();
      }
      while (c != '\n' && c != EOF);
    }
    if (c == '\t' || c == ' ' || c == '\n' || c == EOF)
    {
      if (!buf.empty())
      {
        args.push_back(buf);
        buf.clear();
      }
      if (c == EOF || (c == '\n' && args.size() > 0))
        break;
    }
    else
    {
      buf += char(c);
    }
  }
  return (args.size() > 0);
}

std::string InputFile::getLineNumberString() const
{
  char    buf[64];
  std::sprintf(buf, "%lu", (unsigned long) lineNum);
  return std::string(buf);
}

// ----------------

// If sampleRate is greater than 0, a 'd' prefix will convert a dB value to
// raw amplitude, 's' will convert time from samples to seconds, and 't' will
// convert cycle time in samples to frequency in Hz.

static bool parseNumber(double& x, const char *s, double sampleRate = 0.0)
{
  x = 0.0;
  char    *endptr = NULL;
  char    prefix = '\0';
  if (!s)
    return false;
  while (*s == ' ' || *s == '\t')
    s++;
  if (sampleRate > 0.0)
  {
    if (*s == 'd' || *s == 's' || *s == 't')
      prefix = *(s++);
    if (*s == 'D' || *s == 'S' || *s == 'T')
      prefix = *(s++) + ('a' - 'A');
  }
  if (!s[0])
    return false;
  x = std::strtod(s, &endptr);
  if (!endptr || endptr == s || *endptr != '\0')
    return false;
  if (prefix == 'd')
    x = std::pow(10.0, x / 20.0);
  else if (prefix == 's')
    x = x / sampleRate;
  else if (prefix == 't')
    x = sampleRate / x;
  return true;
}

static bool parseBinaryInteger(unsigned int& x, const char *s)
{
  x = 0U;
  if (!s)
    return false;
  while (*s == ' ' || *s == '\t')
    s++;
  if (!s[0])
    return false;
  while (*s == '0' || *s == '1')
  {
    if (x >= 0x80000000U)
      return false;
    x = (x << 1) | (unsigned int) (*s & 1);
    s++;
  }
  while (*s == ' ' || *s == '\t')
    s++;
  return (*s == '\0');
}

static void parseInteger(int& x, const char *s, const char *name,
                         int minValue, int maxValue)
{
  long    tmp = 0L;
  char    *endptr = NULL;
  if (s && s[0])
  {
    tmp = std::strtol(s, &endptr, 0);
    if (endptr && endptr != s && *endptr == '\0' &&
        tmp >= minValue && tmp <= maxValue)
    {
      x = int(tmp);
      return;
    }
  }
  throw std::runtime_error(std::string("invalid ") + name);
}

// ----------------

struct Event
{
  Generator *gen;
  long    startTime;
  size_t  refCount_;
  size_t  *refCount;
  // ----------------
  Event();
  Event(const Event& r);
  Event(InputFile& f, double sampleRate, int nBits);
  ~Event()
  {
    if (*refCount < 2)
    {
      (*refCount) = 0;
      delete gen;
    }
    else
    {
      (*refCount)--;
    }
  }
  Event& operator=(const Event& r);
  bool operator<(const Event& r) const
  {
    return (startTime < r.startTime);
  }
};

Event::Event()
  : gen(NULL),
    startTime(0L),
    refCount_(1),
    refCount(&refCount_)
{
}

Event::Event(const Event& r)
  : gen(r.gen),
    startTime(r.startTime),
    refCount_(r.refCount_),
    refCount(r.refCount)
{
  (*refCount)++;
}

// sine TIME DURATION STARTFREQ STARTAMP ENDFREQ ENDAMP PHASE AMPL AMPR
// mls TIME DURATION POLY AMPL AMPR
// jtest TIME DURATION TOGGLEINT
// square STARTTIME DURATION FREQ PHASE AMPL AMPR
// sawtooth STARTTIME DURATION FREQ PHASE AMPL AMPR
// triangle STARTTIME DURATION FREQ PHASE AMPL AMPR
// pulse STARTTIME DURATION FREQ PHASE AMPL AMPR
// noise STARTTIME DURATION SLOPE AMPL AMPR

Event::Event(InputFile& f, double sampleRate, int nBits)
  : gen(NULL),
    startTime(0L),
    refCount_(1),
    refCount(&refCount_)
{
  std::vector< std::string >  args;
  if (!f.getLine(args))
    return;
  for (size_t i = 0; i < args[0].length(); i++)
  {
    if (std::isupper(args[0][i]))
      args[0][i] = std::tolower(args[0][i]);
  }
  if (args.size() < 3 || args.size() > 10)
  {
    throw std::runtime_error(std::string("invalid number of arguments at "
                                         "line ") + f.getLineNumberString());
  }
  double  params[9];
  for (size_t i = 1; i < args.size() && i < 10; i++)
  {
    if (!parseNumber(params[i - 1], args[i].c_str(), sampleRate))
    {
      throw std::runtime_error(std::string("invalid numeric format at line ")
                               + f.getLineNumberString());
    }
  }
  if (params[0] < 0.0)  // limit length and duration to non-negative values
  {
    params[1] += params[0];
    params[0] = 0.0;
  }
  if (params[1] < 0.0)
    params[1] = 0.0;
  startTime = long(params[0] * sampleRate + 0.5);
  if (args[0] == "sine")
  {
    if (args.size() == 7)
    {
      gen = new SineGenerator(sampleRate, params[2], params[4], params[5],
                              params[3], params[1]);
    }
    else if (args.size() == 10)
    {
      if (!((params[2] * params[4]) > 0.0))
      {
        throw std::runtime_error(std::string("invalid sine generator "
                                             "frequency envelope at line ")
                                 + f.getLineNumberString());
      }
      if (!((params[3] * params[5]) > 0.0))
      {
        throw std::runtime_error(std::string("invalid sine generator "
                                             "amplitude envelope at line ")
                                 + f.getLineNumberString());
      }
      gen = new SineSweepGenerator(sampleRate,
                                   params[2], params[4], params[3], params[5],
                                   params[7], params[8], params[6], params[1]);
    }
    else
    {
      throw std::runtime_error(std::string("invalid number of arguments at "
                                           "line ") + f.getLineNumberString());
    }
  }
  else if (args[0] == "mls")
  {
    if (args.size() != 6)
    {
      throw std::runtime_error(std::string("invalid number of arguments at "
                                           "line ") + f.getLineNumberString());
    }
    unsigned int  poly = 0U;
    if (!parseBinaryInteger(poly, args[3].c_str()))
    {
      throw std::runtime_error(std::string("invalid binary number format at "
                                           "line ") + f.getLineNumberString());
    }
    // check if the polynomial gives a maximum length sequence
    {
      int     nTaps = 0;
      int     maxTap = 0;
      for (unsigned char n = 0; n <= 31; n++)
      {
        if (poly & (1U << n))
        {
          nTaps++;
          maxTap = n + 1;
        }
      }
      if (!nTaps || (nTaps & 1) != 0)
      {
        throw std::runtime_error(std::string("invalid LFSR polynomial at line ")
                                 + f.getLineNumberString());
      }
      for (unsigned char i = 2; i <= maxTap; i++)
      {
        int     d = 0;
        for (unsigned char j = i; j <= maxTap; j += i)
        {
          if (poly & (1U << (j - 1)))
            d++;
        }
        if (d == nTaps)
        {
          throw std::runtime_error(std::string("invalid LFSR polynomial "
                                               "at line ")
                                   + f.getLineNumberString());
        }
      }
    }
    gen = new MLSGenerator(sampleRate, poly, params[3], params[4], params[1]);
  }
  else if (args[0] == "noise")
  {
    if (args.size() != 6)
    {
      throw std::runtime_error(std::string("invalid number of arguments at "
                                           "line ") + f.getLineNumberString());
    }
    if (!(params[2] >= -2.0 && params[2] <= 2.0))
    {
      throw std::runtime_error(std::string("invalid noise generator slope at "
                                           "line ") + f.getLineNumberString());
    }
    gen = new NoiseGenerator(sampleRate,
                             params[2], params[3], params[4], params[1]);
  }
  else if (args[0] == "jtest")
  {
    if (args.size() != 4)
    {
      throw std::runtime_error(std::string("invalid number of arguments at "
                                           "line ") + f.getLineNumberString());
    }
    gen = new JTestGenerator(sampleRate, nBits, params[2], params[1]);
  }
  else if (args[0] == "square" || args[0] == "sawtooth" ||
           args[0] == "triangle" || args[0] == "pulse")
  {
    if (args.size() != 7)
    {
      throw std::runtime_error(std::string("invalid number of arguments at "
                                           "line ") + f.getLineNumberString());
    }
    WaveGenerator::WaveForm waveForm =
        (args[0] == "square" ? WaveGenerator::SQUARE :
         (args[0] == "sawtooth" ? WaveGenerator::SAWTOOTH :
          (args[0] == "triangle" ? WaveGenerator::TRIANGLE :
           WaveGenerator::PULSE)));
    gen = new WaveGenerator(sampleRate, waveForm, params[2],
                            params[4], params[5], params[3], params[1]);
  }
  else
  {
    throw std::runtime_error(std::string("invalid event type at line ")
                             + f.getLineNumberString());
  }
}

Event& Event::operator=(const Event& r)
{
  gen = r.gen;
  startTime = r.startTime;
  refCount_ = r.refCount_;
  refCount = r.refCount;
  (*refCount)++;
  return *this;
}

// ----------------

class WAVOutputFile
{
 protected:
  static const int  maxChannels = 256;
  std::FILE     *f;
  unsigned char wavHeader[44];
  int           nBits;
  int           nChannels;
  size_t        framesWritten;
  // --------
  void writeHeaderWORD(int offs, unsigned short n);
  void writeHeaderDWORD(int offs, unsigned int n);
  void updateHeader();
 public:
  WAVOutputFile(const char *fileName,
                int sampleRate, int nBits_, int nChannels_);
  virtual ~WAVOutputFile();
  void writeData(const double *buf, size_t nFrames);
};

void WAVOutputFile::writeHeaderWORD(int offs, unsigned short n)
{
  wavHeader[offs] = (unsigned char) (n & 0xFF);
  wavHeader[offs + 1] = (unsigned char) ((n >> 8) & 0xFF);
}

void WAVOutputFile::writeHeaderDWORD(int offs, unsigned int n)
{
  wavHeader[offs] = (unsigned char) (n & 0xFFU);
  wavHeader[offs + 1] = (unsigned char) ((n >> 8) & 0xFFU);
  wavHeader[offs + 2] = (unsigned char) ((n >> 16) & 0xFFU);
  wavHeader[offs + 3] = (unsigned char) ((n >> 24) & 0xFFU);
}

void WAVOutputFile::updateHeader()
{
  if (std::fseek(f, 0L, SEEK_END) < 0)
    throw std::runtime_error(std::string("error seeking WAV output file"));
  long    fileSize = std::ftell(f);
  if (fileSize < 0L)
    throw std::runtime_error(std::string("error seeking WAV output file"));
  fileSize = (fileSize > 44L ? fileSize : 44L);
  writeHeaderDWORD(4, (unsigned int) (fileSize - 8L));
  writeHeaderDWORD(40, (unsigned int) (fileSize - 44L));
  if (std::fseek(f, 0L, SEEK_SET) < 0)
    throw std::runtime_error(std::string("error seeking WAV output file"));
  if (std::fwrite(&(wavHeader[0]), 1, sizeof(wavHeader), f)
      != sizeof(wavHeader))
  {
    throw std::runtime_error(std::string("error writing WAV output file"));
  }
  if (std::fflush(f) != 0)
    throw std::runtime_error(std::string("error writing WAV output file"));
  if (std::fseek(f, 0L, SEEK_END) < 0)
    throw std::runtime_error(std::string("error seeking WAV output file"));
}

WAVOutputFile::WAVOutputFile(const char *fileName,
                             int sampleRate, int nBits_, int nChannels_)
{
  if (sampleRate < 1000 || sampleRate > 1000000)
    throw std::runtime_error(std::string("invalid WAV file sample rate"));
  if (nBits_ < 8 || nBits_ > 32 || (nBits_ & 7) != 0)
    throw std::runtime_error(std::string("invalid WAV file bits per sample"));
  if (nChannels_ < 1 || nChannels_ > maxChannels)
  {
    throw std::runtime_error(std::string("invalid WAV file "
                                         "number of channels"));
  }
  if (!fileName || !fileName[0] || !(f = std::fopen(fileName, "wb")))
    throw std::runtime_error(std::string("error opening output file"));
  nBits = nBits_;
  nChannels = nChannels_;
  framesWritten = 0;
  std::memset(&(wavHeader[0]), 0, 44);
  std::memcpy(&(wavHeader[0]), "RIFF", 4);
  std::memcpy(&(wavHeader[8]), "WAVEfmt ", 8);
  std::memcpy(&(wavHeader[36]), "data", 4);
  writeHeaderDWORD(16, 16);             // length of "fmt " chunk
  writeHeaderWORD(20, 0x0001);          // WAVE_FORMAT_PCM
  writeHeaderWORD(22, (unsigned short) nChannels_);
  writeHeaderDWORD(24, (unsigned int) sampleRate);
  writeHeaderDWORD(28,
                   (unsigned int) (sampleRate * (nBits_ >> 3) * nChannels_));
  writeHeaderWORD(32, (unsigned short) ((nBits_ >> 3) * nChannels_));
  writeHeaderWORD(34, (unsigned short) nBits_);
  updateHeader();
}

WAVOutputFile::~WAVOutputFile()
{
  try
  {
    updateHeader();
  }
  catch (...)
  {
  }
  std::fclose(f);
}

void WAVOutputFile::writeData(const double *buf, size_t nFrames)
{
  double  scaleFac = double(1U << (nBits - 1));
  double  minVal = -scaleFac;
  double  maxVal = scaleFac - 1.0;
  size_t  sampleSize = size_t(nBits >> 3);
  size_t  frameSize = sampleSize * size_t(nChannels);
  unsigned char frameBuf[maxChannels << 2];
  for (size_t i = 0; i < nFrames; i++)
  {
    unsigned char *bufp = &(frameBuf[0]);
    for (int j = 0; j < nChannels; j++)
    {
      double        tmp = *(buf++) * scaleFac;
      tmp = (tmp > minVal ? (tmp < maxVal ? tmp : maxVal) : minVal);
      unsigned int  tmp2 = (unsigned int) int(tmp + (tmp < 0.0 ? -0.5 : 0.5));
      if (nBits == 8)
        tmp2 += 0x80U;
      for (size_t k = 0; k < sampleSize; k++, tmp2 >>= 8)
        *(bufp++) = (unsigned char) (tmp2 & 0xFFU);
    }
    if (std::fwrite(&(frameBuf[0]), sizeof(unsigned char), frameSize, f)
        != frameSize)
    {
      throw std::runtime_error(std::string("error writing WAV output file"));
    }
    if ((++framesWritten & 0x7FFF) == 0)
      updateHeader();
  }
}

// ----------------

int main(int argc, char **argv)
{
  if (argc != 5)
  {
    std::fprintf(stderr, "Usage: %s INFILE OUTFILE SAMPLERATE BITSPERSAMPLE\n",
                 argv[0]);
    return -1;
  }
  try
  {
    int     sampleRate = 0;
    int     nBits = 0;
    parseInteger(sampleRate, argv[3], "sample rate", 1000, 1000000);
    parseInteger(nBits, argv[4], "bits per sample", 8, 32);
    if (nBits & 7)
      throw std::runtime_error(std::string("invalid bits per sample"));
    // load input file
    std::vector< Event >  events;
    {
      InputFile f(argv[1]);
      while (true)
      {
        Event   tmp(f, double(sampleRate), nBits);
        if (!tmp.gen)
          break;
        events.push_back(tmp);
      }
    }
    std::stable_sort(events.begin(), events.end());
    // open output file
    WAVOutputFile outFile(argv[2], sampleRate, nBits, 2);
    // render audio output
    size_t  curEvent = 0;
    long    curTime = 0L;
    bool    wasSineSweep = false;
    std::list< Generator * >  activeGens;
    DitherGenerator ditherGen(double(sampleRate), nBits,
                              double(0x7FFFFFFF) / double(sampleRate));
    double  outBuf[2];
    while (curEvent < events.size() || activeGens.begin() != activeGens.end())
    {
      while (curEvent < events.size() && events[curEvent].startTime <= curTime)
      {
        activeGens.push_back(events[curEvent].gen);
        curEvent++;
      }
      outBuf[0] = 0.0;
      outBuf[1] = 0.0;
      ditherGen.process(outBuf[0], outBuf[1]);
      bool    isSineSweep = false;
      for (std::list< Generator * >::iterator i = activeGens.begin();
           i != activeGens.end(); )
      {
        if (!((*i)->process(outBuf[0], outBuf[1])))
        {
          i = activeGens.erase(i);
        }
        else
        {
          isSineSweep = (typeid(*(*i)) == typeid(SineSweepGenerator));
          i++;
        }
      }
      if (curEvent >= events.size() && activeGens.begin() == activeGens.end())
        break;
      if (isSineSweep && !wasSineSweep)
      {
        outBuf[0] += 0.5;
        outBuf[1] += 0.5;
      }
      wasSineSweep = isSineSweep;
      outFile.writeData(&(outBuf[0]), 1);
      curTime++;
    }
  }
  catch (std::exception& e)
  {
    fprintf(stderr, "%s: error: %s\n", argv[0], e.what());
    return -1;
  }
  return 0;
}

