
#include "headers.hpp"
#include "dsplib.cpp"
#include "options.cpp"

#define MAX_THD_ORDER           64
#define LEVEL_ADJUST_BASE       -8.9124308
#define FFT_SIZE                65536
#define MIN_WINDOW_SIZE         64
#define MAX_WINDOW_SIZE         262144

struct AnalysisParameters
{
  double  sampleRateIn;
  double  bandWidth;
  bool    isTHDPlusN;
  bool    isTHDVsLevel;
  double  sweepDuration;
  double  minFreq;              // for THD vs. frequency only
  double  maxFreq;              //   -"-
  double  winSizeMultMinF;      //   -"-
  double  winSizeMultMaxF;      //   -"-
  double  minLevel;             // for THD vs. level only
  double  maxLevel;             //   -"-
  double  thdLvlFreq;           //   -"-
  double  winSizeMultLvl;       //   -"-
  double  maxLevelFreq;         //   -"-
  double  maxOverlap;
  int     thdLevelMode;         // 0: absolute
                                // 1: relative to fundamental
                                // 2: relative to total RMS
  int     thdLimit;             // maximum order THD to test
};

class THDAnalyzer
{
 protected:
  const AnalysisParameters& params;
  AudioFilter_FileInput&    sf;
  std::vector< double >     cosBuf;
  std::vector< double >     sinBuf;
  std::vector< std::vector< float > >   sampleBufs;
  long        bufWritePos;
  long        bufFilePos;
  size_t      framesRemaining;
  long        prvPos;
  long        prvWinSize;
  long double phase;
  // --------
  size_t getNextBlock(long pos);
  inline size_t getFrame(long pos)
  {
    if (pos >= (bufFilePos + long(MAX_WINDOW_SIZE)))
      return getNextBlock(pos);
    return size_t((bufWritePos + (pos - bufFilePos))
                  & long(MAX_WINDOW_SIZE - 1));
  }
 public:
  THDAnalyzer(AudioFilter_FileInput& sf_, size_t framesRemaining_,
              const AnalysisParameters& params_);
  virtual ~THDAnalyzer();
  bool analyze(std::vector< double >& lvl, std::vector< double >& thd,
               double freq, std::vector< double >& x,
               std::vector< std::vector< double > >& cSum,
               std::vector< std::vector< double > >& sSum,
               bool noMaxOverlap = true);
};

size_t THDAnalyzer::getNextBlock(long pos)
{
  while (pos >= (bufFilePos + long(MAX_WINDOW_SIZE)))
  {
    long    p = bufFilePos + long(MAX_WINDOW_SIZE);
    size_t  i = size_t(bufWritePos);
    bufFilePos++;
    bufWritePos = (bufWritePos + 1L) & long(MAX_WINDOW_SIZE - 1);
    if (framesRemaining < 1)
    {
      cosBuf[i] = 0.0;
      sinBuf[i] = 0.0;
      for (size_t j = 0; j < sampleBufs.size(); j++)
        sampleBufs[j][i] = 0.0f;
      continue;
    }
    framesRemaining--;
    sf.process();
    if (params.isTHDVsLevel)
    {
      long double omega =
          twoPiL * params.thdLvlFreq / (long double) params.sampleRateIn;
      if (omega <= piL)
      {
        cosBuf[i] = double(std::cos(phase));
        sinBuf[i] = double(std::sin(phase));
        phase += omega;
        if (phase >= piL)
          phase -= twoPiL;
      }
      else
      {
        cosBuf[i] = 0.0;
        sinBuf[i] = 0.0;
      }
      // in THD vs. level mode, apply inverse envelope to normalize levels
      double  x = (double(p) + 0.5)
                  / (params.sweepDuration * params.sampleRateIn);
      double  w = std::pow(params.minLevel / params.maxLevel, x)
                  / params.minLevel;
      for (size_t j = 0; j < sampleBufs.size(); j++)
        sampleBufs[j][i] = float(sf.getOutputBuffer()[j] * w);
    }
    else
    {
      long double x = ((long double) p + 0.5L)
                      / ((long double) params.sweepDuration
                         * params.sampleRateIn);
      long double frq = std::pow((long double) params.maxFreq
                                 / params.minFreq, x)
                        * (long double) params.minFreq;
      long double omega = twoPiL * frq / (long double) params.sampleRateIn;
      if (omega <= piL)
      {
        cosBuf[i] = double(std::cos(phase));
        sinBuf[i] = double(std::sin(phase));
        phase += omega;
        if (phase >= piL)
          phase -= twoPiL;
      }
      else
      {
        cosBuf[i] = 0.0;
        sinBuf[i] = 0.0;
      }
      for (size_t j = 0; j < sampleBufs.size(); j++)
        sampleBufs[j][i] = float(sf.getOutputBuffer()[j]);
    }
  }
  return size_t((bufWritePos + (pos - bufFilePos)) & long(MAX_WINDOW_SIZE - 1));
}

THDAnalyzer::THDAnalyzer(AudioFilter_FileInput& sf_, size_t framesRemaining_,
                         const AnalysisParameters& params_)
  : params(params_),
    sf(sf_),
    cosBuf(MAX_WINDOW_SIZE, 0.0),
    sinBuf(MAX_WINDOW_SIZE, 0.0),
    sampleBufs(size_t(sf_.getChannelCount())),
    bufWritePos(0L),
    bufFilePos(-(long(MAX_WINDOW_SIZE))),
    framesRemaining(framesRemaining_),
    prvPos(0L),
    prvWinSize(0L),
    phase(0.0L)
{
  for (int i = 0; i < sf.getChannelCount(); i++)
    sampleBufs[i].resize(MAX_WINDOW_SIZE, 0.0f);
}

THDAnalyzer::~THDAnalyzer()
{
}

bool THDAnalyzer::analyze(std::vector< double >& lvl,
                          std::vector< double >& thd, double freq,
                          std::vector< double >& x,
                          std::vector< std::vector< double > >& cSum,
                          std::vector< std::vector< double > >& sSum,
                          bool noMaxOverlap)
{
  size_t  channels = sampleBufs.size();
  for (size_t i = 0; i < channels; i++)
  {
    lvl[i] = 0.0;
    thd[i] = 0.0;
  }
  double  winSizeMult = 0.0;
  long    windowSize = 0L;
  long    pos = 0L;
  if (!params.isTHDVsLevel)
  {
    double  f = std::log(freq / params.minFreq)
                / std::log(params.maxFreq / params.minFreq);
    winSizeMult =
        params.winSizeMultMinF
        * std::pow(params.winSizeMultMaxF / params.winSizeMultMinF, f);
    winSizeMult = double(long(winSizeMult + 0.5));
    windowSize = long(params.sampleRateIn * winSizeMult / freq + 0.5);
    windowSize = (windowSize > long(MIN_WINDOW_SIZE) ?
                  (windowSize < long(MAX_WINDOW_SIZE) ?
                   windowSize : long(MAX_WINDOW_SIZE))
                  : long(MIN_WINDOW_SIZE));
    pos = long(params.sampleRateIn * params.sweepDuration * f
               - (double(windowSize) * 0.5) + 0.5);
  }
  else
  {
    winSizeMult = double(long(params.winSizeMultLvl + 0.5));
    windowSize =
        long(params.sampleRateIn * winSizeMult / params.thdLvlFreq + 0.5);
    windowSize = (windowSize > long(MIN_WINDOW_SIZE) ?
                  (windowSize < long(MAX_WINDOW_SIZE) ?
                   windowSize : long(MAX_WINDOW_SIZE))
                  : long(MIN_WINDOW_SIZE));
    pos = long(params.sampleRateIn * params.sweepDuration
               * (std::log(freq / params.maxLevelFreq)
                  / std::log(params.maxLevel / params.minLevel) + 1.0)
               - (double(windowSize) * 0.5) + 0.5);
    freq = params.thdLvlFreq;
  }
  if (prvWinSize > 0L && !noMaxOverlap)
  {
    double  o = (double(pos - prvPos) * params.maxOverlap)
                / (double(windowSize + prvWinSize) * 0.5);
    if (o < 1.0)
      return false;
  }
  prvPos = pos;
  prvWinSize = windowSize;
  int     hCnt = (params.isTHDPlusN ? 1 : params.thdLimit);
  for (int i = 0; i < params.thdLimit; i++)
  {
    double  f = double(i + 1) * freq;
    if (f >= params.bandWidth || f >= (params.sampleRateIn * 0.5))
    {
      hCnt = i;
      break;
    }
    for (size_t j = 0; j < channels; j++)
    {
      cSum[j][i] = 0.0;
      sSum[j][i] = 0.0;
    }
  }
  {
    long        startPos = pos;
    long        endPos = startPos + windowSize;
    startPos = (startPos >= 0L ? startPos : 0L);
    endPos = (endPos < (bufFilePos + long(framesRemaining + MAX_WINDOW_SIZE)) ?
              endPos : (bufFilePos + long(framesRemaining + MAX_WINDOW_SIZE)));
    CosineGen   w((double) windowSize);
    if (!params.isTHDPlusN && hCnt > 0)
    {
      if (hCnt == 1)
      {
        for (long p = startPos; p < endPos; p++)
        {
          double  w_ = cosineToNuttallWindow(w.sample());
          size_t  i = getFrame(p);
          double  c = cosBuf[i];
          double  s = sinBuf[i];
          for (size_t j = 0; j < channels; j++)
          {
            double  x_ = double(sampleBufs[j][i]) * w_;
            cSum[j][0] += (x_ * c);
            sSum[j][0] += (x_ * s);
          }
        }
      }
      else if (channels == 1)
      {
        for (long p = startPos; p < endPos; p++)
        {
          double  w_ = cosineToNuttallWindow(w.sample());
          size_t  i = getFrame(p);
          double  c = cosBuf[i];
          double  s = sinBuf[i];
          double  x_ = double(sampleBufs[0][i]) * w_;
          double  c_ = c * 2.0;
          double  prv2_c = x_;
          double  prv2_s = 0.0;
          c *= x_;
          s *= x_;
          cSum[0][0] += c;
          sSum[0][0] += s;
          int     h = 1;
          do
          {
            double  prv_c = c;
            double  prv_s = s;
            c = c * c_ - prv2_c;
            s = s * c_ - prv2_s;
            prv2_c = prv_c;
            prv2_s = prv_s;
            cSum[0][h] += c;
            sSum[0][h] += s;
          }
          while (++h < hCnt);
        }
      }
      else if (channels == 2)
      {
        for (long p = startPos; p < endPos; p++)
        {
          double  w_ = cosineToNuttallWindow(w.sample());
          size_t  i = getFrame(p);
          double  c = cosBuf[i];
          double  s = sinBuf[i];
          double  xl = double(sampleBufs[0][i]) * w_;
          double  xr = double(sampleBufs[1][i]) * w_;
          cSum[0][0] += (xl * c);
          cSum[1][0] += (xr * c);
          sSum[0][0] += (xl * s);
          sSum[1][0] += (xr * s);
          double  c_ = c * 2.0;
          double  prv2_c = 1.0;
          double  prv2_s = 0.0;
          int     h = 1;
          do
          {
            double  prv_c = c;
            double  prv_s = s;
            c = c * c_ - prv2_c;
            s = s * c_ - prv2_s;
            prv2_c = prv_c;
            prv2_s = prv_s;
            cSum[0][h] += (xl * c);
            cSum[1][h] += (xr * c);
            sSum[0][h] += (xl * s);
            sSum[1][h] += (xr * s);
          }
          while (++h < hCnt);
        }
      }
      else
      {
        for (long p = startPos; p < endPos; p++)
        {
          double  w_ = cosineToNuttallWindow(w.sample());
          size_t  i = getFrame(p);
          double  c = cosBuf[i];
          double  s = sinBuf[i];
          for (size_t j = 0; j < channels; j++)
          {
            double  x_ = double(sampleBufs[j][i]) * w_;
            x[j] = x_;
            cSum[j][0] += (x_ * c);
            sSum[j][0] += (x_ * s);
          }
          double  c_ = c * 2.0;
          double  prv2_c = 1.0;
          double  prv2_s = 0.0;
          int     h = 1;
          do
          {
            double  prv_c = c;
            double  prv_s = s;
            c = c * c_ - prv2_c;
            s = s * c_ - prv2_s;
            prv2_c = prv_c;
            prv2_s = prv_s;
            for (size_t j = 0; j < channels; j++)
            {
              cSum[j][h] += (x[j] * c);
              sSum[j][h] += (x[j] * s);
            }
          }
          while (++h < hCnt);
        }
      }
    }
    else if (params.isTHDPlusN)
    {
      // THD+N mode (FIXME: this does not work correctly)
      if (hCnt < 1)
      {
        for (long p = startPos; p < endPos; p++)
        {
          double  w_ = cosineToNuttallWindow(w.sample());
          size_t  i = getFrame(p);
          for (size_t j = 0; j < channels; j++)
          {
            double  x_ = double(sampleBufs[j][i]) * w_;
            thd[j] += (x_ * x_);
          }
        }
      }
      else
      {
        for (long p = startPos; p < endPos; p++)
        {
          double  w_ = cosineToNuttallWindow(w.sample());
          size_t  i = getFrame(p);
          double  c = cosBuf[i];
          double  s = sinBuf[i];
          for (size_t j = 0; j < channels; j++)
          {
            double  x_ = double(sampleBufs[j][i]) * w_;
            thd[j] += (x_ * x_);
            cSum[j][0] += (x_ * c);
            sSum[j][0] += (x_ * s);
          }
        }
      }
    }
    double  scale = 2.0 / (0.355768 * double(windowSize));
    scale = scale * scale;
    if (hCnt > 0)
    {
      for (size_t i = 0; i < channels; i++)
      {
        lvl[i] = ((cSum[i][0] * cSum[i][0]) + (sSum[i][0] * sSum[i][0]))
                 * scale;
      }
    }
    if (params.isTHDPlusN)
    {
      double  scale2 = 2.0 / (0.255829165552 * double(windowSize));
      for (size_t i = 0; i < channels; i++)
        thd[i] = thd[i] * scale2 - lvl[i];
    }
    else
    {
      for (size_t i = 0; i < channels; i++)
      {
        for (int j = 1; j < hCnt; j++)
          thd[i] += ((cSum[i][j] * cSum[i][j]) + (sSum[i][j] * sSum[i][j]));
        thd[i] *= scale;
      }
    }
  }
  for (size_t i = 0; i < channels; i++)
  {
    lvl[i] = (lvl[i] > 1.0e-16 ? (lvl[i] < 10000.0 ? lvl[i] : 10000.0) : 0.0);
    thd[i] = (thd[i] > 1.0e-16 ? (thd[i] < 10000.0 ? thd[i] : 10000.0) : 0.0);
    if (params.thdLevelMode == 1)
    {
      if (thd[i] < 1.0e-16)
        thd[i] = 0.0;
      else if (thd[i] < lvl[i])
        thd[i] = std::sqrt(thd[i] / lvl[i]);
      else
        thd[i] = 1.0;
    }
    else if (params.thdLevelMode == 2)
    {
      if (thd[i] < 1.0e-16)
        thd[i] = 0.0;
      else
        thd[i] = std::sqrt(thd[i] / (thd[i] + lvl[i]));
    }
    else
    {
      thd[i] = std::sqrt(thd[i]);
    }
    lvl[i] = std::sqrt(lvl[i]);
  }
  return true;
}

// ----------------------------------------------------------------------------

int main(int argc, char **argv)
{
  try
  {
    AnalysisParameters  params;
    params.sampleRateIn = 96000.0;
    params.bandWidth = 50000.0;
    params.isTHDPlusN = false;
    params.isTHDVsLevel = false;
    params.sweepDuration = 60.0;
    params.minFreq = 10.0;
    params.maxFreq = 22000.0;
    params.winSizeMultMinF = 8.0;
    params.winSizeMultMaxF = 256.0;
    params.minLevel = 0.001;
    params.maxLevel = 1.0;
    params.thdLvlFreq = 1000.0;
    params.winSizeMultLvl = 32.0;
    params.maxLevelFreq = 20000.0;
    params.maxOverlap = 20.0;
    params.thdLevelMode = 0;
    params.thdLimit = 7;
    std::string inFileName;
    std::string outFileName;
    double      sampleRateInScale = 1.0;
    double      sampleRateOut = 44100.0;
    double      skipTime = 0.0;
    float       triggerLevel = 0.08f;           // disabled if <= 0
    bool        normalizeLevel = true;
    double      lvlAdjust = 0.0;                // in dB
    {
      CommandLineParser opts(argv[0]);
      opts.addFileName("INFILE", "input file name", inFileName);
      opts.addFileName("OUTFILE", "output file name", outFileName);
      opts.addOption("infile", "i",
                     "input file name",
                     inFileName);
      opts.addOption("outfile", "o",
                     "output file name",
                     outFileName);
      opts.addOption("sample-rate", "r",
                     "output sample rate in Hz, or resample ratio if negative "
                     "(e.g.\n"
                     "        -2.0 will resample 44100 Hz to 88200 Hz)",
                     sampleRateOut, -100.0, 1000000.0);
      opts.addOption("isrm", "m",
                     "factor to multiply assumed input sample rate by; "
                     "it only affects\n"
                     "        the resample ratio, but not the conversion of "
                     "skip time to\n"
                     "        samples",
                     sampleRateInScale, 0.1, 10.0);
      opts.addOption("skip", "k", "start time in seconds or -samples",
                     skipTime);
      opts.addOption("bandwidth", "bw",
                     "ignore harmonics above the specified frequency in Hz",
                     params.bandWidth, 100.0, 100000.0);
      opts.addOption("triggerlevel", "tl",
                     "trigger impulse threshold as a fraction of 0 dBFS, "
                     "disabled if\n"
                     "        zero or negative",
                     triggerLevel, -1.0f, 1.0f);
      opts.addOption("normalizelevel", "n",
                     "normalize average level at 500-2000 Hz to 0 dB",
                     normalizeLevel);
      opts.addOption("thd+n", "thdn",
                     "calculate THD+N instead of THD; this mode is broken, "
                     "do not use\n"
                     "        it",
                     params.isTHDPlusN);
      opts.addOption("thdvslevel", "l",
                     "enable THD vs. level mode instead of THD vs. frequency",
                     params.isTHDVsLevel);
      opts.addOption("sweepduration", "d",
                     "duration of test signal in seconds",
                     params.sweepDuration, 1.0, 1000.0);
      opts.addOption("minfreq", "fmin",
                     "start frequency of sweep in Hz (for THD vs. frequency)",
                     params.minFreq, 1.0, 100000.0);
      opts.addOption("maxfreq", "fmax",
                     "end frequency of sweep",
                     params.maxFreq, 1.0, 100000.0);
      opts.addOption("winsizemultminfreq", "wsmin",
                     "window size in cycles at the start frequency",
                     params.winSizeMultMinF, 2.0, 4096.0);
      opts.addOption("winsizemultmaxfreq", "wsmax",
                     "window size in cycles at the end frequency",
                     params.winSizeMultMaxF, 2.0, 4096.0);
      opts.addOption("minlevel", "lmin",
                     "start level (as a fraction of 0 dBFS) "
                     "in THD vs. level mode",
                     params.minLevel, 0.000001, 1.0);
      opts.addOption("maxlevel", "lmax",
                     "end level in THD vs. level mode",
                     params.maxLevel, 0.000001, 1.0);
      opts.addOption("thdvslvlfreq", "f",
                     "signal frequency in THD vs. level mode",
                     params.thdLvlFreq, 1.0, 100000.0);
      opts.addOption("thdvslvlwinsizemult", "ws",
                     "window size in cycles in THD vs. level mode",
                     params.winSizeMultLvl, 2.0, 4096.0);
      opts.addOption("maxlevelfreq", "lmaxf",
                     "output (FFT) frequency at end level "
                     "in THD vs. level mode",
                     params.maxLevelFreq, 100.0, 500000.0);
      opts.addOption("overlap", "ol",
                     "interpolate output if the window overlap exceeds this "
                     "value",
                     params.maxOverlap, 4.0, 1000.0);
      opts.addOption("thdcalcmode", "tm",
                     "THD calculation mode:\n"
                     "            0: absolute (the same 0 dB reference as for "
                     "the fundamental)\n"
                     "            1: relative to fundamental\n"
                     "            2: relative to total power",
                     params.thdLevelMode, 0, 2);
      opts.addOption("thdlimit", "hmax",
                     "highest order THD to calculate",
                     params.thdLimit, 2, MAX_THD_ORDER);
      opts.addOption("leveladjust", "a",
                     "adjust output level by the specified amount (in dB)",
                     lvlAdjust, -80.0, 80.0);
      if (!opts.parseCommandLine(argc, argv))
        return 0;
      if (!((!params.isTHDVsLevel && params.maxFreq > (params.minFreq * 1.1)) ||
            (params.isTHDVsLevel && params.maxLevel > (params.minLevel * 1.1))))
      {
        throw std::runtime_error(std::string("invalid sweep parameters"));
      }
      lvlAdjust += LEVEL_ADJUST_BASE;
    }

    // load input file
    AudioFilter_FileInput  inFile(inFileName.c_str());
    params.sampleRateIn = inFile.getSampleRate() * sampleRateInScale;
    sf_count_t  filePos =
        (skipTime < 0.0 ?
         sf_count_t(-skipTime + 0.5)
         : sf_count_t(skipTime * inFile.getSampleRate() + 0.5));
    inFile.setPosition(filePos, false);
    int     channels = inFile.getChannelCount();
    sf_count_t  framesLeft = inFile.getFileLength() - filePos;
    if (triggerLevel > 0.0)
    {
      std::vector< float >  prvFrame(size_t(channels), 0.0f);
      while (framesLeft > 0)
      {
        if (!inFile.process())
        {
          framesLeft = 0;
          break;
        }
        filePos++;
        framesLeft--;
        bool    isTriggered = true;
        for (int i = 0; i < channels; i++)
        {
          float   x = float(inFile.getOutputBuffer()[i]);
          if (!(prvFrame[i] >= triggerLevel && prvFrame[i] >= x))
            isTriggered = false;
          prvFrame[i] = x;
        }
        if (isTriggered)
        {
          filePos = filePos - 2;
          framesLeft = framesLeft + 2;
          break;
        }
      }
    }
    {
      sf_count_t  sweepFrames =
          sf_count_t(params.sweepDuration * params.sampleRateIn + 0.5);
      framesLeft = (framesLeft < sweepFrames ? framesLeft : sweepFrames);
    }
    if (channels < 1 || framesLeft < 1)
      throw std::runtime_error(std::string("there is no data to analyze"));
    if (sampleRateOut < 0.0)
      sampleRateOut = -sampleRateOut * inFile.getSampleRate();
    if (!(sampleRateOut >= 1000.0 && sampleRateOut <= 1000000.0))
      throw std::runtime_error(std::string("invalid output sample rate"));
    inFile.setPosition(filePos, false);

    THDAnalyzer analyzer(inFile, size_t(framesLeft), params);

    std::vector< std::vector< double > >  fftBufs(size_t(channels) << 1);
    double  normalizeMinFreq = 0.0;
    double  normalizeMaxFreq = 0.0;
    std::vector< double > levels(size_t(channels), 0.0);
    long    levelCnt = 0L;
    if (normalizeLevel)
    {
      if (!params.isTHDVsLevel)
      {
        normalizeMinFreq = 500.0;
        normalizeMaxFreq = 2000.0;
      }
      else
      {
        normalizeMinFreq = std::sqrt(params.minLevel / params.maxLevel);
        normalizeMaxFreq = std::sqrt(normalizeMinFreq);
        normalizeMinFreq *= params.maxLevelFreq;
        normalizeMaxFreq *= params.maxLevelFreq;
      }
    }
    for (size_t i = 0; i < fftBufs.size(); i++)
      fftBufs[i].resize(FFT_SIZE, 0.0);
    {
      std::vector< double > lvl(size_t(channels), 0.0);
      std::vector< double > thd(size_t(channels), 0.0);
      std::vector< double > x(size_t(channels), 0.0);
      std::vector< std::vector< double > >  cSum((size_t) channels);
      std::vector< std::vector< double > >  sSum((size_t) channels);
      for (int i = 0; i < channels; i++)
      {
        cSum[i].resize(MAX_THD_ORDER, 0.0);
        sSum[i].resize(MAX_THD_ORDER, 0.0);
      }
      int     lastBin = 0;
      for (int i = 1; i <= (FFT_SIZE >> 1); i++)
      {
        double  f = double(i) * sampleRateOut / double(FFT_SIZE);
        if (!analyzer.analyze(lvl, thd, f, x, cSum, sSum,
                              (i == 1 || i == (FFT_SIZE >> 1))))
        {
          continue;
        }
        int     n = (i < (FFT_SIZE >> 1) ? (i << 1) : 1);
        for (int j = 0; j < channels; j++)
        {
          fftBufs[j << 1][n] = ((i & 1) == 0 ? lvl[j] : -(lvl[j]));
          fftBufs[(j << 1) + 1][n] = ((i & 1) == 0 ? thd[j] : -(thd[j]));
        }
        if (f >= normalizeMinFreq && f <= normalizeMaxFreq)
        {
          for (int j = 0; j < channels; j++)
            levels[j] += (lvl[j] * lvl[j]);
          levelCnt += long(channels);
        }
        for (int k = (lastBin + 1); k < i; k++)
        {
          // interpolate previous bins
          f = double(k - lastBin) / double(i - lastBin);
          int     n0 = (lastBin < (FFT_SIZE >> 1) ? (lastBin << 1) : 1);
          int     n_ = (k < (FFT_SIZE >> 1) ? (k << 1) : 1);
          for (int j = 0; j < channels; j++)
          {
            double  lvl0 = std::fabs(fftBufs[j << 1][n0]);
            double  lvl1 = std::fabs(fftBufs[j << 1][n]);
            double  thd0 = std::fabs(fftBufs[(j << 1) + 1][n0]);
            double  thd1 = std::fabs(fftBufs[(j << 1) + 1][n]);
            lvl[j] = lvl0 + ((lvl1 - lvl0) * f);
            thd[j] = thd0 + ((thd1 - thd0) * f);
          }
          f = double(k) * sampleRateOut / double(FFT_SIZE);
          for (int j = 0; j < channels; j++)
          {
            fftBufs[j << 1][n_] = ((k & 1) == 0 ? lvl[j] : -(lvl[j]));
            fftBufs[(j << 1) + 1][n_] = ((k & 1) == 0 ? thd[j] : -(thd[j]));
          }
          if (f >= normalizeMinFreq && f <= normalizeMaxFreq)
          {
            for (int j = 0; j < channels; j++)
              levels[j] += (lvl[j] * lvl[j]);
            levelCnt += long(channels);
          }
        }
        lastBin = i;
      }
    }
    for (size_t i = 0; i < fftBufs.size(); i++)
      realFFT(&(fftBufs[i].front()), FFT_SIZE, true);
    double  scaleFac1 = std::pow(10.0, lvlAdjust / 20.0)    // for fundamental
                        * double(FFT_SIZE >> 1);
    double  scaleFac2 = scaleFac1;                          // for THD products
    if (levelCnt > 0L)
    {
      double  maxLevel = 0.0;
      for (int i = 0; i < channels; i++)
        maxLevel = (levels[i] > maxLevel ? levels[i] : maxLevel);
      if (maxLevel > 0.0)
      {
        double  avgLevel = 0.0;
        for (int i = 0; i < channels; i++)
          avgLevel += (levels[i] >= (maxLevel * 0.25) ? levels[i] : maxLevel);
        avgLevel = std::sqrt(avgLevel / double(levelCnt));
        scaleFac1 /= avgLevel;
        if (params.thdLevelMode < 1)
          scaleFac2 = scaleFac1;
      }
    }

    // write output file
    {
      AudioOutputFile   outFile(outFileName.c_str(), sampleRateOut, channels);
      std::vector< double > buf(size_t(channels), 0.0);
      for (size_t i = 0; i < FFT_SIZE; i++)
      {
        for (int j = 0; j < channels; j++)
          buf[j] = fftBufs[j << 1][i] * scaleFac1;
        outFile.writeSampleFrame(&(buf.front()));
      }
      for (size_t i = 0; i < FFT_SIZE; i++)
      {
        for (int j = 0; j < channels; j++)
          buf[j] = fftBufs[(j << 1) + 1][i] * scaleFac2;
        outFile.writeSampleFrame(&(buf.front()));
      }
    }
  }
  catch (std::exception& e)
  {
    std::fprintf(stderr, "%s: %s\n", argv[0], e.what());
    return -1;
  }
  return 0;
}

